// MA95.CPP : WebSphere MQ Support Pac MA95 - REXX support for z/OS and Windows
//
//   (C) Copyright IBM Corporation 1997, 2010
//
// This is a nearly full implementation of the MQ API with these
//      omissions :
//
//      INQ and SET only take a single attribute.
//
//      MQCONNX not implemented
//
//      MQBEGIN not implemented
//
// The MQ API is extended in several ways:
//
//
//      * The QM Handle is not externally referenced, as this is held
//            within this program
//
//      * Windows version knows about threads. It will allow multiple
//            Rexx threads to talk to multiple (or the same) Queue Managers
//
//      * Structures that get updated are managed via seperate
//            input and output areas, so avoiding the need call
//            to do all that tedious rebuilding for each
//
//      * Output areas contain .ZLIST which contains a list of all the
//            components within the output stem (without the dot)
//
//      * A Browse function is provided which just returns the
//            message data
//
//      * A Header Extraction function will interprete Dead Letter
//           and Transmission Headers from suitable messages
//
//      * Expansion of Events into components (like the Header Extraction)
//
//      * Interpretation of Trigger Messages and Trigger Data
//
//      * Function to only set up MQ Constants
//
//
//   In order to use this Rexx/MQ Interface, initialization function
//        must be called before usage.
//
//   For Windows only this is done by a RxFuncAdd('RXMQINIT','RXMQx','RXMQINIT')
//        call to let Rexx know about the DLL, and then a
//        rcc = RXMQINIT() to call the initialization function.
//        As part of this initialization function, all the other
//        RXMQ functions are registered with Rexx.
//
//   There is no equivalent termination function, as the MQ routines
//        attach to process termination to stop access at this
//        time, and a Termination operation would interfere
//        with this processing. However, a function called
//        RXMQTERM will deregister the interface functions WITHOUT
//        doing end-of-process actions.
//
//   All the RXMQ functions return a standard Rexx Return string.
//        This is structured so that the numeric Return Code
//        (which may be negative) is obtained by a word(RCC,1) call.
//
//   The Return Code for an operation can be negative to
//        show that this DLL has detected the error, otherwise
//        it will be the MQ Completion Code
//
//        The Return String is in text format as follows:
//
//            Word 1 : Return Code
//            Word 2 : MQ Completion Code (or 0 if MQ not done)
//            Word 2 : MQ Reason     Code (or 0 if MQ not done)
//            Word 4 : RXMQ... function being run
//            Word > : OK or an helpful error message
//
//   In addition, the current (ie: the settings last set) values are
//      available in these variables:
//
//   Windows:
//          RXMQN.LASTRC    -> current operation Return Code
//          RXMQN.LASTCC    -> current operation MQ Completion Code
//          RXMQN.LASTAC    -> current operation MQ Reason     Code
//          RXMQN.LASTOP    -> current operation RXMQ function
//          RXMQN.LASTMSG   -> current operation Return String
//          or
//          RXMQT.LASTRC    -> current operation Return Code
//          RXMQT.LASTCC    -> current operation MQ Completion Code
//          RXMQT.LASTAC    -> current operation MQ Reason     Code
//          RXMQT.LASTOP    -> current operation RXMQ function
//          RXMQT.LASTMSG   -> current operation Return String
//   MVS:
//          RXMQV.LASTRC    -> current operation Return Code
//          RXMQV.LASTCC    -> current operation MQ Completion Code
//          RXMQV.LASTAC    -> current operation MQ Reason     Code
//          RXMQV.LASTOP    -> current operation RXMQ function
//          RXMQV.LASTMSG   -> current operation Return String
//   Both:
//          RXMQ.LASTRC     -> current operation Return Code
//          RXMQ.LASTCC     -> current operation MQ Completion Code
//          RXMQ.LASTAC     -> current operation MQ Reason     Code
//          RXMQ.LASTOP     -> current operation RXMQ function
//          RXMQ.LASTMSG    -> current operation Return String
//
//   As a nice little bonus, there are also defined lots of variables
//        called RXMQ.RCMAP.nn , where nn is a MQAC number, whose
//        value is the name of the Reason Code (MQRC_ERROR_THING).
//        These variables can then be used in the Rexx Exec to get
//        the name from the Reason code via a
//                 interpret 'name = RXMQ.RCMAP.'word(rcc,1)
//        statement.
//
//        A TERM function leaves the MQ Rexx Variables known; if removed via
//        a DROP call, they may be restored via RXMQCONS (registering if
//        apt via RxFuncAdd('RXMQCONS',RXQMx','RXMQCONS') if needed).
//
//
//   Tracing MAY be enabled (if this module has been compiled with
//        the relevant setting of the DEBUG macro) by setting
//        values into a Rexx Variable call RXMQTRACE. The settings are:
//
//                              CONN  -> mqconn
//                              DISC  -> mqdisc
//                              OPEN  -> mqopen
//                              CLOSE -> mqclose
//                              GET   -> mqget
//                              PUT   -> mqput
//                              PUT1  -> mqput1
//                              INQ   -> mqinq
//                              SET   -> mqset
//                              SUB   -> mqsub
//                              CMIT  -> mqcmit
//                              BACK  -> mqback
//
//                              BRO   -> Browse extension
//                              HXT   -> Header extraction extension
//                              EVENT -> Event determination extension
//                              TM    -> Trigger message extension
//                              COM   -> Command interface
//                              MQV   -> Debug a RXMQV
//
//                              MMD   -> Rexx stem var --> MQMD
//                              MOD   -> Rexx stem var --> MQOD
//                              MPO   -> Rexx stem var --> MQPMO
//                              MGO   -> Rexx stem var --> MQGMO
//                              MSD   -> Rexx stem var --> MQSD
//
//                              BMD   -> MQMD  --> Rexx stem var
//                              BOD   -> MQOD  --> Rexx stem var
//                              BPO   -> MQPMO --> Rexx stem var
//                              BGO   -> MQGMO --> Rexx stem var
//                              BSD   -> MQSD  --> Rexx stem var
//
//                              SK    -> Return Code processing
//                              TR    -> Thread based processing
//
//                              INIT  -> initialization processing
//                              TERM  -> Deregistration processing
//
//                              *     -> Trace everything!!!
//
//
//  Thread Support: In order to provide Thread support for non-IBM Rexxs,
//                  the rahmqnTR Class library records the current thread upon
//                  which the MQCONN & MQOPEN is issued. This is then
//                  available for policing by the other functions. This
//                  results in calls to see that the Connection/Open thread
//                  is the current one for which an access is made.
//
//                   RC -98 indicates that the thread is already connected
//                      -97                     owning thread is not the
//                                              one doing the access
//
//                  This also means that the scanning and allocation of
//                  objects from the global process-based objects must be
//                  protected against multiple accesses. Therefore, an
//                  anomous private Mutex is created at INIT time (which is
//                  thus available in all threads for the process) to
//                  engender this single threading. The Mutex is removed
//                  in the TERM call. This locking is active OVER MQ calls
//                  in the CONN, OPEN, CLOSE and DISC functions.
//
//                   RC -96 indicates that the Mutex was not acquired
//                      -95                                  released
//
//                  The TIMEOUT for the acquisition is set to MUTEXWAIT. This
//                      should be set to a big enough value to cope with long
//                      remote-type accesses.
//
//
//
//   Support  : This program is not supported in any way by IBM.
//              It is distributed only to show techniques for
//              Message Queueing (if so distributed).
//              It may be freely modified and adapted by your
//              installation.
//
//
//

//
// Standard Header file includes
//
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <time.h>
 #include <memory.h>
 #include <ctype.h>
 #include <errno.h>

 #ifdef __MVS__
  #include <iostream.h>
 #else
  #include <iostream>
 #endif

#if _MSC_VER
   #include <afxmt.h>
#endif

//
// Rexx Header file includes
//
#define INCL_RXSHV
#define INCL_RXFUNC
#include <rexx.h>

//
// MQ/Windows/MVS Header file includes
//

#include <cmqc.h>
#include <cmqcfc.h>

using namespace std;

#ifdef __MVS__
//
// REXX Functions Windows/MVS type
// And redefine entry points so that externalized entries are in assembler stub
//
  #define RXMQINIT CPPMINIT
  #define RXMQTERM CPPMTERM
  #define RXMQCONS CPPMCONS
  #define RXMQCONN CPPMCONN
  #define RXMQOPEN CPPMOPEN
  #define RXMQCLOS CPPMCLOS
  #define RXMQDISC CPPMDISC
  #define RXMQCMIT CPPMCMIT
  #define RXMQBACK CPPMBACK
  #define RXMQPUT  CPPMPUT
  #define RXMQPUT1 CPPMPUT1
  #define RXMQGET  CPPMGET
  #define RXMQINQ  CPPMINQ
  #define RXMQSET  CPPMSET
  #define RXMQSUB  CPPMSUB
  #define RXMQBRWS CPPMBRWS
  #define RXMQHXT  CPPMHXT
  #define RXMQEVNT CPPMEVNT
  #define RXMQTM   CPPMTM
  #define RXMQC    CPPMC
  #define RXMQV    CPPMV
  #define RXMQVC   CPPMVC
//
// Make this structure preserved between calls
//
  #pragma variable(gmqo, NORENT)

//
// Dummy function for MVS environment
//
  MQULONG GetCurrentThreadId() {return 1;}
//
// Function to print to stdout (REXX terminal or standard output)
//
#define PRINTF(s) RXSAY(s, strlen(s));
#else
#define PRINTF(s) printf(s); \
                  fflush(NULL) ;  // Required to flush buffer
#endif


#define FTYPE extern "C" int APIENTRY

#define RXMQPARM ( PSZ       afuncname /*Rexx Func name       */\
                 , MQLONG    aargc     /*     Number of parms */\
                 , PRXSTRING aargv     /*     Parms           */\
                 , PSZ       aqname    /*     Cur Queue       */\
                 , PRXSTRING aretstr   /*     Return String   */\
                 )

//
// Trace/Return variables in Windows/MVS
//
#ifdef __MVS__
  #define PREFIX   "RXMQV."
  #define TRACEVAR "RXMQVTRACE"
  #define RCMAP    "RXMQV.RCMAP.%lu"
#endif

#ifdef _RXMQN
  #define PREFIX   "RXMQN."
  #define TRACEVAR "RXMQNTRACE"
  #define RCMAP    "RXMQN.RCMAP.%lu"
  #define THISDLL  "RXMQN"
 //
 // This is a table of functions to be registered with Rexx
 //
  static PSZ funcs??(??) = {"RXMQV"       ,
                            "RXMQCONN"    ,  "RXMQNCONN"   ,
                            "RXMQDISC"    ,  "RXMQNDISC"   ,
                            "RXMQOPEN"    ,  "RXMQNOPEN"   ,
                            "RXMQCLOS"    ,  "RXMQNCLOSE"  ,
                            "RXMQGET"     ,  "RXMQNGET"    ,
                            "RXMQPUT"     ,  "RXMQNPUT"    ,
                            "RXMQINQ"     ,  "RXMQNINQ"    ,
                            "RXMQSET"     ,  "RXMQNSET"    ,
#ifdef _v7
                            "RXMQSUB"     ,  "RXMQNSUB"    ,
#endif
                            "RXMQCMIT"    ,  "RXMQNCMIT"   ,
                            "RXMQBACK"    ,  "RXMQNBACK"   ,
                            "RXMQBRWS"    ,  "RXMQNBROWSE" ,
                            "RXMQHXT"     ,  "RXMQNHXT"    ,
                            "RXMQEVNT"    ,  "RXMQNEVENT"  ,
                            "RXMQTM"      ,  "RXMQNTM"     ,
                            "RXMQPUT1"    ,  "RXMQNPUT1"   ,
                            "RXMQC"       ,  "RXMQNC"      ,
                            "RXMQCONS"    ,  "RXMQNCONS"   ,
                            "RXMQTERM"    ,  "RXMQNTERM"
              } ;
 static MQULONG numfuncs = sizeof(funcs)/sizeof(char *) ;
#endif

#ifdef _RXMQT
  #define PREFIX   "RXMQT."
  #define RCMAP    "RXMQT.RCMAP.%lu"
  #define TRACEVAR "RXMQTTRACE"
  #define THISDLL  "RXMQT"
//
// This is a table of functions to be registered with Rexx
//
 static PSZ funcs??(??) = {"RXMQV"       ,
                           "RXMQCONN"    ,  "RXMQTCONN"   ,
                           "RXMQDISC"    ,  "RXMQTDISC"   ,
                           "RXMQOPEN"    ,  "RXMQTOPEN"   ,
                           "RXMQCLOS"    ,  "RXMQTCLOSE"  ,
                           "RXMQGET"     ,  "RXMQTGET"    ,
                           "RXMQPUT"     ,  "RXMQTPUT"    ,
                           "RXMQINQ"     ,  "RXMQTINQ"    ,
                           "RXMQSET"     ,  "RXMQTSET"    ,
#ifdef _v7
                           "RXMQSUB"     ,  "RXMQTSUB"    ,
#endif
                           "RXMQCMIT"    ,  "RXMQTCMIT"   ,
                           "RXMQBACK"    ,  "RXMQTBACK"   ,
                           "RXMQBRWS"    ,  "RXMQTBROWSE" ,
                           "RXMQHXT"     ,  "RXMQTHXT"    ,
                           "RXMQEVNT"    ,  "RXMQTEVENT"  ,
                           "RXMQTM"      ,  "RXMQTTM"     ,
                           "RXMQPUT1"    ,  "RXMQTPUT1"   ,
                           "RXMQC"       ,  "RXMQTC"      ,
                           "RXMQCONS"    ,  "RXMQTCONS"   ,
                           "RXMQTERM"    ,  "RXMQTTERM"
                          } ;
 static MQULONG numfuncs = sizeof(funcs)/sizeof(char *) ;
#endif

//
// Define type for output messages struct
//
 typedef struct _RETMSG
                {
                 int    retcode  ;  // Internal (minus) retcode
                 char * retmsgc  ;  // Text part of message
                } RETMSG         ;
 typedef RETMSG * PRETMSG        ;

//
// Define global structure defaults
//
 MQOD    od_default    = {MQOD_DEFAULT}   ;
 MQSD    sd_default    = {MQSD_DEFAULT}   ;
 MQMD2   md_default    = {MQMD2_DEFAULT}  ;
 MQPMO   pmo_default   = {MQPMO_DEFAULT}  ;
 MQGMO   gmo_default   = {MQGMO_DEFAULT}  ;

//
//  Global debug variable for controlling current trace status
//
 MQULONG tracebits;                   // Variable to contain current trace status

  #define ZERO  0x00000000
  #define CONN  0x80000000
  #define DISC  0x40000000
  #define OPEN  0x20000000
  #define CLOSE 0x10000000
  #define GET   0x08000000
  #define PUT   0x04000000
  #define PUT1  0x02000000
  #define INQ   0x01000000
  #define SET   0x00800000
  #define CMIT  0x00400000
  #define BACK  0x00200000
  #define BRO   0x00100000
  #define HXT   0x00080000
  #define EVENT 0x00040000
  #define TM    0x00020000
  #define MQV   0x00010000
  #define MMD   0x00008000
  #define MOD   0x00004000
  #define MPO   0x00002000
  #define MGO   0x00001000
  #define BMD   0x00000800
  #define BOD   0x00000400
  #define BPO   0x00000200
  #define BGO   0x00000100
  #define SK    0x00000080
  #define TR    0x00000040
  #define INIT  0x00000020
  #define TERM  0x00000010
  #define COM   0x00000008
  #define SUB   0x00000004
  #define MSD   0x00000002
  #define BSD   0x00000001
  #define ALL   0xFFFFFFFF

//
//  The DEBUG macro is used to enable tracing. This may be perminantly
//            enabled/disabled, or dynamically under the control
//            of the RXMQTRACE Rexx variable.
//
//
//  DEBUG settings :
//
//  DEBUG  -> if ...    for dynamic debugging based on RXMQTRACE
//  DEBUG  -> cout <<   for static trace everything debugging
//  DEBUG  -> ;         for stop all tracing
//
//  DEBUGW -> if ...    for dyna1ic debugging based on RXMQTRACE
//  DEBUGW ->           for stat2c trace everything debugging
//  DEBUGW -> ;         for stop3all tracing
//
//  DEBUGS -> cout <<   for    tracing in istrace (lots of lines!!!)
//  DEBUGS -> ;         for no tracing in istrace routine
//
//
//  The setting are thus:
//
// #define DEBUG(p1,p2) if (tracebits & p1 ) cout << p2 << endl
// #define DEBUG(p1,p2) cout << p2 << endl
// #define DEBUG(p1,p2)
//
// #define DEBUGW(p1,p2,p3,p4) if (tracebits & p1) { cout.p2 ; cout.p3 ; cout.p4 ; }
// #define DEBUGW(p1,p2,p3,p4) { cout.p2 ; cout.p3 ; cout.p4 ; }
// #define DEBUGW(p1,p2,p3,p4)
//
// #define DEBUGS(p1) cout << p1 << endl
// #define DEBUGS(p1)
//

#define DEBUG(p1,p2) if (tracebits & p1 ) cout << p2 << endl
#define DEBUGW(p1,p2,p3,p4) if (tracebits & p1 ) { cout.p2 ; cout.p3 ; cout.p4 ; }
#define DEBUGS(p1)


//
// Global definitions
//
//  MINQS     -> Start scan of Queues allowed to process (not 0!)
//  MAXQS     -> maximum number of Queues allowed to process
//  MUTEXWAIT -> wait time for critical section mutex (in ms)
//
 #define MINQS 1
 #define MAXQS 100
 #define MUTEXWAIT 30000
 #define MAXCOMMLEN 5000

//
// All the actual MQ calls are preformed within the rahmqnTR
//     class library. This is brought in at compile time
//     for ease of customer compilation. There is no problem
//     about using the Linkage editor to bind at link time
//     if this is so desired.
//
//     !!! Note that this version of rahmqn has info about Threads
//     !!! within it. It is the responsibility of the caller to ensure
//     !!! that the current thread is the one that was used when the MQCONN
//     !!! was issued (if this is not done, MQ will return Invalid Handle
//     !!! for the QM messages etc.)
//

//
// rahmqnTR Class Library - Thread aware
//
class rahmqntr
 {
   public :

    char        qm  ??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name
    char        q   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name
    MQHCONN     qmh                              ; //QM handle
    MQHOBJ      qh                               ; //Q  handle
    MQHOBJ      sh                               ; //Subscription handle for MQSUB
    MQLONG      rc                               ; //Return Code
    MQLONG      ac                               ; //Reason Code
    MQLONG      lastrc                           ; //Prev Return Code
    MQLONG      lastac                           ; //Prev Reason Code
    MQULONG     threadid                         ; //Thread for MQCONN


    rahmqntr ( void )               ;  //Constructor for new  use of QM
    rahmqntr ( MQHCONN const uqmh)  ;  //Constructor for same use of QM

    MQLONG      conn ( char * const uqm )   ;  //Connect with QM name supplied

    MQLONG      disc ( void )               ;  //Disconnect from QM

    void        cleanup ( void )            ;  //Close and Disconnect

    MQOD        odopen                      ;  //OD for MQOPEN usage
    MQLONG      open (char * const uq    ,     //  with Q name supplied
                      MQLONG const uopts )  ;  //
    MQLONG      open (MQOD * const uod   ,     //  with OD and options
                      MQLONG const uopts )  ;  //

    MQLONG      sub  (MQSD * const subdesc) ;  //  MQSUB with SD

    MQLONG      close (MQLONG const uopts ) ;  //  MQCLOSE call

    MQLONG      commit   ( void ) ; //Syncpoint UOW
    MQLONG      rollback ( void ) ; //Backout   UOW

    MQLONG      inq ( MQLONG   const  uselnum ,             //MQINQ call
                      MQLONG          usels   ,
                      MQLONG   const  uinum   ,
                      MQLONG *        uiattr  ,
                      MQLONG   const  ucharl  ,
                      char   *        ucattr  ) ;

    MQLONG      set ( MQLONG   const uselnum  ,             //MQSET call
                      MQLONG         usels    ,
                      MQLONG   const uinum    ,
                      MQLONG *       uiattr   ,
                      MQLONG   const ucharl   ,
                      char   *       ucattr   ) ;

    MQMD2       pmd                             ;           //PUT Message Descriptor
    MQLONG      put ( MQPMO  * const  uopts   ,             //MQPUT call
                      MQLONG   const  bufferl ,
                      char   * const  buffer  ) ;

    MQLONG      put ( MQMD2  *        upmd    ,
                      MQPMO  * const  uopts   ,
                      MQLONG   const  bufferl ,
                      char   * const  buffer  ) ;

    MQLONG      put1 ( MQMD2  *        upmd    ,             //MQPUT1 call
                       MQOD   * const  uod     ,
                       MQPMO  * const  uopts   ,
                       MQLONG   const  bufferl ,
                       char   * const  buffer  ) ;

    MQMD2       bmd                                  ;      //Browse Message Descriptor
    MQGMO       bmo                                  ;      //Browse MQGET options
    MQLONG      browse ( MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  ) ;

    MQMD2       gmd                               ;         //GET Message Descriptor
    MQLONG      get ( MQGMO  * const  uopts     ,           //MQGET call
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  ) ;

    MQLONG      get ( MQMD2  *        ugmd      ,
                      MQGMO  * const  uopts     ,
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  ) ;


    double      gettimeconn     () { return (time_conn     ) ; }
    double      gettimedisc     () { return (time_disc     ) ; }
    double      gettimeopen     () { return (time_open     ) ; }
    double      gettimesub      () { return (time_sub      ) ; }
    double      gettimeclose    () { return (time_close    ) ; }
    double      gettimeinq      () { return (time_inq      ) ; }
    double      gettimeset      () { return (time_set      ) ; }
    double      gettimeput      () { return (time_put      ) ; }
    double      gettimeput1     () { return (time_put1     ) ; }
    double      gettimeget      () { return (time_get      ) ; }
    double      gettimecommit   () { return (time_commit   ) ; }
    double      gettimerollback () { return (time_rollback ) ; }
    double      getcputconn     () { return (cput_conn     / CLOCKS_PER_SEC) ; }
    double      getcputdisc     () { return (cput_disc     / CLOCKS_PER_SEC) ; }
    double      getcputopen     () { return (cput_open     / CLOCKS_PER_SEC) ; }
    double      getcputsub      () { return (cput_sub      / CLOCKS_PER_SEC) ; }
    double      getcputclose    () { return (cput_close    / CLOCKS_PER_SEC) ; }
    double      getcputinq      () { return (cput_inq      / CLOCKS_PER_SEC) ; }
    double      getcputset      () { return (cput_set      / CLOCKS_PER_SEC) ; }
    double      getcputput      () { return (cput_put      / CLOCKS_PER_SEC) ; }
    double      getcputput1     () { return (cput_put1     / CLOCKS_PER_SEC) ; }
    double      getcputget      () { return (cput_get      / CLOCKS_PER_SEC) ; }
    double      getcputcommit   () { return (cput_commit   / CLOCKS_PER_SEC) ; }
    double      getcputrollback () { return (cput_rollback / CLOCKS_PER_SEC) ; }

 private :

    double      time_conn      ;
    double      time_disc      ;
    double      time_open      ;
    double      time_sub       ;
    double      time_close     ;
    double      time_inq       ;
    double      time_set       ;
    double      time_put       ;
    double      time_put1      ;
    double      time_get       ;
    double      time_commit    ;
    double      time_rollback  ;

    double      cput_conn      ;
    double      cput_disc      ;
    double      cput_open      ;
    double      cput_sub       ;
    double      cput_close     ;
    double      cput_inq       ;
    double      cput_set       ;
    double      cput_put       ;
    double      cput_put1      ;
    double      cput_get       ;
    double      cput_commit    ;
    double      cput_rollback  ;
 } ;

//
// rahmqntr member function definitions
//

//
// Constructor functions
//
//     without a parm, then this is a new use of a QM
//     with          , then this is an additional access on a
//                     previously connected QM
//
rahmqntr::rahmqntr ( void )
{
  memset(qm,0,sizeof(qm))   ;
  memset(q ,0,sizeof(q ))   ;
  qmh      = 0              ;
  qh       = 0              ;
  sh       = 0              ;
  rc       = 0              ;
  ac       = 0              ;
  lastrc   = 0              ;
  lastac   = 0              ;
  threadid = 0              ;
  memcpy(&odopen, &od_default, sizeof(MQOD)) ;
  memset(&pmd,0,sizeof(pmd))                 ;
  memset(&bmd,0,sizeof(bmd))                 ;
  memset(&gmd,0,sizeof(gmd))                 ;

  cput_conn      = 0 ;
  cput_disc      = 0 ;
  cput_open      = 0 ;
  cput_sub       = 0 ;
  cput_close     = 0 ;
  cput_inq       = 0 ;
  cput_set       = 0 ;
  cput_put       = 0 ;
  cput_put1      = 0 ;
  cput_get       = 0 ;
  cput_commit    = 0 ;
  cput_rollback  = 0 ;
  time_conn      = 0 ;
  time_disc      = 0 ;
  time_open      = 0 ;
  time_sub       = 0 ;
  time_close     = 0 ;
  time_inq       = 0 ;
  time_set       = 0 ;
  time_put       = 0 ;
  time_put1      = 0 ;
  time_get       = 0 ;
  time_commit    = 0 ;
  time_rollback  = 0 ;

  return             ;
}

rahmqntr::rahmqntr ( MQHCONN const uqmh )
{
  memset(qm,0,sizeof(qm))   ;
  memset(q ,0,sizeof(q ))   ;
  qmh      = uqmh           ;
  qh       = 0              ;
  sh       = 0              ;
  rc       = 0              ;
  ac       = 0              ;
  lastrc   = 0              ;
  lastac   = 0              ;
  threadid = 0              ;
  memcpy(&odopen, &od_default, sizeof(MQOD)) ;
  memset(&pmd,0,sizeof(pmd))                 ;
  memset(&bmd,0,sizeof(bmd))                 ;
  memset(&gmd,0,sizeof(gmd))                 ;

  cput_conn      = 0 ;
  cput_disc      = 0 ;
  cput_open      = 0 ;
  cput_sub       = 0 ;
  cput_close     = 0 ;
  cput_inq       = 0 ;
  cput_set       = 0 ;
  cput_put       = 0 ;
  cput_put1      = 0 ;
  cput_get       = 0 ;
  cput_commit    = 0 ;
  cput_rollback  = 0 ;
  time_conn      = 0 ;
  time_disc      = 0 ;
  time_open      = 0 ;
  time_sub       = 0 ;
  time_close     = 0 ;
  time_inq       = 0 ;
  time_set       = 0 ;
  time_put       = 0 ;
  time_put1      = 0 ;
  time_get       = 0 ;
  time_commit    = 0 ;
  time_rollback  = 0 ;

  return             ;
}

//
// rahmqntr destructor functions
//
//   these are not provided
//
//    (Note that a disc call removes access to ALL
//     instances of rahmqnTR sharing the Handle. as I
//     am not aiming at a 'proper' C++ implementation
//     I have not provided code to cope with this event).
//

//
// rahmqntr functions to access a Queue Manager
//

//
// conn  will connect to a Queue Manager
//       saving the QMH
//
//       Parm is Queue Manager Name which is the
//       one to be used
//
MQLONG  rahmqntr::conn ( char * const uqm )
{
  time_t c1, c2, t1, t2         ;

  strncpy(qm,uqm,sizeof(qm)-1)  ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCONN ( qm, &qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  if (rc == 0 ) threadid = GetCurrentThreadId() ;

  cput_conn += 1.0000 * ( c2 - c1 )      ;
  time_conn += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// disc  will disconnect from a Queue Manager
//       deleting the QMH
//
MQLONG  rahmqntr::disc ( void )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQDISC ( &qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_disc += 1.0000 * ( c2 - c1 )      ;
  time_disc += 1.0000 * difftime(t2,t1)  ;

  qmh      = 0 ;
  threadid = 0 ;

  return (ac)  ;
}

//
// commit will Syncpoint ALL accesses to the
//       QM on the given QMH (this will effect
//       other instances of the Class if the
//       QMH is being shared) in the current UOW
//
MQLONG  rahmqntr::commit ( void )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCMIT ( qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_commit += 1.0000 * ( c2 - c1 )      ;
  time_commit += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// rollback will backout ALL accesses to the
//       QM on the given QMH (this will effect
//       other instances of the Class if the
//       QMH is being shared) in the current UOW
//
MQLONG  rahmqntr::rollback ( void )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQBACK ( qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_rollback += 1.0000 * ( c2 - c1 )      ;
  time_rollback += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// open  will provide access to a Queue
//
//       The OPEN options are always supplied
//           by the caller.
//
MQLONG  rahmqntr::open ( char * const uq
                       , MQLONG const uopts
                       )
{
  time_t c1, c2, t1, t2       ;

  strncpy(q,uq,sizeof(q)-1)   ;

  memcpy(&odopen, &od_default, sizeof(MQOD)) ;
  strcpy(odopen.ObjectName,q)                ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQOPEN ( qmh, &odopen, uopts, &qh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_open += 1.0000 * ( c2 - c1 )      ;
  time_open += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

MQLONG  rahmqntr::open ( MQOD * const uod
                       , MQLONG const uopts
                       )
{
  time_t c1, c2, t1, t2         ;

  memcpy(&odopen,uod,sizeof(odopen)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQOPEN ( qmh, &odopen, uopts, &qh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_open += 1.0000 * ( c2 - c1 )      ;
  time_open += 1.0000 * difftime(t2,t1)  ;

  memcpy(uod,&odopen,sizeof(odopen))     ;
  return (ac) ;
}

#ifdef _v7
//
// sub  will provide access to subscriptions
//
//       The MQSD is always supplied by the caller.
//
MQLONG  rahmqntr::sub ( MQSD * const subdesc )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQSUB ( qmh, subdesc, &qh, &sh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_sub += 1.0000 * ( c2 - c1 )      ;
  time_sub += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}
#endif

//
// close will stop access to a Queue
//
//       The CLOSE options are always supplied
//           by the caller.
//
MQLONG  rahmqntr::close ( MQLONG const uopts )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCLOSE ( qmh, &qh, uopts, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_close += 1.0000 * ( c2 - c1 )      ;
  time_close += 1.0000 * difftime(t2,t1)  ;

  qh = 0      ;

  return (ac) ;
}

//
// inq   will inquire upon the Attributes of
//       a Queue etc.
//
//       Note that ALL the Selectors stuff
//       has to be supplied by the caller
//
MQLONG  rahmqntr::inq ( MQLONG   const  uselnum
                      , MQLONG          usels
                      , MQLONG   const  uinum
                      , MQLONG *        uiattr
                      , MQLONG   const  ucharl
                      , char   *        ucattr
                      )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQINQ ( qmh, qh, uselnum, &usels ,
                   uinum  , uiattr ,
                   ucharl , ucattr ,
                        &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_inq += 1.0000 * ( c2 - c1 )      ;
  time_inq += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// set   will set the Attributes of a Queue etc.
//
//       Note that ALL the Selectors stuff
//       has to be supplied by the caller
//
MQLONG  rahmqntr::set ( MQLONG   const   uselnum
                      , MQLONG           usels
                      , MQLONG   const   uinum
                      , MQLONG *         uiattr
                      , MQLONG   const   ucharl
                      , char   *         ucattr
                      )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQSET ( qmh, qh, uselnum , &usels  ,
                   uinum   , uiattr  ,
                   ucharl  , ucattr  ,
                          &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_set += 1.0000 * ( c2 - c1 )      ;
  time_set += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// put       will write a Message into a Queue
//
//           The Message descriptor may either
//           be supplied or the instance version
//           used.
//
MQLONG  rahmqntr::put ( MQPMO  * const  uopts
                      , MQLONG   const  bufferl
                      , char   * const  buffer
                      )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, &pmd, uopts, bufferl, buffer, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqntr::put ( MQMD2  *        upmd
                      , MQPMO  * const  uopts
                      , MQLONG   const  bufferl
                      , char   * const  buffer
                      )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, upmd, uopts, bufferl, buffer, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


//
// put1       will open the Queue, write one message into it
//               and close the Queue
//
MQLONG  rahmqntr::put1 ( MQMD2  *        upmd
                       , MQOD   * const  uod
                       , MQPMO  * const  uopts
                       , MQLONG   const  bufferl
                       , char   * const  buffer
                       )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT1 ( qmh, uod, upmd, uopts, bufferl, buffer, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put1 += 1.0000 * ( c2 - c1 )      ;
  time_put1 += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
//    Browse    will non-destructively read a Queue
//
//              Use an instance Message dewscriptor
//
MQLONG  rahmqntr::browse ( MQLONG   const  bufferl
                         , char   * const  buffer
                         , MQLONG *        udatalen
                         )
{
  time_t c1, c2, t1, t2         ;

  memcpy(&bmd, &md_default, sizeof(MQMD2))                   ;
  memcpy(&bmo, &gmo_default, sizeof(MQGMO))                  ;
  bmo.Options        = MQGMO_NO_WAIT+MQGMO_BROWSE_NEXT+MQGMO_ACCEPT_TRUNCATED_MSG+
                       MQGMO_FAIL_IF_QUIESCING               ;
  bmo.WaitInterval   = MQWI_UNLIMITED                        ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, &bmd, &bmo, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
//    Get      will read a Message from a Queue
//
MQLONG  rahmqntr::get ( MQGMO  * const  uopts
                      , MQLONG   const  bufferl
                      , char   * const  buffer
                      , MQLONG *        udatalen
                      )
{
  time_t c1, c2, t1, t2         ;

  memcpy(&gmd, &md_default, sizeof(MQMD2))                   ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, &gmd, uopts, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqntr::get ( MQMD2  *        ugmd
                      , MQGMO  * const  uopts
                      , MQLONG   const  bufferl
                      , char   * const  buffer
                      , MQLONG *        udatalen
                      )
{
  time_t c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, ugmd, uopts, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}
//
// End of rahmqnTR Class definition and functions
//

//
// Global Variables
//

rahmqntr      gmqo??(MAXQS+2??)                            ; //Queue Objects

#if _MSC_VER
CMutex      * mutexbase = new CMutex(FALSE,NULL,NULL)      ; //Mutex based sync Object
CSingleLock * mutex     = new CSingleLock(mutexbase,FALSE) ; //Mutex Handle to lock gmqo access
#endif

//
// Polymorphic C++ functions
//

//
// Output Stream functions - convert MQ objects into cout
//
//        The MQOD, MQMD, MQPMO, MQGMO and MQSD structures are
//            enabled for use with the << output stream
//            operators via the usual operator overloadings.
//
//        Each field is dumped in character and hex.
//

//
// Print a MQOD  - Object descriptor
//

ostream& operator<<(ostream& os, MQOD * od)
 {

  MQULONG   i    ; //scanner

  os << "MQOD Structid            :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << od->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)od->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     Version             :"  ;
        os << od->Version ;
  os << " <"  ;
        os << hex << od->Version ;
  os << "> " ;
  os << endl ;

  os << "     ObjectType          :"  ;
        os << od->ObjectType ;
  os << " <"  ;
        os << hex << od->ObjectType ;
  os << "> " ;
  os << endl ;

  os << "     ObjectName          :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << od->ObjectName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)od->ObjectName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     ObjectQMgrName      :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << od->ObjectQMgrName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)od->ObjectQMgrName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     DynamicQName        :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << od->DynamicQName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)od->DynamicQName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     AlternateUserId     :"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << od->AlternateUserId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << hex << (unsigned short)od->AlternateUserId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     RecsPresent         :"  ;
        os << od->RecsPresent ;
  os << " <"  ;
        os << hex << od->RecsPresent ;
  os << "> " ;
  os << endl ;

  os << "     KnownDestCount      :"  ;
        os << od->KnownDestCount ;
  os << " <"  ;
        os << hex << od->KnownDestCount ;
  os << "> " ;
  os << endl ;

  os << "     UnknownDestCount    :"  ;
        os << od->UnknownDestCount ;
  os << " <"  ;
        os << hex << od->UnknownDestCount ;
  os << "> " ;
  os << endl ;

  os << "     InvalidDestCount    :"  ;
        os << od->InvalidDestCount ;
  os << " <"  ;
        os << hex << od->InvalidDestCount ;
  os << "> " ;
  os << endl ;

  os << "     ObjectRecOffset     :"  ;
        os << od->ObjectRecOffset ;
  os << " <"  ;
        os << hex << od->ObjectRecOffset ;
  os << "> " ;
  os << endl ;

  os << "     ResponseRecOffset   :"  ;
        os << od->ResponseRecOffset ;
  os << " <"  ;
        os << hex << od->ResponseRecOffset ;
  os << "> " ;
  os << endl ;

  os << "     ObjectRecPtr        :"  ;
        os << od->ObjectRecPtr ;
  os << " <"  ;
        os << hex << od->ObjectRecPtr ;
  os << "> " ;
  os << endl ;

  os << "     ResponseRecPtr      :"  ;
        os << od->ResponseRecPtr ;
  os << " <"  ;
        os << hex << od->ResponseRecPtr ;
  os << "> " ;
  os << endl ;

  os << "     AlternateSecurityId :"  ;
        for(i=0; i < sizeof(MQBYTE40) ;
            os << od->AlternateSecurityId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE40) ;
            os << hex << (unsigned short)od->AlternateSecurityId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     ResolvedQName       :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << od->ResolvedQName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)od->ResolvedQName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     ResolvedQMgrName    :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << od->ResolvedQMgrName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)od->ResolvedQMgrName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.Ptr    :"  ;
        os << od->ObjectString.VSPtr ;
  os << " <"  ;
        os << hex << od->ObjectString.VSPtr ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.Off    :"  ;
        os << od->ObjectString.VSOffset ;
  os << " <"  ;
        os << hex << od->ObjectString.VSOffset ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.Len    :"  ;
        os << od->ObjectString.VSLength ;
  os << " <"  ;
        os << hex << od->ObjectString.VSLength ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.Buf    :"  ;
        os << od->ObjectString.VSBufSize ;
  os << " <"  ;
        os << hex << od->ObjectString.VSBufSize ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.CCSI   :"  ;
        os << od->ObjectString.VSCCSID ;
  os << " <"  ;
        os << hex << od->ObjectString.VSCCSID ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.Ptr :"  ;
        os << od->SelectionString.VSPtr ;
  os << " <"  ;
        os << hex << od->SelectionString.VSPtr ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.Off :"  ;
        os << od->SelectionString.VSOffset ;
  os << " <"  ;
        os << hex << od->SelectionString.VSOffset ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.Len :"  ;
        os << od->SelectionString.VSLength ;
  os << " <"  ;
        os << hex << od->SelectionString.VSLength ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.Buf :"  ;
        os << od->SelectionString.VSBufSize ;
  os << " <"  ;
        os << hex << od->SelectionString.VSBufSize ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.CCSI:"  ;
        os << od->SelectionString.VSCCSID ;
  os << " <"  ;
        os << hex << od->SelectionString.VSCCSID ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.Ptr :"  ;
        os << od->ResObjectString.VSPtr ;
  os << " <"  ;
        os << hex << od->ResObjectString.VSPtr ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.Off :"  ;
        os << od->ResObjectString.VSOffset ;
  os << " <"  ;
        os << hex << od->ResObjectString.VSOffset ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.Len :"  ;
        os << od->ResObjectString.VSLength ;
  os << " <"  ;
        os << hex << od->ResObjectString.VSLength ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.Buf :"  ;
        os << od->ResObjectString.VSBufSize ;
  os << " <"  ;
        os << hex << od->ResObjectString.VSBufSize ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.CCSI:"  ;
        os << od->ResObjectString.VSCCSID ;
  os << " <"  ;
        os << hex << od->ResObjectString.VSCCSID ;
  os << "> " ;
  os << endl ;

  os << "     ResolvedType        :"  ;
        os << od->ResolvedType ;
  os << " <"  ;
        os << hex << od->ResolvedType ;
  os << "> " ;
  os << endl ;

  return (os) ;
 }

//
// Print a MQGMO - Get Message Options block
//

ostream& operator<<(ostream& os, MQGMO * gmo)
 {

  MQULONG   i    ; //scanner

  os << "MQGMO Structid        :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << gmo->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)gmo->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      Version         :"  ;
        os << gmo->Version ;
  os << " <"  ;
        os << hex << gmo->Version ;
  os << "> " ;
  os << endl ;

  os << "      Options         :"  ;
        os << gmo->Options   ;
  os << " <"  ;
        os << hex << gmo->Options ;
  os << "> " ;
  os << endl ;

  os << "      WaitInterval    :"  ;
        os << gmo->WaitInterval ;
  os << " <"  ;
        os << hex << gmo->WaitInterval ;
  os << "> " ;
  os << endl ;

  os << "      Signal1         :"  ;
        os << gmo->Signal1 ;
  os << " <"  ;
        os << hex << gmo->Signal1 ;
  os << "> " ;
  os << endl ;

  os << "      Signal2         :"  ;
        os << gmo->Signal2 ;
  os << " <"  ;
        os << hex << gmo->Signal2 ;
  os << "> " ;
  os << endl ;

  os << "      ResolvedQName   :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << gmo->ResolvedQName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)gmo->ResolvedQName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      MatchOptions    :"  ;
        os << gmo->MatchOptions ;
  os << " <"  ;
        os << hex << gmo->MatchOptions ;
  os << "> " ;
  os << endl ;

  os << "      GroupStatus     :"  ;
            os << gmo->GroupStatus;
  os << " <"  ;
            os << hex << (unsigned short)gmo->GroupStatus;
  os << "> " ;
  os << endl ;

  os << "      SegmentStatus   :"  ;
            os << gmo->SegmentStatus;
  os << " <"  ;
            os << hex << (unsigned short)gmo->SegmentStatus;
  os << "> " ;
  os << endl ;

  os << "      Segmentation    :"  ;
            os << gmo->Segmentation;
  os << " <"  ;
            os << hex << (unsigned short)gmo->Segmentation;
  os << "> " ;
  os << endl ;

  os << "      Reserved1       :"  ;
            os << gmo->Reserved1;
  os << " <"  ;
            os << hex << (unsigned short)gmo->Reserved1;
  os << "> " ;
  os << endl ;

  os << "      MsgToken        :"  ;
        for(i=0; i < sizeof(MQBYTE16) ;
            os << gmo->MsgToken??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE16) ;
            os << hex << (unsigned short)gmo->MsgToken??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ReturnedLength  :"  ;
        os << gmo->ReturnedLength ;
  os << " <"  ;
        os << hex << gmo->ReturnedLength ;
  os << "> " ;
  os << endl ;

  os << "      MsgHandle       :"  ;
        os << gmo->MsgHandle ;
  os << " <"  ;
        os << hex << gmo->MsgHandle ;
  os << "> " ;
  os << endl ;

  return (os) ;

 }

//
// Print a MQPMO - Put Message Options block
//

ostream& operator<<(ostream& os, MQPMO * pmo )
 {

  MQULONG   i    ; //scanner

  os << "MQPMO Structid          :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << pmo->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)pmo->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      Version           :"  ;
        os << pmo->Version ;
  os << " <"  ;
        os << hex << pmo->Version ;
  os << "> " ;
  os << endl ;

  os << "      Options           :"  ;
        os << pmo->Options ;
  os << " <"  ;
        os << hex << pmo->Options ;
  os << "> " ;
  os << endl ;

  os << "      Timeout           :"  ;
        os << pmo->Timeout ;
  os << " <"  ;
        os << hex << pmo->Timeout ;
  os << "> " ;
  os << endl ;

  os << "      Context           :"  ;
        os << pmo->Context ;
  os << " <"  ;
        os << hex << pmo->Context ;
  os << "> " ;
  os << endl ;

  os << "      KnownDestCount    :"  ;
        os << pmo->KnownDestCount ;
  os << " <"  ;
        os << hex << pmo->KnownDestCount ;
  os << "> " ;
  os << endl ;

  os << "      UnknownDestCount  :"  ;
        os << pmo->UnknownDestCount ;
  os << " <"  ;
        os << hex << pmo->UnknownDestCount ;
  os << "> " ;
  os << endl ;

  os << "      InvalidDestCount  :"  ;
        os << pmo->InvalidDestCount ;
  os << " <"  ;
        os << hex << pmo->InvalidDestCount ;
  os << "> " ;
  os << endl ;

  os << "      ResolvedQName     :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << pmo->ResolvedQName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)pmo->ResolvedQName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ResolvedQMgrName  :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << pmo->ResolvedQMgrName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)pmo->ResolvedQMgrName??(i++??));
  os << "> " ;
  os << endl ;

  os << "      RecsPresent       :"  ;
        os << pmo->RecsPresent ;
  os << " <"  ;
        os << hex << pmo->RecsPresent ;
  os << "> " ;
  os << endl ;

  os << "      PutMsgRecFields   :"  ;
        os << pmo->PutMsgRecFields ;
  os << " <"  ;
        os << hex << pmo->PutMsgRecFields ;
  os << "> " ;
  os << endl ;

  os << "      PutMsgRecOffset   :"  ;
        os << pmo->PutMsgRecOffset ;
  os << " <"  ;
        os << hex << pmo->PutMsgRecOffset ;
  os << "> " ;
  os << endl ;

  os << "      ResponseRecOffset :"  ;
        os << pmo->ResponseRecOffset ;
  os << " <"  ;
        os << hex << pmo->ResponseRecOffset ;
  os << "> " ;
  os << endl ;

  os << "      PutMsgRecPtr      :"  ;
        os << pmo->PutMsgRecPtr ;
  os << " <"  ;
        os << hex << pmo->PutMsgRecPtr ;
  os << "> " ;
  os << endl ;

  os << "      ResponseRecPtr    :"  ;
        os << pmo->ResponseRecPtr ;
  os << " <"  ;
        os << hex << pmo->ResponseRecPtr ;
  os << "> " ;
  os << endl ;

  os << "      OriginalMsgHandle :"  ;
        os << pmo->OriginalMsgHandle ;
  os << " <"  ;
        os << hex << pmo->OriginalMsgHandle ;
  os << "> " ;
  os << endl ;

  os << "      NewMsgHandle      :"  ;
        os << pmo->NewMsgHandle      ;
  os << " <"  ;
        os << hex << pmo->NewMsgHandle ;
  os << "> " ;
  os << endl ;

  os << "      Action            :"  ;
        os << pmo->Action            ;
  os << " <"  ;
        os << hex << pmo->Action     ;
  os << "> " ;
  os << endl ;

  os << "      PubLevel          :"  ;
        os << pmo->PubLevel          ;
  os << " <"  ;
        os << hex << pmo->PubLevel   ;
  os << "> " ;
  os << endl ;

  return (os) ;

 }

//
// Print a MQMD  - MQ Message Descriptor
//
ostream& operator<<(ostream& os, MQMD2 * md )
 {

  MQULONG   i    ; //scanner

  os << "MQMD  Structid        :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << md->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)md->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      Version         :"  ;
        os << md->Version ;
  os << " <"  ;
        os << hex << md->Version ;
  os << "> " ;
  os << endl ;

  os << "      Report          :"  ;
        os << md->Report ;
  os << " <"  ;
        os << hex << md->Report ;
  os << "> " ;
  os << endl ;

  os << "      MsgType         :"  ;
        os << md->MsgType ;
  os << " <"  ;
        os << hex << md->MsgType ;
  os << "> " ;
  os << endl ;

  os << "      Expiry          :"  ;
        os << md->Expiry  ;
  os << " <"  ;
        os << hex << md->Expiry ;
  os << "> " ;
  os << endl ;

  os << "      Feedback        :"  ;
        os << md->Feedback ;
  os << " <"  ;
        os << hex << md->Feedback ;
  os << "> " ;
  os << endl ;

  os << "      Encoding        :"  ;
        os << md->Encoding ;
  os << " <"  ;
        os << hex << md->Encoding ;
  os << "> " ;
  os << endl ;

  os << "      CodedCharSetId  :"  ;
        os << md->CodedCharSetId ;
  os << " <"  ;
        os << hex << md->CodedCharSetId ;
  os << "> " ;
  os << endl ;

  os << "      Format          :"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << md->Format??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << hex << (unsigned short)md->Format??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      Priority        :"  ;
        os << md->Priority ;
  os << " <"  ;
        os << hex << md->Priority ;
  os << "> " ;
  os << endl ;

  os << "      Persistence     :"  ;
        os << md->Persistence ;
  os << " <"  ;
        os << hex << md->Persistence ;
  os << "> " ;
  os << endl ;

  os << "      MsgId           :"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << md->MsgId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << hex << (unsigned short)md->MsgId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      CorrelId        :"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << md->CorrelId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << hex << (unsigned short)md->CorrelId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      BackoutCount    :"  ;
        os << md->BackoutCount ;
  os << " <"  ;
        os << hex << md->BackoutCount ;
  os << "> " ;
  os << endl ;

  os << "      ReplyToQ        :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << md->ReplyToQ??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)md->ReplyToQ??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ReplyToQMgr     :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << md->ReplyToQMgr??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)md->ReplyToQMgr??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      UserIdentifier  :"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << md->UserIdentifier??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << hex << (unsigned short)md->UserIdentifier??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      AccountingToken :"  ;
        for(i=0; i < sizeof(MQBYTE32) ;
            os << md->AccountingToken??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE32) ;
            os << hex << (unsigned short)md->AccountingToken??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ApplIdentityData:"  ;
        for(i=0; i < sizeof(MQCHAR32) ;
            os << md->ApplIdentityData??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR32) ;
            os << hex << (unsigned short)md->ApplIdentityData??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      PutApplType     :"  ;
        os << md->PutApplType ;
  os << " <"  ;
        os << hex << md->PutApplType ;
  os << "> " ;
  os << endl ;

  os << "      PutApplName     :"  ;
        for(i=0; i < sizeof(MQCHAR28) ;
            os << md->PutApplName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR28) ;
            os << hex << (unsigned short)md->PutApplName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      PutDate         :"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << md->PutDate??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << hex << (unsigned short)md->PutDate??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      PutTime         :"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << md->PutTime??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR8) ;
            os << hex << (unsigned short)md->PutTime??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      ApplOriginData  :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << md->ApplOriginData??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)md->ApplOriginData??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      GroupId         :"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << md->GroupId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << hex << (unsigned short)md->GroupId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "      MsgSeqNumber    :"  ;
        os << md->MsgSeqNumber ;
  os << " <"  ;
        os << hex << md->MsgSeqNumber ;
  os << "> " ;
  os << endl ;

  os << "      Offset          :"  ;
        os << md->Offset ;
  os << " <"  ;
        os << hex << md->Offset ;
  os << "> " ;
  os << endl ;

  os << "      MsgFlags        :"  ;
        os << md->MsgFlags ;
  os << " <"  ;
        os << hex << md->MsgFlags ;
  os << "> " ;
  os << endl ;

  os << "      OriginalLength  :"  ;
        os << md->OriginalLength ;
  os << " <"  ;
        os << hex << md->OriginalLength ;
  os << "> " ;
  os << endl ;

  return (os) ;
 }

//
// Print a MQSD  - Subscription descriptor
//

ostream& operator<<(ostream& os, MQSD * sd)
 {

  MQULONG   i    ; //scanner

  os << "MQSD Structid            :"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << sd->StrucId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR4) ;
            os << hex << (unsigned short)sd->StrucId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     Version             :"  ;
        os << sd->Version ;
  os << " <"  ;
        os << hex << sd->Version ;
  os << "> " ;
  os << endl ;

  os << "     Options             :"  ;
        os << sd->Options ;
  os << " <"  ;
        os << hex << sd->Options ;
  os << "> " ;
  os << endl ;

  os << "     ObjectName          :"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << sd->ObjectName??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR48) ;
            os << hex << (unsigned short)sd->ObjectName??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     AlternateUserId     :"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << sd->AlternateUserId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR12) ;
            os << hex << (unsigned short)sd->AlternateUserId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     AlternateSecurityId :"  ;
        for(i=0; i < sizeof(MQBYTE40) ;
            os << sd->AlternateSecurityId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE40) ;
            os << hex << (unsigned short)sd->AlternateSecurityId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     SubExpiry           :"  ;
        os << sd->SubExpiry ;
  os << " <"  ;
        os << hex << sd->SubExpiry ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.Ptr    :"  ;
        os << sd->ObjectString.VSPtr ;
  os << " <"  ;
        os << hex << sd->ObjectString.VSPtr ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.Off    :"  ;
        os << sd->ObjectString.VSOffset ;
  os << " <"  ;
        os << hex << sd->ObjectString.VSOffset ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.Len    :"  ;
        os << sd->ObjectString.VSLength ;
  os << " <"  ;
        os << hex << sd->ObjectString.VSLength ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.Buf    :"  ;
        os << sd->ObjectString.VSBufSize ;
  os << " <"  ;
        os << hex << sd->ObjectString.VSBufSize ;
  os << "> " ;
  os << endl ;

  os << "     ObjectString.CCSI   :"  ;
        os << sd->ObjectString.VSCCSID ;
  os << " <"  ;
        os << hex << sd->ObjectString.VSCCSID ;
  os << "> " ;
  os << endl ;

  os << "     SubName.Ptr         :"  ;
        os << sd->SubName.VSPtr ;
  os << " <"  ;
        os << hex << sd->SubName.VSPtr ;
  os << "> " ;
  os << endl ;

  os << "     SubName.Off         :"  ;
        os << sd->SubName.VSOffset ;
  os << " <"  ;
        os << hex << sd->SubName.VSOffset ;
  os << "> " ;
  os << endl ;

  os << "     SubName.Len         :"  ;
        os << sd->SubName.VSLength ;
  os << " <"  ;
        os << hex << sd->SubName.VSLength ;
  os << "> " ;
  os << endl ;

  os << "     SubName.Buf         :"  ;
        os << sd->SubName.VSBufSize ;
  os << " <"  ;
        os << hex << sd->SubName.VSBufSize ;
  os << "> " ;
  os << endl ;

  os << "     SubName.CCSI        :"  ;
        os << sd->SubName.VSCCSID ;
  os << " <"  ;
        os << hex << sd->SubName.VSCCSID ;
  os << "> " ;
  os << endl ;

  os << "     SubUserData.Ptr     :"  ;
        os << sd->SubUserData.VSPtr ;
  os << " <"  ;
        os << hex << sd->SubUserData.VSPtr ;
  os << "> " ;
  os << endl ;

  os << "     SubUserData.Off     :"  ;
        os << sd->SubUserData.VSOffset ;
  os << " <"  ;
        os << hex << sd->SubUserData.VSOffset ;
  os << "> " ;
  os << endl ;

  os << "     SubUserData.Len     :"  ;
        os << sd->SubUserData.VSLength ;
  os << " <"  ;
        os << hex << sd->SubUserData.VSLength ;
  os << "> " ;
  os << endl ;

  os << "     SubUserData.Buf     :"  ;
        os << sd->SubUserData.VSBufSize ;
  os << " <"  ;
        os << hex << sd->SubUserData.VSBufSize ;
  os << "> " ;
  os << endl ;

  os << "     SubUserData.CCSI    :"  ;
        os << sd->SubUserData.VSCCSID ;
  os << " <"  ;
        os << hex << sd->SubUserData.VSCCSID ;
  os << "> " ;
  os << endl ;

  os << "     SubCorrelId         :"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << sd->SubCorrelId??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE24) ;
            os << hex << (unsigned short)sd->SubCorrelId??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     PubPriority         :"  ;
        os << sd->PubPriority ;
  os << " <"  ;
        os << hex << sd->PubPriority ;
  os << "> " ;
  os << endl ;

  os << "     PubAccountingToken  :"  ;
        for(i=0; i < sizeof(MQBYTE32) ;
            os << sd->PubAccountingToken??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQBYTE32) ;
            os << hex << (unsigned short)sd->PubAccountingToken??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     PubApplIdentityData :"  ;
        for(i=0; i < sizeof(MQCHAR32) ;
            os << sd->PubApplIdentityData??(i++??)) ;
  os << " <"  ;
        for(i=0; i < sizeof(MQCHAR32) ;
            os << hex << (unsigned short)sd->PubApplIdentityData??(i++??)) ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.Ptr :"  ;
        os << sd->SelectionString.VSPtr ;
  os << " <"  ;
        os << hex << sd->SelectionString.VSPtr ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.Off :"  ;
        os << sd->SelectionString.VSOffset ;
  os << " <"  ;
        os << hex << sd->SelectionString.VSOffset ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.Len :"  ;
        os << sd->SelectionString.VSLength ;
  os << " <"  ;
        os << hex << sd->SelectionString.VSLength ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.Buf :"  ;
        os << sd->SelectionString.VSBufSize ;
  os << " <"  ;
        os << hex << sd->SelectionString.VSBufSize ;
  os << "> " ;
  os << endl ;

  os << "     SelectionString.CCSI:"  ;
        os << sd->SelectionString.VSCCSID ;
  os << " <"  ;
        os << hex << sd->SelectionString.VSCCSID ;
  os << "> " ;
  os << endl ;

  os << "     SubLevel            :"  ;
        os << sd->SubLevel ;
  os << " <"  ;
        os << hex << sd->SubLevel ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.Ptr :"  ;
        os << sd->ResObjectString.VSPtr ;
  os << " <"  ;
        os << hex << sd->ResObjectString.VSPtr ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.Off :"  ;
        os << sd->ResObjectString.VSOffset ;
  os << " <"  ;
        os << hex << sd->ResObjectString.VSOffset ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.Len :"  ;
        os << sd->ResObjectString.VSLength ;
  os << " <"  ;
        os << hex << sd->ResObjectString.VSLength ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.Buf :"  ;
        os << sd->ResObjectString.VSBufSize ;
  os << " <"  ;
        os << hex << sd->ResObjectString.VSBufSize ;
  os << "> " ;
  os << endl ;

  os << "     ResObjectString.CCSI:"  ;
        os << sd->ResObjectString.VSCCSID ;
  os << " <"  ;
        os << hex << sd->ResObjectString.VSCCSID ;
  os << "> " ;
  os << endl ;

  return (os) ;
 }


//
// Internal Functions
//

//
// Functions that Fetch and Set Rexx Variables in desired format
//
//
// Fetch integer value from REXX function parameter
//
void parm_to_ulong ( RXSTRING   parm    // parameter REXX string
                   , MQLONG *   number  // received value
                   )
{
 *number = 0;
 for(MQULONG i=0; i<parm.strlength; i++)
  {
   if((parm.strptr??(i??) < '0') || (parm.strptr??(i??) > '9')) return;
   else *number = (*number)*10 + ((parm.strptr??(i??)) & 0x0f);
  }
}
//
// Fetch MQLONG value from REXX stem variable
//
void stem_to_long ( MQULONG    traceid      // trace id of caller
                  , RXSTRING   stem         // stem variable name high
                  , char       name??(??)   // stem variable name low
                  , MQLONG *   number       // received value
                  )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(100??) ;  // Char version of variable value
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 memset(&varvalc,0,sizeof(varvalc))       ; // Clear REXX variable value buffer
 sv1.shvnext     = 0                      ; // Fetch only one variable
 sv1.shvcode     = RXSHV_SYFET            ; // Fetch operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 sv1.shvvalue.strptr    = varvalc                   ; // Set pointer to value buffer for REXX
 sv1.shvvalue.strlength = sizeof(varvalc)           ;
 sv1.shvvaluelen        = sv1.shvvalue.strlength    ; // Set max value length for REXX

 sv1rc           = RexxVariablePool(&sv1)           ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " " << varnamc
             << " ->" << varvalc
             << "<- " << sv1.shvvalue.strlength       // Returned length
             << "/"   << sv1.shvvaluelen )          ; // Requested length

 if (    (sv1rc                       == RXSHV_OK)
      && (strlen(varvalc)             != 0       ) )
          sscanf(varvalc,"%ld",number)              ; // long int value

 return ;
} // End of stem_to_long

//
// Fetch MQINT64 value from REXX stem variable
//
void stem_to_int64 ( MQULONG    traceid      // trace id of caller
                   , RXSTRING   stem         // stem variable name high
                   , char       name??(??)   // stem variable name low
                   , MQINT64 *  number       // received value
                   )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(100??) ;  // Char version of variable value
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 memset(&varvalc,0,sizeof(varvalc))                                ; // Clear REXX variable value buffer
 sv1.shvnext     = 0                                               ; // Fetch only one variable
 sv1.shvcode     = RXSHV_SYFET                                     ; // Fetch operation
 sv1.shvret      = 0                                               ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name)    ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))                 ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                           ; // REXX variable name length

 sv1.shvvalue.strptr    = varvalc                   ; // Set pointer to value buffer for REXX
 sv1.shvvalue.strlength = sizeof(varvalc)           ;
 sv1.shvvaluelen        = sv1.shvvalue.strlength    ; // Set max value length for REXX

 sv1rc           = RexxVariablePool(&sv1)           ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc
             << " ->" << varvalc
             << "<- " << sv1.shvvalue.strlength       // Returned length
             << "/"   << sv1.shvvaluelen )          ; // Requested length

 if (    (sv1rc               == RXSHV_OK)
      && (strlen(varvalc)     != 0       ) )
          sscanf(varvalc,"%lld",number)             ; // long int value

 return ;
} // End of stem_to_int64

//
// Fetch single MQCHAR value from REXX variable
//
void stem_to_char ( MQULONG    traceid      // trace id of caller
                  , RXSTRING   stem         // stem variable name high
                  , char       name??(??)   // stem variable name low
                  , MQCHAR *   letter       // received value
                  )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(100??) ;  // Char version of variable value
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 memset(&varvalc,0,sizeof(varvalc))       ; // Clear REXX variable value buffer
 sv1.shvnext     = 0                      ; // Fetch only one variable
 sv1.shvcode     = RXSHV_SYFET            ; // Fetch operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))        ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                  ; // REXX variable name length

 sv1.shvvalue.strptr    = varvalc                         ; // Set pointer to value buffer for REXX
 sv1.shvvalue.strlength = sizeof(varvalc)                 ;
 sv1.shvvaluelen        = sv1.shvvalue.strlength          ; // Set max value length for REXX

 sv1rc           = RexxVariablePool(&sv1)                 ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc
             << " ->" << varvalc
             << "<- " << sv1.shvvalue.strlength             // Returned length
             << "/"   << sv1.shvvaluelen )                ; // Requested length

 if (    (sv1rc                          == RXSHV_OK)
         && (strlen(varvalc)             != 0       ))
         *letter = varvalc??(0??)                         ; // char value

 return ;
} // End of stem_to_char


//
// Fetch string MQCHAR value from REXX variable
//
void stem_to_string ( MQULONG    traceid      // trace id of caller
                    , RXSTRING   stem         // stem variable name high
                    , char       name??(??)   // stem variable name low
                    , MQCHAR     string??(??) // received value
                    , int        size         // max size of value
                    )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(900??) ;  // Char version of variable value
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 memset(&varvalc,0,sizeof(varvalc))       ; // Clear REXX variable value buffer
 sv1.shvnext     = 0                      ; // Fetch only one variable
 sv1.shvcode     = RXSHV_SYFET            ; // Fetch operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 sv1.shvvalue.strptr    = varvalc                ; // Set pointer to value buffer for REXX
 sv1.shvvalue.strlength = size                   ;
 sv1.shvvaluelen        = sv1.shvvalue.strlength ; // Set max accepted value length

 sv1rc           = RexxVariablePool(&sv1)        ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc
             << " ->" << varvalc
             << "<- " << sv1.shvvalue.strlength    // Returned length
             << "/"   << sv1.shvvaluelen )       ; // Requested length

 if (sv1rc == RXSHV_OK)
    memcpy(string, &varvalc, size)               ; // Char string value

 return ;
} // End of stem_to_string

//
// Fetch MQBYTE value from REXX variable
// Use this only for MQBYTExx variables
//
void stem_to_bytes  ( MQULONG    traceid      // trace id of caller
                    , RXSTRING   stem         // stem variable name high
                    , char       name??(??)   // stem variable name low
                    , MQBYTE     string??(??) // received value
                    , int        size         // max size of value
                    )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(100??) ;  // Char version of variable value
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 memset(&varvalc,0,sizeof(varvalc))       ; // Clear REXX variable value buffer
 sv1.shvnext     = 0                      ; // Fetch only one variable
 sv1.shvcode     = RXSHV_SYFET            ; // Fetch operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 sv1.shvvalue.strptr    = varvalc                ; // Set pointer to value buffer for REXX
 sv1.shvvalue.strlength = size                   ;
 sv1.shvvaluelen        = sv1.shvvalue.strlength ; // Set max accepted value length

 sv1rc           = RexxVariablePool(&sv1)        ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc
             << " ->" << varvalc
             << "<- " << sv1.shvvalue.strlength    // Returned length
             << "/"   << sv1.shvvaluelen )       ; // Requested length

 if (sv1rc == RXSHV_OK)
    memcpy(string, &varvalc, size)               ; // Byte string value

 return ;
} // End of stem_to_bytes

//
// Fetch long MQBYTE value from REXX variable
// Buffer is provided by the caller
//
MQULONG stem_to_data ( MQULONG    traceid      // trace id of caller
                     , RXSTRING   stem         // stem variable name high
                     , char       name??(??)   // stem variable name low
                     , MQBYTE     string??(??) // received value
                     , int        size         // max size of value
                     )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 sv1.shvnext     = 0                      ; // Fetch only one variable
 sv1.shvcode     = RXSHV_SYFET            ; // Fetch operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 sv1.shvvalue.strptr    = (char *) string        ; // Set pointer to value buffer for REXX
 sv1.shvvalue.strlength = size                   ;
 sv1.shvvaluelen        = sv1.shvvalue.strlength ; // Set max accepted value length

 sv1rc           = RexxVariablePool(&sv1)        ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc
             << " ->" << string
             << "<- " << sv1.shvvalue.strlength    // Returned length
             << "/"   << sv1.shvvaluelen )       ; // Requested length

 if (sv1rc == RXSHV_OK)
   return (sv1.shvvalue.strlength)               ;
 else return 0                                   ;
} // End of stem_to_data

//
// Fetch MQCHARV value from REXX variable
//
void stem_to_strinv ( MQULONG    traceid     // trace id of caller
                    , RXSTRING   stem        // stem variable name high
                    , char       name??(??)  // stem variable name low
                    , MQCHARV *  string      // received value
                    )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 RXSTRING                varname          ;  // REXX variable name
 MQULONG                 dosrc = 0        ;  // RC from malloc

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(varname,varnamc,strlen(varnamc))                  ; // Construct REXX variable name structure

 stem_to_long  (traceid, varname, ".0" ,    &string->VSBufSize)    ; // Assume it is buffer length
 stem_to_long  (traceid, varname, ".CCSI" , &string->VSCCSID)      ; // CCSID may be required

 //
 // Allocate data buffer to store stem.name.1 variable data.
 // Buffer will be freed during call to stem_from_strinv
 //
 if ( string->VSBufSize != 0 )
   {
    DEBUG(PUT, "Doing malloc for " << string->VSBufSize << " bytes ")      ;
    string->VSPtr = malloc(string->VSBufSize)                              ;
    if ( string->VSPtr == NULL ) dosrc = errno                             ;
    DEBUG(traceid,"malloc RC " << dosrc << " PTR is "<<hex<< string->VSPtr);
    string->VSLength = stem_to_data(traceid, varname, ".1",
                                   (MQBYTE *)string->VSPtr, string->VSBufSize) ;
   }

 return ;
} // End of stem_to_strinv

//
// Set REXX variable to MQLONG value
//
void stem_from_long ( MQULONG    traceid      // trace id of caller
                    , char       zlist??(??)  // .ZLIST string for accumulation
                    , RXSTRING   stem         // stem variable name high
                    , char       name??(??)   // stem variable name low
                    , MQLONG     number       // value to set
                    )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(100??) ;  // Char version of variable value
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 sv1.shvnext     = 0                      ; // Set only one variable
 sv1.shvcode     = RXSHV_SYSET            ; // Set operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 sprintf(varvalc,"%ld",number)                       ; // Convert long int to char
 MAKERXSTRING(sv1.shvvalue,varvalc,strlen(varvalc))  ; // Construct REXX variable value structure
 sv1.shvvaluelen = sv1.shvvalue.strlength            ; // Set actual value length for REXX

 sv1rc           = RexxVariablePool(&sv1)            ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc << " ->" << varvalc
             << "<- " << strlen(varvalc)
             << "/"   << sv1.shvvalue.strlength )    ;

 if (   ( (sv1rc == RXSHV_OK) || (sv1rc == RXSHV_NEWV) )
     && (zlist != NULL))
   {
     strcat(zlist," ")        ;
     strcat(zlist,name)       ;
   }
 return ;
} // End of stem_from_long

//
// Set REXX variable to MQINT64 value
//
void stem_from_int64 ( MQULONG    traceid      // trace id of caller
                     , char       zlist??(??)  // .ZLIST string for accumulation
                     , RXSTRING   stem         // stem variable name high
                     , char       name??(??)   // stem variable name low
                     , MQINT64    number       // value to set
                     )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(100??) ;  // Char version of variable value
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 sv1.shvnext     = 0                      ; // Set only one variable
 sv1.shvcode     = RXSHV_SYSET            ; // Set operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 sprintf((char *)varvalc,"%lld",number)              ; // Convert MQINT64 to char
 MAKERXSTRING(sv1.shvvalue,varvalc,strlen(varvalc))  ; // Construct REXX variable value structure
 sv1.shvvaluelen = sv1.shvvalue.strlength            ; // Set actual value length for REXX

 sv1rc           = RexxVariablePool(&sv1)            ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc
             << " ->" << varvalc
             << "<- " << strlen(varvalc)
             << "/"   << sv1.shvvalue.strlength )    ;

 if (   ((sv1rc == RXSHV_OK) || (sv1rc == RXSHV_NEWV))
     && (zlist != NULL))
 {
     strcat(zlist," ")                      ;
     strcat(zlist,(const char *)name)       ;
 }
 return ;
} // End of stem_from_int64

//
// Set REXX variable to single MQCHAR value
//
void stem_from_char ( MQULONG    traceid      // trace id of caller
                    , char       zlist??(??)  // .ZLIST string for accumulation
                    , RXSTRING   stem         // stem variable name high
                    , char       name??(??)   // stem variable name low
                    , MQCHAR     letter       // value to set
                    )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 sv1.shvnext     = 0           ; // Set only one variable
 sv1.shvcode     = RXSHV_SYSET ; // Set operation
 sv1.shvret      = 0           ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 MAKERXSTRING(sv1.shvvalue,&letter,sizeof(MQCHAR))  ; // Construct REXX variable value structure
 sv1.shvvaluelen = sv1.shvvalue.strlength           ; // Set actual value length for REXX

 sv1rc           = RexxVariablePool(&sv1)           ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc
             << " ->" << letter
             << "<- " << 1
             << "/"   << sv1.shvvalue.strlength )   ;

 if (   ((sv1rc == RXSHV_OK) || (sv1rc == RXSHV_NEWV))
     && (zlist != NULL))
 {
     strcat(zlist," ")                      ;
     strcat(zlist,(const char *)name)       ;
 }

 return ;
} // End of stem_from_char

//
// Set REXX variable to MQCHAR string value
//
void stem_from_string ( MQULONG    traceid      // trace id of caller
                      , char       zlist??(??)  // .ZLIST string for accumulation
                      , RXSTRING   stem         // stem variable name high
                      , char       name??(??)   // stem variable name low
                      , MQCHAR     string??(??) // value to set
                      , int        size         // full size of value
                      )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(900??) ;  // Char version of variable value
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 sv1.shvnext     = 0                      ; // Set only one variable
 sv1.shvcode     = RXSHV_SYSET            ; // Set operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 sprintf(varvalc,"%.*s",size,string)                 ; // Make full size variable
 MAKERXSTRING(sv1.shvvalue,varvalc,strlen(varvalc))  ; // Construct REXX variable value structure
 sv1.shvvaluelen = sv1.shvvalue.strlength            ; // Set actual value length for REXX

 sv1rc           = RexxVariablePool(&sv1)            ; // Call REXX variable interface

 DEBUG(traceid, "rc " << sv1rc
             << " "   << varnamc
             << " ->" << varvalc
             << "<- " << strlen((const char *)varvalc)
             << "/"   << sv1.shvvalue.strlength )    ;

 if (   ((sv1rc == RXSHV_OK) || (sv1rc == RXSHV_NEWV))
     && (zlist != NULL))
 {
     strcat(zlist," ")                      ;
     strcat(zlist,(const char *)name)       ;
 }

 return ;
} // End of stem_from_string

//
// Set REXX variable to MQBYTE string value
//
void stem_from_bytes  ( MQULONG    traceid      // trace id of caller
                      , char       zlist??(??)  // .ZLIST string for accumulation
                      , RXSTRING   stem         // stem variable name high
                      , char       name??(??)   // stem variable name low
                      , MQBYTE     string??(??) // value to set
                      , int        size         // full size of value
                      )
{
 char                    varnamc??(100??) ;  // Char version of variable name
 SHVBLOCK                sv1              ;  // REXX var interface CB
 MQULONG                 sv1rc            ;  // REXX var interface RC

 sv1.shvnext     = 0                      ; // Set only one variable
 sv1.shvcode     = RXSHV_SYSET            ; // Set operation
 sv1.shvret      = 0                      ; // Zero out RC

 sprintf(varnamc,"%.*s%s",(int)stem.strlength,stem.strptr,name) ; // Construct REXX variable name
 MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))              ; // Construct REXX variable name structure
 sv1.shvnamelen  = sv1.shvname.strlength                        ; // REXX variable name length

 MAKERXSTRING(sv1.shvvalue,string,size)                         ; // Construct REXX variable value structure
 sv1.shvvaluelen = sv1.shvvalue.strlength                       ; // Set actual value length for REXX

 sv1rc           = RexxVariablePool(&sv1)                       ; // Call REXX variable interface

/* DEBUG(traceid, "rc " << sv1rc << " " << varnamc << " ->" << string
                      << "<- " << size
                      << "/"   << sv1.shvvalue.strlength ) ;
*/
 if (   ((sv1rc == RXSHV_OK) || (sv1rc == RXSHV_NEWV))
     && (zlist != NULL))
 {
     strcat(zlist," ")                      ;
     strcat(zlist,(const char *)name)       ;
 }

 return ;
} // End of stem_from_bytes

//
// Set REXX variable to MQCHARV value
//
void stem_from_strinv ( MQULONG    traceid      // trace id of caller
                      , char       zlist??(??)  // .ZLIST string for accumulation
                      , RXSTRING   stem         // stem variable name high
                      , char       name??(??)   // stem variable name low
                      , MQCHARV    string       // value to set
                      )
{
 char                   varnamc??(100??) ;      // Char version of variable name

 stem_from_long  (traceid, zlist, stem,
                  strcat(strcpy(varnamc,name),".0"),
                  string.VSLength)                          ;
 stem_from_long  (traceid, zlist, stem,
                  strcat(strcpy(varnamc,name),".CCSI"),
                  string.VSCCSID)                           ;
 //
 // Free data buffer for stem.name.1 variable data, if allocated.
 //
 if ( string.VSPtr != 0 )
    {
     stem_from_bytes (traceid, zlist, stem,
                      strcat(strcpy(varnamc,name),".1"),
                      (MQBYTE *)string.VSPtr, string.VSLength)  ;
     DEBUG(traceid, "Freemaining area " << hex << string.VSPtr) ;
     free(string.VSPtr)                                         ;
    }

 return ;
} // End of stem_from_strinv

//
// set_entry is called at the beginning of each external function execution
//
// Tracing is controlled by the DEBUG macro.
//
//         If the dynamic usage of DEBUG is set, then the
//            set_entry routine determines whether or not data
//            is written to cout (based on the RXMQTRACE Rexx var).
//

//
// set_entry will obtain the trace setting currently in effect,
//             and set tracebits variable for later checking in DEBUG macros
//
//         Known settings are :
//
//                              CONN  -> mqconn
//                              DISC  -> mqdisc
//                              OPEN  -> mqopen
//                              CLOSE -> mqclose
//                              GET   -> mqget
//                              PUT   -> mqput
//                              INQ   -> mqinq
//                              SET   -> mqset
//                              CMIT  -> mqcmit
//                              BACK  -> mqback
//
//                              BRO   -> Browse extension
//                              HXT   -> Header extraction extension
//                              EVENT -> Event determination extension
//                              TM    -> Trigger message extension
//                              COM   -> Command interface
//                              MQV   -> Debug a RXMQV
//
//                              MMD   -> Rexx stem var --> MQMD
//                              MOD   -> Rexx stem var --> MQOD
//                              MPO   -> Rexx stem var --> MQPMO
//                              MGO   -> Rexx stem var --> MQGMO
//
//                              BMD   -> MQMD  --> Rexx stem var
//                              BOD   -> MQOD  --> Rexx stem var
//                              BPO   -> MQPMO --> Rexx stem var
//                              BGO   -> MQGMO --> Rexx stem var
//
//                              SK    -> Return Code processing
//                              TR    -> Thread based processing
//
//                              INIT  -> initialization processing
//                              TERM  -> Deregistration processing
//
//                              *     -> Trace everything!!!
//
//
//  These settings are extracted from the RXMQTRACE Rexx Variable
//    (so RXMQTRACE = 'INQ OPEN' will trace Open and Inq operations)
//
//
//
void  set_entry ( const char * func        // Current function executed
                , MQULONG      traceid     // Trace bitmask
                )
{
 RXSTRING              varname          ;  // Variable name
 char                  varvalc??(200??) ;  // Char version of variable

 DEBUGS( "Entering set_entry"    ) ;
 DEBUGS( "Test is <" << test << ">" ) ;

 tracebits = ZERO;

 memset(&varvalc,0,sizeof(varvalc))                             ; // Clear REXX variable
 MAKERXSTRING(varname,TRACEVAR,strlen(TRACEVAR))                ; // Make name like RXMQNTRACE
 stem_to_string(ZERO, varname, "", varvalc, sizeof(varvalc))    ; // Get old-style trace variable
 strcat(varvalc," ")                                            ; // Add a blank to tail

 if ( strlen((const char *)varvalc) != 0)
    {
     if ( strstr((const char *)varvalc,"* "    ) != NULL) tracebits |= ALL   ;
     if ( strstr((const char *)varvalc,"CONN " ) != NULL) tracebits |= CONN  ;
     if ( strstr((const char *)varvalc,"DISC " ) != NULL) tracebits |= DISC  ;
     if ( strstr((const char *)varvalc,"OPEN " ) != NULL) tracebits |= OPEN  ;
     if ( strstr((const char *)varvalc,"CLOSE ") != NULL) tracebits |= CLOSE ;
     if ( strstr((const char *)varvalc,"GET "  ) != NULL) tracebits |= GET   ;
     if ( strstr((const char *)varvalc,"PUT "  ) != NULL) tracebits |= PUT   ;
     if ( strstr((const char *)varvalc,"PUT1 " ) != NULL) tracebits |= PUT1  ;
     if ( strstr((const char *)varvalc,"INQ "  ) != NULL) tracebits |= INQ   ;
     if ( strstr((const char *)varvalc,"SET "  ) != NULL) tracebits |= SET   ;
     if ( strstr((const char *)varvalc,"CMIT " ) != NULL) tracebits |= CMIT  ;
     if ( strstr((const char *)varvalc,"BACK " ) != NULL) tracebits |= BACK  ;
     if ( strstr((const char *)varvalc,"BRO "  ) != NULL) tracebits |= BRO   ;
     if ( strstr((const char *)varvalc,"HXT "  ) != NULL) tracebits |= HXT   ;
     if ( strstr((const char *)varvalc,"EVENT ") != NULL) tracebits |= EVENT ;
     if ( strstr((const char *)varvalc,"TM "   ) != NULL) tracebits |= TM    ;
     if ( strstr((const char *)varvalc,"COM "  ) != NULL) tracebits |= COM   ;
     if ( strstr((const char *)varvalc,"MQV "  ) != NULL) tracebits |= MQV   ;
     if ( strstr((const char *)varvalc,"MMD "  ) != NULL) tracebits |= MMD   ;
     if ( strstr((const char *)varvalc,"MOD "  ) != NULL) tracebits |= MOD   ;
     if ( strstr((const char *)varvalc,"MPO "  ) != NULL) tracebits |= MPO   ;
     if ( strstr((const char *)varvalc,"MGO "  ) != NULL) tracebits |= MGO   ;
     if ( strstr((const char *)varvalc,"BMD "  ) != NULL) tracebits |= BMD   ;
     if ( strstr((const char *)varvalc,"BOD "  ) != NULL) tracebits |= BOD   ;
     if ( strstr((const char *)varvalc,"BPO "  ) != NULL) tracebits |= BPO   ;
     if ( strstr((const char *)varvalc,"BGO "  ) != NULL) tracebits |= BGO   ;
     if ( strstr((const char *)varvalc,"SK "   ) != NULL) tracebits |= SK    ;
     if ( strstr((const char *)varvalc,"TR "   ) != NULL) tracebits |= TR    ;
     if ( strstr((const char *)varvalc,"INIT " ) != NULL) tracebits |= INIT  ;
     if ( strstr((const char *)varvalc,"TERM " ) != NULL) tracebits |= TERM  ;
    }

 memset(&varvalc,0,sizeof(varvalc))                             ; // Clear REXX variable
 MAKERXSTRING(varname,"RXMQTRACE",strlen("RXMQTRACE"))          ; // Make name like RXMQTRACE
 stem_to_string(ZERO, varname, "", varvalc, sizeof(varvalc))    ; // Get new-style trace variable
 strcat(varvalc," ")                                            ; // Add a blank to tail

 if ( strlen((const char *)varvalc) != 0)
    {
     if ( strstr((const char *)varvalc,"* "    ) != NULL) tracebits |= ALL   ;
     if ( strstr((const char *)varvalc,"CONN " ) != NULL) tracebits |= CONN  ;
     if ( strstr((const char *)varvalc,"DISC " ) != NULL) tracebits |= DISC  ;
     if ( strstr((const char *)varvalc,"OPEN " ) != NULL) tracebits |= OPEN  ;
     if ( strstr((const char *)varvalc,"CLOSE ") != NULL) tracebits |= CLOSE ;
     if ( strstr((const char *)varvalc,"GET "  ) != NULL) tracebits |= GET   ;
     if ( strstr((const char *)varvalc,"PUT "  ) != NULL) tracebits |= PUT   ;
     if ( strstr((const char *)varvalc,"PUT1 " ) != NULL) tracebits |= PUT1  ;
     if ( strstr((const char *)varvalc,"INQ "  ) != NULL) tracebits |= INQ   ;
     if ( strstr((const char *)varvalc,"SET "  ) != NULL) tracebits |= SET   ;
     if ( strstr((const char *)varvalc,"SUB "  ) != NULL) tracebits |= SUB   ;
     if ( strstr((const char *)varvalc,"CMIT " ) != NULL) tracebits |= CMIT  ;
     if ( strstr((const char *)varvalc,"BACK " ) != NULL) tracebits |= BACK  ;
     if ( strstr((const char *)varvalc,"BRO "  ) != NULL) tracebits |= BRO   ;
     if ( strstr((const char *)varvalc,"HXT "  ) != NULL) tracebits |= HXT   ;
     if ( strstr((const char *)varvalc,"EVENT ") != NULL) tracebits |= EVENT ;
     if ( strstr((const char *)varvalc,"TM "   ) != NULL) tracebits |= TM    ;
     if ( strstr((const char *)varvalc,"COM "  ) != NULL) tracebits |= COM   ;
     if ( strstr((const char *)varvalc,"MQV "  ) != NULL) tracebits |= MQV   ;
     if ( strstr((const char *)varvalc,"MMD "  ) != NULL) tracebits |= MMD   ;
     if ( strstr((const char *)varvalc,"MOD "  ) != NULL) tracebits |= MOD   ;
     if ( strstr((const char *)varvalc,"MPO "  ) != NULL) tracebits |= MPO   ;
     if ( strstr((const char *)varvalc,"MGO "  ) != NULL) tracebits |= MGO   ;
     if ( strstr((const char *)varvalc,"MSD "  ) != NULL) tracebits |= MSD   ;
     if ( strstr((const char *)varvalc,"BMD "  ) != NULL) tracebits |= BMD   ;
     if ( strstr((const char *)varvalc,"BOD "  ) != NULL) tracebits |= BOD   ;
     if ( strstr((const char *)varvalc,"BPO "  ) != NULL) tracebits |= BPO   ;
     if ( strstr((const char *)varvalc,"BGO "  ) != NULL) tracebits |= BGO   ;
     if ( strstr((const char *)varvalc,"BSD "  ) != NULL) tracebits |= BSD   ;
     if ( strstr((const char *)varvalc,"SK "   ) != NULL) tracebits |= SK    ;
     if ( strstr((const char *)varvalc,"TR "   ) != NULL) tracebits |= TR    ;
     if ( strstr((const char *)varvalc,"INIT " ) != NULL) tracebits |= INIT  ;
     if ( strstr((const char *)varvalc,"TERM " ) != NULL) tracebits |= TERM  ;
    }

// Perform typical entry actions

 DEBUG (traceid, "Entering " << func )                ; // Entering trace
 DEBUGS( "Leaving set_entry" ) ;

 return ;
} // End of set_entry function


//
// Return processing functions
//
//      set_return   : sets up the return variables
//
//          LASTRC    -> current operation Return Code
//          LASTCC    -> current operation MQ Completion Code
//          LASTAC    -> current operation MQ Reason     Code
//          LASTOP    -> current operation RXMQ Function
//          LASTMSG   -> current operation text message
//
//          The Return Code for the operation can be negative to
//              show that this interface has detected the error,
//              otherwise it will be the MQ Completion Code
//
//          The Message is in text format as follows:
//
//            Arg 1 : Return Code
//            Arg 2 : MQ Completion Code (or 0 if MQ not done)
//            Arg 2 : MQ Reason     Code (or 0 if MQ not done)
//            Arg 4 : RXMQ... function being run
//            Arg 5 : OK or an helpful error message
//            Arg 6 : Trace id of caller
//
//
void set_return ( const MQLONG   rc       //Function return Code
                , const MQLONG   cc       //MQ Completion Code
                , const MQLONG   ac       //MQ Reason Code
                , char         * op       //Function name
                , PRETMSG        pRetMsg  //Function message table
                , PRXSTRING      aretstr  //REXX Return String
                , MQULONG        traceid  //trace id of caller
                , char         * moremsg  //additional message
                )
{
 int                     i                ;
 RXSTRING                varname_new      ;  // Variable name
 RXSTRING                varname_old      ;  // Variable name

 DEBUG(SK, "Entering set_return" ) ;

 MAKERXSTRING(varname_new, "RXMQ.", sizeof("RXMQ.")-1)  ;
 MAKERXSTRING(varname_old, PREFIX,  sizeof(PREFIX)-1)   ;

 stem_from_long  (SK, NULL, varname_new, "LASTRC", rc);
 stem_from_long  (SK, NULL, varname_old, "LASTRC", rc);

 stem_from_long  (SK, NULL, varname_new, "LASTCC", cc);
 stem_from_long  (SK, NULL, varname_old, "LASTCC", cc);

 stem_from_long  (SK, NULL, varname_new, "LASTAC", ac);
 stem_from_long  (SK, NULL, varname_old, "LASTAC", ac);

 stem_from_string(SK, NULL, varname_new, "LASTOP", op, strlen(op)) ;
 stem_from_string(SK, NULL, varname_old, "LASTOP", op, strlen(op)) ;

 if (rc < 0)
   {
    for (i = 0; ; i++)
      {
       if (pRetMsg[i].retcode == rc ) break ;
       if (pRetMsg[i].retcode == -99) break ;
      }
    sprintf(aretstr->strptr,  "%ld %ld %ld %-s %s",
                                rc, cc, ac, op, pRetMsg[i].retmsgc);
   }
 else sprintf(aretstr->strptr,"%lu %lu %lu %-s %s %s",
                                rc, cc, ac, op,
           ( (cc == MQCC_OK      ) ? "OK"      :
             (cc == MQCC_WARNING ) ? "WARNING" :
             (cc == MQCC_FAILED  ) ? "FAILED"  :
                                     "UNKNOWN" ),
                                moremsg        ) ;

 aretstr->strlength  = strlen(aretstr->strptr)     ;
 stem_from_string(SK, NULL, varname_new, "LASTMSG",
                  aretstr->strptr, aretstr->strlength);
 stem_from_string(SK, NULL, varname_old, "LASTMSG",
                  aretstr->strptr, aretstr->strlength);

 DEBUG (traceid, "Leaving " << op )      ; // Leaving trace

 DEBUG(SK, "Leaving set_return" ) ;

 return ;
} // End of set_return function

//
// Functions that manipulate MQ objects and Rexx Variables
//
//           The make_??_from_stem  functions build the ??
//                                  MQ object by trying to reference
//                                  a stem variable (whose name ending
//                                  with a dot) whose name is supplied.
//                                  If the .xx part is present, then
//                                  the contents are copied into the
//                                  MQ Object.
//
//
//            The make_stem_from_?? functions work the other way around.
//                                  they are supplied with the name of
//                                  a stem variable (including the dot)
//                                  and build a Rexx representatation of
//                                  the MQ object by setting xx.yy
//                                  variables, where the yy is a
//                                  component of the MQ object. In addition,
//                                  (.)ZLIST is set to a list of the components
//                                  in the stem (without the leading dot).
//
//
//   The names of the .extensions are described in the functions.
//
//   Note that the make_stem_from_?? functions will ALWAYS build
//        all of the .extensions, whereas the make_??_from_stem
//        functions can cope with the omission of extensions.
//
//
// make_od_from_stem will return an Object Descriptor from the
//                   contents of a Stem Variable:
//
//                     .VER  -> Version
//                     .OT   -> ObjectType
//                     .ON   -> ObjectName
//                     .OQM  -> ObjectQMgrName
//                     .DQN  -> DynamicQueue
//                     .AUID -> AlternateUserid
//                     .RP   -> RecsPresent
//                     .KDC  -> KnownDestCount (output only)
//                     .UDC  -> UnknownDestCount (output only)
//                     .IDC  -> InvalidDestCount (output only)
//                     .ORO  -> ObjectRecOffset (not implemented)
//                     .RRO  -> ResponseRecOffset (not implemented)
//                     .ORP  -> ObjectRecPtr (not implemented)
//                     .RRP  -> ResponseRecPtr (not implemented)
//                     .ASID -> AlternateSecurityId (Windows only)
//                     .RQN  -> ResolvedQName (output only)
//                     .RQMN -> ResolvedQMgrName (output only)
//                     .OS   -> ObjectString
//                     .SS   -> SelectionString
//                     .ROS  -> ResObjectString
//                     .RT   -> ResolvedType
//

void make_od_from_stem ( MQOD     * odopen //Object desc to be built
                       , RXSTRING   stem   //name of stem variable
                       )
{
 DEBUG(MOD, "Entering make_od_from_stem " ) ;

 memcpy(odopen, &od_default, sizeof(MQOD));

//If the given variable is not a stem. variable, then take the
//   quicker option of assuming it's a Queue to open

 if ( stem.strptr??(stem.strlength-1??) != '.' )
   {
     memcpy(&odopen->ObjectName,
            stem.strptr,
            strlen((const char *)stem.strptr) ) ;
     DEBUG(MOD,  odopen ) ;
     return ;
    }

 //The given variable is a stem. variable, so get its contents

 // Version 1 of MQOD
 stem_to_long  (MOD, stem, "VER" , &odopen->Version)                               ;
 stem_to_long  (MOD, stem, "OT"  , &odopen->ObjectType)                            ;
 stem_to_string(MOD, stem, "ON"  ,  odopen->ObjectName,          sizeof(MQCHAR48)) ;
 stem_to_string(MOD, stem, "OQM" ,  odopen->ObjectQMgrName,      sizeof(MQCHAR48)) ;
 stem_to_string(MOD, stem, "DQN" ,  odopen->DynamicQName,        sizeof(MQCHAR48)) ;
 stem_to_string(MOD, stem, "AUID",  odopen->AlternateUserId,     sizeof(MQCHAR12)) ;
 // Version 2 of MQOD
 stem_to_long  (MOD, stem, "RP"  , &odopen->RecsPresent)                           ;
 stem_to_long  (MOD, stem, "KDC" , &odopen->KnownDestCount)                        ;
 stem_to_long  (MOD, stem, "UDC" , &odopen->UnknownDestCount)                      ;
 stem_to_long  (MOD, stem, "IDC" , &odopen->InvalidDestCount)                      ;
 // Version 3 of MQOD
 stem_to_bytes (MOD, stem, "ASID",  odopen->AlternateSecurityId, sizeof(MQBYTE40)) ;
 stem_to_string(MOD, stem, "RQN" ,  odopen->ResolvedQName,       sizeof(MQCHAR48)) ;
 stem_to_string(MOD, stem, "RQMN",  odopen->ResolvedQMgrName,    sizeof(MQCHAR48)) ;
 // Version 4 of MQOD
 stem_to_strinv(MOD, stem, "OS"  , &odopen->ObjectString)                          ;
 stem_to_strinv(MOD, stem, "SS"  , &odopen->SelectionString)                       ;
 stem_to_strinv(MOD, stem, "ROS" , &odopen->ResObjectString)                       ;
 stem_to_long  (MOD, stem, "RT"  , &odopen->ResolvedType)                          ;

 DEBUG(MOD, odopen ) ;

 DEBUG(MOD,  "Leaving make_od_from_stem " ) ;

 return ;
} // End of make_od_from_stem function

//
// make_stem_from_od will return an Object Descriptor as the
//                   contents of a Stem Variable:
//
//                     .VER  -> Version
//                     .OT   -> ObjectType
//                     .ON   -> ObjectName
//                     .OQM  -> ObjectQMgrName
//                     .DQN  -> DynamicQueue
//                     .AUID -> AlternateUserId
//                     .RP   -> RecsPresent
//                     .KDC  -> KnownDestCount
//                     .UDC  -> UnknownDestCount
//                     .IDC  -> InvalidDestCount
//                     .ORO  -> ObjectRecOffset (not implemented)
//                     .RRO  -> ResponseRecOffset (not implemented)
//                     .ORP  -> ObjectRecPtr (not implemented)
//                     .RRP  -> ResponseRecPtr (not implemented)
//                     .ASID -> AlternateSecurityId (Windows only)
//                     .RQN  -> ResolvedQName
//                     .RQMN -> ResolvedQMgrName
//                     .OS   -> ObjectString
//                     .SS   -> SelectionString
//                     .ROS  -> ResObjectString
//                     .RT   -> ResolvedType
//
//                     .ZLIST -> 'VER OT ON OQM DQN AUID
//                                RP
//                                KDC UDC IDC
//                                ASID
//                                RQN RQMN
//                                OS. SS. ROS. RT
//

void make_stem_from_od ( MQOD     * odopen //Object desc to be built
                       , RXSTRING   stem   //name of stem variable
                       )
{
 char                        zlist??(200??)   ;  // Char version of .ZLIST

 DEBUG(BOD, "Entering make_stem_from_od " ) ;

 DEBUG(BOD, odopen )                        ;

 zlist??(0??) = '\0'                        ;

 // Version 1 of MQOD
 stem_from_long  (BOD, zlist, stem, "VER" , odopen->Version)                               ;
 stem_from_long  (BOD, zlist, stem, "OT"  , odopen->ObjectType)                            ;
 stem_from_string(BOD, zlist, stem, "ON"  , odopen->ObjectName,          sizeof(MQCHAR48)) ;
 stem_from_string(BOD, zlist, stem, "OQM" , odopen->ObjectQMgrName,      sizeof(MQCHAR48)) ;
 stem_from_string(BOD, zlist, stem, "DQN" , odopen->DynamicQName,        sizeof(MQCHAR48)) ;
 stem_from_string(BOD, zlist, stem, "AUID", odopen->AlternateUserId,     sizeof(MQCHAR12)) ;
 // Version 2 of MQOD
 stem_from_long  (BOD, zlist, stem, "RP"  , odopen->RecsPresent)                           ;
 stem_from_long  (BOD, zlist, stem, "KDC" , odopen->KnownDestCount)                        ;
 stem_from_long  (BOD, zlist, stem, "UDC" , odopen->UnknownDestCount)                      ;
 stem_from_long  (BOD, zlist, stem, "IDC" , odopen->InvalidDestCount)                      ;
 // Version 3 of MQOD
 stem_from_bytes (BOD, zlist, stem, "ASID", odopen->AlternateSecurityId, sizeof(MQBYTE40)) ;
 stem_from_string(BOD, zlist, stem, "RQN" , odopen->ResolvedQName,       sizeof(MQCHAR48)) ;
 stem_from_string(BOD, zlist, stem, "RQMN", odopen->ResolvedQMgrName,    sizeof(MQCHAR48)) ;
 // Version 4 of MQOD
 stem_from_strinv(BOD, zlist, stem, "OS"  , odopen->ObjectString)                          ;
 stem_from_strinv(BOD, zlist, stem, "SS"  , odopen->SelectionString)                       ;
 stem_from_strinv(BOD, zlist, stem, "ROS" , odopen->ResObjectString)                       ;
 stem_from_long  (BOD, zlist, stem, "RT"  , odopen->ResolvedType)                          ;

 stem_from_string(BOD, zlist, stem, "ZLIST", zlist, strlen(zlist))                         ;

 DEBUG(BOD, "Leaving make_stem_from_od " ) ;

 return ;
} // End of make_stem_from_od function

//
// make_po_from_stem will return a Put Message Options Desc from the
//                   contents of a Stem Variable:
//
//                     .VER  -> Version
//                     .OPT  -> Options
//                     .TIME -> Timeout
//                     .CON  -> Context
//                     .KDC  -> KnownDestCount (output only)
//                     .UDC  -> UnKnownDestCount (output only)
//                     .IDC  -> InvalidDestCount (output only)
//                     .RQN  -> ResolvedQName (output only)
//                     .RQMN -> ResolvedQMgrName (output only)
//                     .RP   -> RecsPresent
//                     .PMRF -> PutMsgRecFields (not implemented)
//                     .PMRO -> PutMsgRecOffset (not implemented)
//                     .RRO  -> ResponseRecOffset (not implemented)
//                     .PMRP -> PutMsgRecPtr (not implemented)
//                     .RRP  -> ResponseRecPtr (not implemented)
//                     .OMH  -> OriginalMsgHandle
//                     .NMH  -> NewMsgHandle
//                     .ACT  -> Action
//                     .PL   -> PubLevel
//

void make_po_from_stem ( MQPMO    * pmo   //PMO to be built
                       , RXSTRING   stem  //name of stem variable
                       )
{
 DEBUG(MPO, "Entering make_po_from_stem " ) ;

 memcpy(pmo, &pmo_default, sizeof(MQPMO))     ;

 // Version 1 of MQPMO
 stem_to_long  (MPO, stem, "VER" , &pmo->Version)                            ;
 stem_to_long  (MPO, stem, "OPT" , &pmo->Options)                            ;
 stem_to_long  (MPO, stem, "TIME", &pmo->Timeout)                            ;
 stem_to_long  (MPO, stem, "CON" , &pmo->Context)                            ;
 stem_to_long  (MPO, stem, "KDC" , &pmo->KnownDestCount)                     ;
 stem_to_long  (MPO, stem, "UDC" , &pmo->UnknownDestCount)                   ;
 stem_to_long  (MPO, stem, "IDC" , &pmo->InvalidDestCount)                   ;
 stem_to_string(MPO, stem, "RQN" ,  pmo->ResolvedQName,    sizeof(MQCHAR48)) ;
 stem_to_string(MPO, stem, "RQMN",  pmo->ResolvedQMgrName, sizeof(MQCHAR48)) ;
 // Version 2 of MQPMO
 stem_to_long  (MPO, stem, "RP"  , &pmo->RecsPresent)                        ;
 // Version 3 of MQPMO
 stem_to_int64 (MPO, stem, "OMH" , &pmo->OriginalMsgHandle)                  ;
 stem_to_int64 (MPO, stem, "NMH" , &pmo->NewMsgHandle)                       ;
 stem_to_long  (MPO, stem, "ACT" , &pmo->Action)                             ;
 stem_to_long  (MPO, stem, "PL"  , &pmo->PubLevel)                           ;

 DEBUG(MPO, pmo ) ;

 DEBUG(MPO, "Leaving make_po_from_stem " ) ;

 return ;
} // End of make_po_from_stem function

//
// make_stem_from_po will return a Put Message Options Desc as the
//                   contents of a Stem Variable:
//
//                     .VER  -> Version
//                     .OPT  -> Options
//                     .TIME -> Timeout
//                     .CON  -> Context
//                     .KDC  -> KnownDestCount
//                     .UDC  -> UnKnownDestCount
//                     .IDC  -> InvalidDestCount
//                     .RQN  -> ResolvedQName
//                     .RQMN -> ResolvedQMgrName
//                     .RP   -> RecsPresent
//                     .PMRF -> PutMsgRecFields (not implemented)
//                     .PMRO -> PutMsgRecOffset (not implemented)
//                     .RRO  -> ResponseRecOffset (not implemented)
//                     .PMRP -> PutMsgRecPtr (not implemented)
//                     .RRP  -> ResponseRecPtr (not implemented)
//                     .OMH  -> OriginalMsgHandle
//                     .NMH  -> NewMsgHandle
//                     .ACT  -> Action
//                     .PL   -> PubLevel
//
//                     .ZLIST -> 'VER OPT CON CON TIME KDC UDC IDC RQN RQMN
//                                RP OMH NMH ACT PL'
//

void make_stem_from_po ( MQPMO    * pmo   //PMO to be built
                       , RXSTRING   stem  //name of stem variable
                       )
{
 char                    zlist??(200??)   ;  // Char version of .ZLIST

 DEBUG(BPO, "Entering make_stem_from_po " ) ;

 DEBUG(BPO,  pmo )                          ;

 zlist??(0??) = '\0'                        ;

 // Version 1 of MQPMO
 stem_from_long  (BPO, zlist, stem, "VER" , pmo->Version)                               ;
 stem_from_long  (BPO, zlist, stem, "OPT" , pmo->Options)                               ;
 stem_from_long  (BPO, zlist, stem, "TIME", pmo->Timeout)                               ;
 stem_from_long  (BPO, zlist, stem, "CON" , pmo->Context)                               ;
 stem_from_long  (BPO, zlist, stem, "KDC" , pmo->KnownDestCount)                        ;
 stem_from_long  (BPO, zlist, stem, "UDC" , pmo->UnknownDestCount)                      ;
 stem_from_long  (BPO, zlist, stem, "IDC" , pmo->InvalidDestCount)                      ;
 stem_from_string(BPO, zlist, stem, "RQN" , pmo->ResolvedQName,       sizeof(MQCHAR48)) ;
 stem_from_string(BPO, zlist, stem, "RQMN", pmo->ResolvedQMgrName,    sizeof(MQCHAR48)) ;
 // Version 2 of MQPMO
 stem_from_long  (BPO, zlist, stem, "RP"  , pmo->RecsPresent)                           ;
 // Version 3 of MQPMO
 stem_from_int64 (BPO, zlist, stem, "OMH" , pmo->OriginalMsgHandle)                     ;
 stem_from_int64 (BPO, zlist, stem, "NMH" , pmo->NewMsgHandle)                          ;
 stem_from_long  (BPO, zlist, stem, "ACT" , pmo->Action)                                ;
 stem_from_long  (BPO, zlist, stem, "PL"  , pmo->PubLevel)                              ;

 stem_from_string(BPO, zlist, stem, "ZLIST", zlist, strlen(zlist))                      ;

 DEBUG(BPO,  "Leaving make_stem_from_po " ) ;

 return ;
} // End of make_stem_from_po function

//
// make_go_from_stem will return a Get Message Options Desc from the
//                   contents of a Stem Variable:
//
//                     .VER   -> Version
//                     .OPT   -> Options
//                     .WAIT  -> WaitInterval
//                     .RQN   -> ResolvedQueueName (output only)
//                     .MOPT  -> MatchOptions
//                     .GS    -> GroupStatus (output only)
//                     .SS    -> SegmentStatus (output only)
//                     .SEG   -> Segmentation (output only)
//                     .MT    -> MsgToken
//                     .RL    -> ReturnedLength (output only)
//                     .MH    -> MsgHandle
//

void make_go_from_stem ( MQGMO    * gmo   //GMO to be built
                       , RXSTRING   stem  //name of stem variable
                       )
{
 DEBUG(MGO,  "Entering make_go_from_stem " ) ;

 memcpy(gmo, &gmo_default, sizeof(MQGMO))    ;

 // Version 1 of MQGMO
 stem_to_long  (MGO, stem, "VER" , &gmo->Version)                         ;
 stem_to_long  (MGO, stem, "OPT" , &gmo->Options)                         ;
 stem_to_long  (MGO, stem, "WAIT", &gmo->WaitInterval)                    ;
 stem_to_string(MGO, stem, "RQN" ,  gmo->ResolvedQName, sizeof(MQCHAR48)) ;
 // Version 2 of MQGMO
 stem_to_long  (MGO, stem, "MOPT", &gmo->MatchOptions)                    ;
 stem_to_char  (MGO, stem, "GS"  , &gmo->GroupStatus)                     ;
 stem_to_char  (MGO, stem, "SS"  , &gmo->SegmentStatus)                   ;
 stem_to_char  (MGO, stem, "SEG" , &gmo->Segmentation)                    ;
 // Version 3 of MQGMO
 stem_to_bytes (MGO, stem, "MT"  ,  gmo->MsgToken,      sizeof(MQBYTE16)) ;
 stem_to_long  (MGO, stem, "RL"  , &gmo->ReturnedLength)                  ;
 // Version 4 of MQGMO
 stem_to_int64 (MGO, stem, "MH"  , &gmo->MsgHandle)                       ;

 DEBUG(MGO,  gmo ) ;

 DEBUG(MGO,  "Leaving make_go_from_stem " ) ;

 return ;
} // End of make_go_from_stem function

//
// make_stem_from_go will return a Get Message Options Desc as the
//                   contents of a Stem Variable:
//
//                     .VER   -> Version
//                     .OPT   -> Options
//                     .WAIT  -> WaitInterval
//                     .RQN   -> ResolvedQueueName
//                     .MOPT  -> MatchOptions
//                     .GS    -> GroupStatus
//                     .SS    -> SegmentStatus
//                     .SEG   -> Segmentation
//                     .MT    -> sgToken
//                     .RL    -> ReturnedLength
//                     .MH    -> MsgHandle
//                     .ZLIST -> 'VER OPT WAIT RQN
//                                MOPT GS SS SEG
//                                MT RL MH'
//

void make_stem_from_go ( MQGMO    * gmo    //GMO to be built
                       , RXSTRING   stem   //name of stem variable
                       )
{
 char                    zlist??(200??)   ;  // Char version of .ZLIST

 DEBUG(BGO,  "Entering make_stem_from_go " ) ;

 DEBUG(BGO,  gmo )                           ;

 zlist??(0??) = '\0'                         ;

// Version 1 of MQGMO
 stem_from_long  (BGO, zlist, stem, "VER" , gmo->Version)                         ;
 stem_from_long  (BGO, zlist, stem, "OPT" , gmo->Options)                         ;
 stem_from_long  (BGO, zlist, stem, "WAIT", gmo->WaitInterval)                    ;
 stem_from_string(BGO, zlist, stem, "RQN" , gmo->ResolvedQName, sizeof(MQCHAR48)) ;
// Version 2 of MQGMO
 stem_from_long  (BGO, zlist, stem, "MOPT", gmo->MatchOptions)                    ;
 stem_from_char  (BGO, zlist, stem, "GS"  , gmo->GroupStatus)                     ;
 stem_from_char  (BGO, zlist, stem, "SS"  , gmo->SegmentStatus)                   ;
 stem_from_char  (BGO, zlist, stem, "SEG" , gmo->Segmentation)                    ;
// Version 3 of MQGMO
 stem_from_bytes (BGO, zlist, stem, "MT"  , gmo->MsgToken,      sizeof(MQBYTE16)) ;
 stem_from_long  (BGO, zlist, stem, "RL"  , gmo->ReturnedLength)                  ;
// Version 4 of MQGMO
 stem_from_int64 (BGO, zlist, stem, "MH"  , gmo->MsgHandle)                       ;

 stem_from_string(BGO, zlist, stem, "ZLIST", zlist, strlen(zlist))                ;

 DEBUG(BGO,  "Leaving make_stem_from_go " ) ;

 return ;
} // End of make_stem_from_go function

//
// make_md_from_stem will return a Put Message Options Desc from the
//                   contents of a Stem Variable:
//
//                     .VER   -> Version
//                     .REP   -> Report
//                     .MSG   -> MsgType
//                     .EXP   -> Expiry
//                     .FBK   -> Feedback
//                     .ENC   -> Encoding
//                     .CCSI  -> CodedCharSetId
//                     .FORM  -> Format
//                     .PRI   -> Priority
//                     .PER   -> Persistence
//                     .MSGID -> MsgId
//                     .CID   -> CorrelId
//                     .BC    -> BackoutCount
//                     .RTOQ  -> ReplyToQ
//                     .RTOQM -> ReplyToQMgr
//                     .UID   -> UserIdentifier
//                     .AT    -> AccountingToken
//                     .AID   -> ApplyIdentityData
//                     .PAT   -> PutApplType
//                     .PAN   -> PutApplName
//                     .PD    -> PutDate
//                     .PT    -> PutTime
//                     .AOD   -> ApplOriginData
//                     .GID   -> GroupId
//                     .MSN   -> MsgSeqNumber
//                     .OFF   -> Offset
//                     .MF    -> MsgFlags
//                     .OL    -> OriginalLength
//
//

void make_md_from_stem ( MQMD2    * md    //MQMD to be built
                       , RXSTRING   stem  //name of stem variable
                       )
{
 DEBUG(MMD,  "Entering make_md_from_stem " ) ;

 memcpy(md, &md_default, sizeof(MQMD2))      ;
 md->Version = MQMD_VERSION_1                ;

 // Version 1 of MQMD
 stem_to_long  (MMD, stem, "VER"  , &md->Version)                           ;
 stem_to_long  (MMD, stem, "REP"  , &md->Report)                            ;
 stem_to_long  (MMD, stem, "MSG"  , &md->MsgType)                           ;
 stem_to_long  (MMD, stem, "EXP"  , &md->Expiry)                            ;
 stem_to_long  (MMD, stem, "FBK"  , &md->Feedback)                          ;
 stem_to_long  (MMD, stem, "ENC"  , &md->Encoding)                          ;
 stem_to_long  (MMD, stem, "CCSI" , &md->CodedCharSetId)                    ;
 stem_to_string(MMD, stem, "FORM" ,  md->Format,           sizeof(MQCHAR8)) ;
 stem_to_long  (MMD, stem, "PRI"  , &md->Priority)                          ;
 stem_to_long  (MMD, stem, "PER"  , &md->Persistence)                       ;
 stem_to_bytes (MMD, stem, "MSGID",  md->MsgId,            sizeof(MQBYTE24));
 stem_to_bytes (MMD, stem, "CID"  ,  md->CorrelId,         sizeof(MQBYTE24));
 stem_to_long  (MMD, stem, "BC"   , &md->BackoutCount)                      ;
 stem_to_string(MMD, stem, "RTOQ" ,  md->ReplyToQ,         sizeof(MQCHAR48));
 stem_to_string(MMD, stem, "RTOQM",  md->ReplyToQMgr,      sizeof(MQCHAR48));
 stem_to_string(MMD, stem, "UID"  ,  md->UserIdentifier,   sizeof(MQCHAR12));
 stem_to_bytes (MMD, stem, "AT"   ,  md->AccountingToken,  sizeof(MQBYTE32));
 stem_to_string(MMD, stem, "AID"  ,  md->ApplIdentityData, sizeof(MQCHAR32));
 stem_to_long  (MMD, stem, "PAT"  , &md->PutApplType)                       ;
 stem_to_string(MMD, stem, "PAN"  ,  md->PutApplName,      sizeof(MQCHAR28));
 stem_to_string(MMD, stem, "PD"   ,  md->PutDate,          sizeof(MQCHAR8)) ;
 stem_to_string(MMD, stem, "PT"   ,  md->PutTime,          sizeof(MQCHAR8)) ;
 stem_to_string(MMD, stem, "AOD"  ,  md->ApplOriginData,   sizeof(MQCHAR4)) ;
 // Version 2 of MQMD
 stem_to_bytes (MMD, stem, "GID"  ,  md->GroupId,          sizeof(MQBYTE24));
 stem_to_long  (MMD, stem, "MSN"  , &md->MsgSeqNumber)                      ;
 stem_to_long  (MMD, stem, "OFF"  , &md->Offset)                            ;
 stem_to_long  (MMD, stem, "MF"   , &md->MsgFlags)                          ;
 stem_to_long  (MMD, stem, "OL"   , &md->OriginalLength)                    ;

 DEBUG(MMD,  md ) ;

 DEBUG(MMD,  "Leaving make_md_from_stem " ) ;

 return ;
} // End of make_md_from_stem function

//
// make_stem_from_md will return a Message Descriptor as the
//                   contents of a Stem Variable:
//
//                     .VER   -> Version
//                     .REP   -> Report
//                     .MSG   -> MsgType
//                     .EXP   -> Expiry
//                     .FBK   -> Feedback
//                     .ENC   -> Encoding
//                     .CCSI  -> CodedCharSetId
//                     .FORM  -> Format
//                     .PRI   -> Priority
//                     .PER   -> Persistence
//                     .MSGID -> MsgId
//                     .CID   -> CorrelId
//                     .BC    -> BackoutCount
//                     .RTOQ  -> ReplyToQ
//                     .RTOQM -> ReplyToQMgr
//                     .UID   -> UserIdentifier
//                     .AT    -> AccountingToken
//                     .AID   -> ApplyIdentityData
//                     .PAT   -> PutApplType
//                     .PAN   -> PutApplName
//                     .PD    -> PutDate
//                     .PT    -> PutTime
//                     .AOD   -> ApplOriginData
//                     .GID   -> GroupId
//                     .MSN   -> MsgSeqNumber
//                     .OFF   -> Offset
//                     .MF    -> MsgFlags
//                     .OL    -> OriginalLength
//
//
//                     .ZLIST -> 'VER REP MSG EXP FBK ENC CCSI FORM PRI PER MSGID CID BC
//                                RTOQ RTOQM UID AT AID PAT PAN PD PT AOD
//                                GID MSN OFF MF OL
//
//

void make_stem_from_md ( MQMD2    * md    //MQMD to be built
                       , RXSTRING   stem  //name of stem variable
                       )
{
 char                         zlist??(200??)   ;  // Char version of .ZLIST

 DEBUG(BMD,  "Entering make_stem_from_md " ) ;

 DEBUG(BMD,  md ) ;

 zlist??(0??) = '\0'                         ;

// Version 1 of MQMD
 stem_from_long  (BMD, zlist, stem, "VER"  , md->Version)                               ;
 stem_from_long  (BMD, zlist, stem, "REP"  , md->Report)                                ;
 stem_from_long  (BMD, zlist, stem, "MSG"  , md->MsgType)                               ;
 stem_from_long  (BMD, zlist, stem, "EXP"  , md->Expiry)                                ;
 stem_from_long  (BMD, zlist, stem, "FBK"  , md->Feedback)                              ;
 stem_from_long  (BMD, zlist, stem, "ENC"  , md->Encoding)                              ;
 stem_from_long  (BMD, zlist, stem, "CCSI" , md->CodedCharSetId)                        ;
 stem_from_string(BMD, zlist, stem, "FORM" , md->Format,              sizeof(MQCHAR8))  ;
 stem_from_long  (BMD, zlist, stem, "PRI"  , md->Priority)                              ;
 stem_from_long  (BMD, zlist, stem, "PER"  , md->Persistence)                           ;
 stem_from_bytes (BMD, zlist, stem, "MSGID", md->MsgId,               sizeof(MQBYTE24)) ;
 stem_from_bytes (BMD, zlist, stem, "CID"  , md->CorrelId,            sizeof(MQBYTE24)) ;
 stem_from_long  (BMD, zlist, stem, "BC"   , md->BackoutCount)                          ;
 stem_from_string(BMD, zlist, stem, "RTOQ" , md->ReplyToQ,            sizeof(MQCHAR48)) ;
 stem_from_string(BMD, zlist, stem, "RTOQM", md->ReplyToQMgr,         sizeof(MQCHAR48)) ;
 stem_from_string(BMD, zlist, stem, "UID"  , md->UserIdentifier,      sizeof(MQCHAR12)) ;
 stem_from_bytes (BMD, zlist, stem, "AT"   , md->AccountingToken,     sizeof(MQBYTE32)) ;
 stem_from_string(BMD, zlist, stem, "AID"  , md->ApplIdentityData,    sizeof(MQCHAR32)) ;
 stem_from_long  (BMD, zlist, stem, "PAT"  , md->PutApplType)                           ;
 stem_from_string(BMD, zlist, stem, "PAN"  , md->PutApplName,         sizeof(MQCHAR28)) ;
 stem_from_string(BMD, zlist, stem, "PD"   , md->PutDate,             sizeof(MQCHAR8))  ;
 stem_from_string(BMD, zlist, stem, "PT"   , md->PutTime,             sizeof(MQCHAR8))  ;
 stem_from_string(BMD, zlist, stem, "AOD"  , md->ApplOriginData,      sizeof(MQCHAR4))  ;
// Version 2 of MQMD
 stem_from_bytes (BMD, zlist, stem, "GID"  , md->GroupId,             sizeof(MQBYTE24)) ;
 stem_from_long  (BMD, zlist, stem, "MSN"  , md->MsgSeqNumber)                          ;
 stem_from_long  (BMD, zlist, stem, "OFF"  , md->Offset)                                ;
 stem_from_long  (BMD, zlist, stem, "MF"   , md->MsgFlags)                              ;
 stem_from_long  (BMD, zlist, stem, "OL"   , md->OriginalLength)                        ;

 stem_from_string(BMD, zlist, stem, "ZLIST", zlist, strlen(zlist))                      ;

 DEBUG(BMD,  "Leaving make_stem_from_md " ) ;

 return ;
} // End of make_stem_from_md function

//
// make_sd_from_stem will return a Subscription Descriptor from the
//                   contents of a Stem Variable:
//
//                     .VER  -> Version
//                     .OPT  -> Options
//                     .ON   -> ObjectName
//                     .AUID -> AlternateUserId
//                     .ASID -> AlternateSecurityId
//                     .SE   -> SubExpiry
//                     .OS   -> ObjectString
//                     .SN   -> SubName
//                     .SUD  -> SubUserData
//                     .SCID -> SubCorrelId
//                     .PP   -> PubPriority
//                     .PAT  -> PubAccountingToken
//                     .PAID -> PubAppIdentityData
//                     .SS   -> SelectionString
//                     .SL   -> SubLevel
//                     .ROS  -> ResObjectString
//

void make_sd_from_stem ( MQSD     * subdesc  //Sub desc to be built
                       , RXSTRING   stem     //name of stem variable
                       )
{
 DEBUG(MSD, "Entering make_sd_from_stem " ) ;

 memcpy(subdesc, &sd_default, sizeof(MQSD));

 //The given variable is a stem. variable, so get its contents

 // Version 1 of MQSD
 stem_to_long  (MSD, stem, "VER" , &subdesc->Version)                               ;
 stem_to_long  (MSD, stem, "OPT" , &subdesc->Options)                               ;
 stem_to_string(MSD, stem, "ON"  ,  subdesc->ObjectName,          sizeof(MQCHAR48)) ;
 stem_to_string(MSD, stem, "AUID",  subdesc->AlternateUserId,     sizeof(MQCHAR12)) ;
 stem_to_bytes (MSD, stem, "ASID",  subdesc->AlternateSecurityId, sizeof(MQBYTE40)) ;
 stem_to_long  (MSD, stem, "SE"  , &subdesc->SubExpiry)                             ;
 stem_to_strinv(MSD, stem, "OS"  , &subdesc->ObjectString)                          ;
 stem_to_strinv(MSD, stem, "SN"  , &subdesc->SubName)                               ;
 stem_to_strinv(MSD, stem, "SUD" , &subdesc->SubUserData)                           ;
 stem_to_bytes (MSD, stem, "SCID",  subdesc->SubCorrelId,         sizeof(MQBYTE24)) ;
 stem_to_long  (MSD, stem, "PP"  , &subdesc->PubPriority)                           ;
 stem_to_bytes (MSD, stem, "PAT" ,  subdesc->PubAccountingToken,  sizeof(MQBYTE32)) ;
 stem_to_string(MSD, stem, "PAID",  subdesc->PubApplIdentityData, sizeof(MQCHAR32)) ;
 stem_to_strinv(MSD, stem, "SS"  , &subdesc->SelectionString)                       ;
 stem_to_long  (MSD, stem, "SL"  , &subdesc->SubLevel)                              ;
 stem_to_strinv(MSD, stem, "ROS" , &subdesc->ResObjectString)                       ;

 DEBUG(MSD, subdesc) ;

 DEBUG(MSD,  "Leaving make_sd_from_stem " ) ;

 return ;
} // End of make_sd_from_stem function

//
// make_stem_from_sd will return a Subscription Descriptor as the
//                   contents of a Stem Variable:
//
//                     .VER  -> Version
//                     .OPT  -> Options
//                     .ON   -> ObjectName
//                     .AUID -> AlternateUserId
//                     .ASID -> AlternateSecurityId
//                     .SE   -> SubExpiry
//                     .OS   -> ObjectString
//                     .SN   -> SubName
//                     .SUD  -> SubUserData
//                     .SCID -> SubCorrelId
//                     .PP   -> PubPriority
//                     .PAT  -> PubAccountingToken
//                     .PAID -> PubAppIdentityData
//                     .SS   -> SelectionString
//                     .SL   -> SubLevel
//                     .ROS  -> ResObjectString
//
//                     .ZLIST -> 'VER OPT ON AUID ASID
//                                SE OS. SN. SUD. SCID
//                                PP PAT PAID
//                                SS. SL ROS.
//

void make_stem_from_sd ( MQSD     * subdesc  //Sub desc to be built
                       , RXSTRING   stem     //name of stem variable
                       )
{
 char                        zlist??(200??)   ;  // Char version of .ZLIST

 DEBUG(BSD, "Entering make_stem_from_sd " ) ;

 DEBUG(BSD, subdesc)                        ;

 zlist??(0??) = '\0'                        ;

 // Version 1 of MQSD
 stem_from_long  (BSD, zlist, stem, "VER" , subdesc->Version)                               ;
 stem_from_long  (BSD, zlist, stem, "OPT" , subdesc->Options)                               ;
 stem_from_string(BSD, zlist, stem, "ON"  , subdesc->ObjectName,          sizeof(MQCHAR48)) ;
 stem_from_string(BSD, zlist, stem, "AUID", subdesc->AlternateUserId,     sizeof(MQCHAR12)) ;
 stem_from_bytes (BSD, zlist, stem, "ASID", subdesc->AlternateSecurityId, sizeof(MQBYTE40)) ;
 stem_from_long  (BSD, zlist, stem, "SE"  , subdesc->SubExpiry)                             ;
 stem_from_strinv(BSD, zlist, stem, "OS"  , subdesc->ObjectString)                          ;
 stem_from_strinv(BSD, zlist, stem, "SN"  , subdesc->SubName)                               ;
 stem_from_strinv(BSD, zlist, stem, "SUD" , subdesc->SubUserData)                           ;
 stem_from_bytes (BSD, zlist, stem, "SCID", subdesc->SubCorrelId,         sizeof(MQBYTE24)) ;
 stem_from_long  (BSD, zlist, stem, "PP"  , subdesc->PubPriority)                           ;
 stem_from_bytes (BSD, zlist, stem, "PAT" , subdesc->PubAccountingToken,  sizeof(MQBYTE32)) ;
 stem_from_string(BSD, zlist, stem, "PAID", subdesc->PubApplIdentityData, sizeof(MQCHAR32)) ;
 stem_from_strinv(BSD, zlist, stem, "SS"  , subdesc->SelectionString)                       ;
 stem_from_long  (BSD, zlist, stem, "SL"  , subdesc->SubLevel)                              ;
 stem_from_strinv(BSD, zlist, stem, "ROS" , subdesc->ResObjectString)                       ;

 stem_from_string(BSD, zlist, stem, "ZLIST", zlist, strlen(zlist))                         ;

 DEBUG(BOD, "Leaving make_stem_from_sd " ) ;

 return ;
} // End of make_stem_from_sd function


//
// Genuine Internal functions, not concerned with specific things
//

//
// setcons - sets up the MQ constants in the Rexx Variable Space
//           Traced by INIT setting
//

void setcons ()
 {

//
// Structure to define MQ literals, to be setup in Variable Space
//

 struct s_define_mq_ints
        {
          char           s_define_mq_ints_name??(255??) ; // Constant name
          MQLONG         s_define_mq_ints_value         ; // Constant value
          MQULONG        s_define_mq_ints_type          ;  // 0 -> number
                                                           // 1 -> MQRC_
                                                           // 2 -> not used
                                                           // 3 -> MQCC_
                                                           // 4 -> MQCA_
        } ;

 typedef struct s_define_mq_ints   t_define_mq_ints ;
 typedef        t_define_mq_ints * p_define_mq_ints ;

 struct s_define_mq_char
        {
          char           s_define_mq_char_name??(255??) ;
          char           s_define_mq_char_value??(255??);
          MQULONG        s_define_mq_char_type          ;  // 0 -> char
        } ;

 typedef struct s_define_mq_char   t_define_mq_char ;
 typedef        t_define_mq_char * p_define_mq_char ;

 struct s_define_mq_byte
        {
          char           s_define_mq_byte_name??(255??) ;
          MQBYTE         s_define_mq_byte_value??(255??);
          MQULONG        s_define_mq_byte_size          ;
        } ;

 typedef struct s_define_mq_byte   t_define_mq_byte ;
 typedef        t_define_mq_byte * p_define_mq_byte ;

 //
 //  Usual local variables
 //

 MQULONG                  i                ;  // Looper
 p_define_mq_ints         pi               ;  // Table pointer
 p_define_mq_char         pc               ;  // Table pointer
 p_define_mq_byte         pb               ;  // Table pointer
 RXSTRING                 varname          ;  // Variable name
 RXSTRING                 varname_new      ;  // Variable name
 RXSTRING                 varname_old      ;  // Variable name
 char                     varvalc??(100??) ;  // Char version of variable

//
// Array and Structure to define MQ literals, to be setup in Variable Space
// Constants are placed in separate header file for ease of use
//

#include <rxmqcons.hpp>

 DEBUG (INIT, "Entering setcons" ) ;

 //
 // initialize all the constants required.
 //
 // This is done by looping through the table of names and
 //      settings, invoking the RexxVariable interface to set up
 //      each literal.
 //
 // Additionally, if the constant represents Completion Code (MQCC_)
 //               then create an number->name entry in the
 //               RXMQ.CCMAP. stem variable
 //
 // Additionally, if the constant represents Reason Code (MQRC_)
 //               then create an number->name entry in the
 //               RXMQ.RCMAP. stem variable
 //
 // Additionally, if the constant represents Event (MQCA_,...)
 //               then create an number->name entry in the
 //               RXMQ.CAMAP. stem variable
 //

 MAKERXSTRING(varname_new, "RXMQ.", sizeof("RXMQ.")-1)  ;
 MAKERXSTRING(varname_old, PREFIX,  sizeof(PREFIX)-1)   ;

 for (i=0; ; i++)
   {
    pi = &define_mq_ints??(i??)                           ;
    if ( pi->s_define_mq_ints_name??(0??) == '?' )  break ;

    MAKERXSTRING(varname,
                 &(pi->s_define_mq_ints_name??(0??)),
                 strlen(pi->s_define_mq_ints_name))         ; // REXX variable name
    stem_from_long  (INIT, NULL, varname, "", pi->s_define_mq_ints_value);

    if ( pi->s_define_mq_ints_type == 1 )    //Only for MQRC_ stuff
     {
      sprintf(varvalc, "RCMAP.%ld", pi->s_define_mq_ints_value)   ; // like RXMQV.RCMAP.0
      stem_from_string(INIT, NULL, varname_new, varvalc,
                       &(pi->s_define_mq_ints_name??(0??)),
                       strlen(pi->s_define_mq_ints_name))   ;
      stem_from_string(INIT, NULL, varname_old, varvalc,
                       &(pi->s_define_mq_ints_name??(0??)),
                       strlen(pi->s_define_mq_ints_name))   ;
      }

    if ( pi->s_define_mq_ints_type == 3 )    //Only for MQCC_ stuff
     {
      sprintf(varvalc, "CCMAP.%ld", pi->s_define_mq_ints_value)   ; // like RXMQV.RCMAP.0
      stem_from_string(INIT, NULL, varname_new, varvalc,
                       &(pi->s_define_mq_ints_name??(0??)),
                       strlen(pi->s_define_mq_ints_name))   ;
     }

    if ( pi->s_define_mq_ints_type == 4 )    //Only for Selector/Event Attributes
     {
      sprintf(varvalc, "CAMAP.%ld", pi->s_define_mq_ints_value)   ; // like RXMQV.RCMAP.0

      varname.strptr    = strstr(&(pi->s_define_mq_ints_name??(0??)), "_") +1;
      varname.strlength = strlen(varname.strptr)            ;

      stem_from_string(INIT, NULL, varname_new, varvalc,
                       varname.strptr, varname.strlength)                 ;
     }
   } // End of Integer initializations Loop

 for (i=0; ; i++)
   {
    pc = &define_mq_char??(i??)                           ;
    if ( pc->s_define_mq_char_name??(0??) == '?' ) break  ;

    MAKERXSTRING(varname,
                 &(pc->s_define_mq_char_name??(0??)),
                 strlen(pc->s_define_mq_char_name))         ; // REXX variable name
    stem_from_string(INIT, NULL, varname, "",
                     &(pc->s_define_mq_char_value??(0??)),
                     strlen(pc->s_define_mq_char_value))    ;
   } // End of Character initializations Loop

 for (i=0; ; i++)
   {
    pb = &define_mq_byte??(i??)                           ;
    if ( pb->s_define_mq_byte_name??(0??) == '?' ) break  ;

    MAKERXSTRING(varname,
                 &(pb->s_define_mq_byte_name??(0??)),
                 strlen(pb->s_define_mq_byte_name))       ; // REXX variable name
    stem_from_bytes(INIT, NULL, varname, "",
                    &(pb->s_define_mq_byte_value??(0??)),
                    pb->s_define_mq_byte_size)            ;
   } // End of Byte initializations Loop

 DEBUG(INIT, "Leaving setcons " ) ;

 return ;

 } // End of setcons function

//
// Routine geteventname - converts the NUMBER of an Event into
//                        a name (for use as a Component Variable),
//                        which is defined event name like MQCA_...
//                        without MQCA_ prefix itself
//
//         If the PCF field number is unknown, then the
//            component is set to the character version of the
//            PCF event number. This only works for EVENT PCFs!!!!
//
//         Tracing is via the EVENT setting
//
//

void geteventname(char * output, const MQLONG pcfnum )
 {
  char                    varnamc??(100??) ;  // Char version of var name
  SHVBLOCK                sv1              ;  // Var interface CB
  MQULONG                 sv1rc            ;  // Var interface RC

  DEBUG(EVENT,"Entering geteventname for " << pcfnum << " " ) ;

  sv1.shvnext     = 0           ; // Fetch only one variable
  sv1.shvcode     = RXSHV_SYFET ; // Fetch operation
  sv1.shvret      = 0           ; // Zero out RC

  sprintf(varnamc,"RXMQ.CAMAP.%lu",pcfnum)            ; // Construct REXX variable name
  MAKERXSTRING(sv1.shvname,varnamc,strlen(varnamc))   ; // Construct REXX variable name structure
  sv1.shvnamelen  = sv1.shvname.strlength             ; // REXX variable name length

  sv1.shvvalue.strptr   = output                      ; // Set pointer to value buffer for REXX
  sv1.shvvalue.strlength= 30                          ; // Set max value length for REXX
  sv1.shvvaluelen       = 30                          ; // Set max accepted value length

  sv1rc           = RexxVariablePool(&sv1)            ; // Call REXX variable interface

  output??(sv1.shvvalue.strlength??) = 0              ; // Ensure zero terminated

  if (   (sv1rc != RXSHV_OK)
      || (sv1.shvname.strlength == 0) )
    sprintf(output,"%lu",pcfnum)                      ;

  DEBUG(EVENT,"Converted " << pcfnum
           << " into " << output << " " )             ;

  DEBUG(EVENT,"Leaving geteventname ") ;

  return ;

 } // End of geteventname function

//
// iscurrthread - tests to see if the Thread the MQCONN was issued upon
//                is the one currently executing.
//
//                Takes a parm of the pointer to the rahmqnTR object
//                and returns TRUE or FALSE accordingly
//
//                Traced by the TR setting
//

int iscurrthread ( rahmqntr * mqotr )
{

 MQULONG   thisid = 0     ;
 int       rc     = FALSE ;

 DEBUG (TR, "Entering iscurrthread, testing object thread " << mqotr->threadid << " " ) ;

 thisid = GetCurrentThreadId() ;
 DEBUG(TR, "GetCurrentThreadId on thread " << thisid << " " ) ;

 if ( mqotr->threadid == thisid ) rc = TRUE ;

 DEBUG(TR, "Leaving iscurrthread "
        << ((rc == TRUE) ? "Threads match " : "Thread mismatch " ) )  ;
 return rc ;

} // End of iscurrthread (object) function


//
// getthreadid  - returns the current thread number
//
//                Traced by the TR setting
//


MQULONG getthreadid ( void )
{

 MQULONG rc = 0 ;

 DEBUG(TR, "Entering getthreadid " ) ;

 rc = GetCurrentThreadId() ;
 DEBUG(TR, "GetCurrentThreadId rc on thread " << rc << " " ) ;

 DEBUG(TR, "Leaving getthreadid " ) ;

 return rc ;

} // End of getthreadid function


//
// isthreadconn - tests to see if the Thread is connected to a Queue Manager
//                and returns FALSE if not, or the handle of the QM to which
//                the thread is connected.
//
//                Traced by the TR setting
//


MQHCONN isthreadconn ( void )
{
 long      i              ;
 MQULONG   thisid = 0     ;
 MQHCONN   rc     = FALSE ;

 DEBUG (TR, "Entering isthreadconn " ) ;

 thisid = GetCurrentThreadId() ;
 DEBUG(TR, "GetCurrentThreadId on thread " << thisid << " " ) ;

 for ( i=MINQS ; i <= MAXQS ; i++ )
    {
     if ( thisid == gmqo??(i??).threadid )
       {
        rc = gmqo??(i??).qmh  ;
        DEBUG(TR, "Match found on " << i <<  " " << rc << " ") ;
        break                 ;
       }
    }

 DEBUG(TR, "Leaving isthreadconn, Thread "
        << ((rc == FALSE) ? "not connected "
                          : "connected " ) )  ;

 return rc ;

} // End of isthreadconn function


//
// External Functions, callable from Rexx
//

//
// initialize the interface - RXMQINIT
//
//    For Windows this function installs the other functions of the DLL.
//    It MUST be called before any of the other Rexx/MQ functions are
//    available for usage.
//
//    This is done by a RxFuncAdd('RXMQINIT','RXMQx',RXMQINIT')
//         call to let Rexx know about the DLL, and then a
//         rcc = RXMQINIT() to call the initialization function.
//
//    The following things are done:
//
//        * Create Rexx variables for ALL known MQ Constants
//        * Create number->name translation variables for MQ ACs
//        * initialize the DLL Global variables
//            (including obtaining the Mutex for Conn/Open/Close processing)
//        * Make the RXMQ... functions known to Rexx
//
//
//
FTYPE  RXMQINIT  RXMQPARM
 {

 MQULONG                  i                ;  // Looper
 MQLONG                   rc = 0           ;  // Function Return Code
 MQLONG                   mqrc = 0         ;  // MQ RC
 MQLONG                   mqac = 0         ;  // MQ AC
 MQULONG                  traceid = INIT   ;  // This function trace id

 RETMSG ReturnMsg[] = {
        { -99, "UNKNOWN FAILURE"}} ;
 //
 // Prepare Copyright message
 //
 char * moremsg = "WMQ Support Pac MA95. Version 1.0.2 "
                  "(C) Copyright IBM Corporation. 1997, 2010.\n";

 set_entry (afuncname, traceid)    ;

 //
 // Firstly, initialize all the constants required by calling setcons.
 //
 // Additionally, if the constant represents Reason Code (MQRC_)
 //               then create an number->name entry in the
 //               RXMQ.RCMAP. stem variable
 //
 // Additionally, if the constant represents Completion Code (MQCC_)
 //               then create an number->name entry in the
 //               RXMQ.CCMAP. stem variable
 //
 // Additionally, if the constant represents Event Name (MQCA_,...)
 //               then create an number->name entry in the
 //               RXMQ.CAMAP. stem variable
 //
 setcons() ;

#ifndef __MVS__
 //
 // Register the DLL functions in Windows
 //
 if ( rc == 0 )
   {
    for (i= 0; i< numfuncs ; i++ )
      {
       rc = RexxRegisterFunctionDll(        funcs??(i??),
                                      (PSZ) THISDLL,
                                            funcs??(i??)) ;
       DEBUG (traceid, "Registration of " << funcs??(i??)
                    << " RC " << rc ) ;
      }
   }
#endif

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,moremsg) ;

return 0 ;
 } // End of RXMQINIT function

//
// RXMQCONS will simply setup the MQ mappings in the Rexx Workspace.
//
//           This is useful in a thread based environment where RXMQINIT
//                has been done elsewhere (so setting the mappings elsewhere)
//                to act, essentially, as the initialization for the thread
//
FTYPE  RXMQCONS  RXMQPARM
 {
 MQLONG                   rc = 0           ;  // Function Return Code
 MQLONG                   mqrc = 0         ;  // MQ RC
 MQLONG                   mqac = 0         ;  // MQ AC
 MQULONG                  traceid = INIT   ;  // This function trace id

 RETMSG ReturnMsg[] = {
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

 //
 // Just initialize all the constants etc. required by calling setcons.
 //
 setcons() ;

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQCONS function

//
// Terminate the interface - RXMQTERM
//
//    For Windows only this function removes access to the functions of the DLL.
//         After calling, no other Rexx/MQ functions are available.
//
//    However, all the REXX name bindings are left in the
//         Workspace; this is done to provide a method of assembling
//         stem variable with the correct settings.
//
//    NO end of process actions are taken. If a prior Syncpoint is
//         not taken, or the Queue Closed, or the QM Disconnected
//         before this routine is invoked, then the usual MQ
//         End of Process action will occur at that time.
//
//
FTYPE  RXMQTERM  RXMQPARM
 {

 MQULONG                  i                ;  // Looper
 MQLONG                   rc = 0           ;  // Function Return Code
 MQLONG                   mqrc = 0         ;  // MQ RC
 MQLONG                   mqac = 0         ;  // MQ AC
 MQULONG                  traceid = TERM   ;  // This function trace id

 RETMSG ReturnMsg[] = {
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

#ifndef __MVS__
//
// Deregister the DLL functions on distributed
//

 for (i= 0; i< numfuncs ; i++ )
   {
    rc = RexxDeregisterFunction(funcs??(i??)) ;
    DEBUG (traceid, "Deregistration of " << funcs??(i??) << " RC " << rc ) ;
   }
#endif

#if _MSC_VER
 MQULONG dosrc = TRUE ;
 DEBUG(traceid,"Deleting the Mutex ") ;
 delete mutex ;
 DEBUG(traceid, " Mutex deletion rc " << dosrc << " (F0 T1) " ) ;

 dosrc = TRUE ;
 DEBUG(traceid,"Deleting the Mutex base ") ;
 delete mutexbase ;
 DEBUG(traceid, " Mutex base deletion rc " << dosrc << " (F0 T1) " ) ;
#endif

// Return a good return Code

// strcpy((char *)returnc,"0 0 0 RXMQTERM OK Rexx MQ Interface terminated." );
//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

#ifdef __MVS__
 return 4;
#else
 return 0;
#endif
 } // End of RXMQTERM function

//
// Do a Connect MQCONN
//
//   Call:   rc = RXMQconn(qmname)
//
FTYPE  RXMQCONN  RXMQPARM
 {

 MQLONG                  i                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = CONN   ;  // This function trace id

 MQLONG                  reqobj = -1      ;  // Connection Object
 char                    qm??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null QM name"},
        {  -3, "Zero length QM name"},
        {  -4, "QM name too long"},
        {  -5, "No available Q Objects"},
        { -95, "Mutex release failed, see reason code"},
        { -96, "Mutex acquisition failed, see reason code"},
        { -98, "Already Connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

 memset(qm, 0, sizeof(qm) )  ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 1 ) )                  rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )     rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) )  rc =  -3 ;
 if ( (rc == 0) && ( aargv??(0??).strlength >
                     MQ_Q_MGR_NAME_LENGTH ) )       rc =  -4 ;
 if ( (rc == 0) && ( isthreadconn() != FALSE) )     rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
   {
    memcpy(qm,aargv??(0??).strptr,aargv??(0??).strlength)   ;
    DEBUG(traceid, "Requested Contact with " << qm << " " ) ;
   }

#if _MSC_VER
//
// During the Scan and Open, the Queue Objects are locked
//       Wait for 30 seconds before timing out.
//       Only in Windows!
//
 MQULONG dosrc = TRUE ;
 MQULONG gotmutex = 0 ;
 if ( rc == 0 )
   {
    DEBUG(traceid, "Acquiring the Mutex " << mutex << " " ) ;
    dosrc = mutex->Lock(MUTEXWAIT) ;
    DEBUG(traceid, "Mutex Acquisition rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex = 1 ;
       DEBUG(traceid, "Now in Critical Section ") ;
      }
    else
      {
       mqac = GetLastError() ;
       rc = -96 ;
      }
   }
#endif
//
// Find a suitable object to do the Connect upon (ie: unconnected as yet)
//
if ( rc == 0 )
  {
   DEBUG(traceid, "Finding an unused Object ") ;

   for ( i=MINQS; i <= MAXQS ; i++ )
     {
      if ( gmqo??(i??).qmh == 0 )
        {
         DEBUG(traceid, "Empty object at " << i << " located" ) ;
         reqobj = i ;
         break      ;
        }
     }

   if ( reqobj == -1 ) rc = -5 ;
  }

//
// Do the MQCONN
//
 if (rc == 0)
   {
    rc   = gmqo??(reqobj??).conn(qm)   ;
    mqrc = gmqo??(reqobj??).rc         ;
    mqac = gmqo??(reqobj??).ac         ;

    DEBUG(traceid, "MQCONN " << reqobj
             << " Handle is "       << gmqo??(reqobj??).qmh
             << " Threadid is "     << gmqo??(reqobj??).threadid
             << " QM is "           << gmqo??(reqobj??).qm
             << " " ) ;
   }

#if _MSC_VER
//
// Now Exit the critical section, if within it
// Only in Windows!
//
 if ( gotmutex != 0 )
   {
    dosrc = TRUE ;
    DEBUG(traceid, "Releasing the Mutex " << mutex << " " ) ;
    dosrc = mutex->Unlock(1,NULL) ;
    DEBUG(traceid, "Mutex Release rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex-- ;
       DEBUG(traceid, "Now left Critical Section ") ;
      }
    else
      {
      mqac = GetLastError() ;
      rc = -95 ;
      }
   }
#endif

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQCONN function

//
// Do a Disconnect   MQDISC
//
//   Call:   rc = RXMQdisc()
//
FTYPE  RXMQDISC  RXMQPARM
 {

 MQULONG                 i                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = DISC   ;  // This function trace id
 MQULONG                 firstdisc = 0    ;  // First time disc flag

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null QM name"},
        { -95, "Mutex release failed, see reason code"},
        { -96, "Mutex acquisition failed, see reason code"},
        { -98, "Not Connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 0 ) )               rc =  -1 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) ) rc = -98 ;

#if _MSC_VER
//
// During the Scan and Open, the Queue Objects are locked
//       Wait for 30 seconds before timing out.
//       Only in Windows!
//
 MQULONG dosrc = TRUE ;
 MQULONG gotmutex = 0 ;
 if ( rc == 0 )
   {
    DEBUG(traceid, "Acquiring the Mutex " << mutex << " " ) ;
    dosrc = mutex->Lock(MUTEXWAIT) ;
    DEBUG(traceid, "Mutex Acquisition rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex = 1 ;
       DEBUG(traceid, "Now in Critical Section ") ;
      }
    else
      {
       mqac = GetLastError() ;
       rc = -96 ;
      }
   }
#endif

//
// Just to be nice, loop through all the open MQ objects and
//         close them before doing the DISC. Do not take any
//         notice of these Return Codes, as they are all supposed
//         to be shut anyway.
//
 if (rc == 0)
   {
    for ( i=MINQS ; i <= MAXQS ; i++ ) // Close Phase
      {
       if ( (gmqo??(i??).qh != 0 ) && ( iscurrthread(&gmqo??(i??)) == TRUE ) )
         {
          gmqo??(i??).close(MQCO_NONE) ;
          gmqo??(i??).qh = 0           ;
         }
      }

#ifndef __MVS__
    for ( i=MINQS ; i <= MAXQS ; i++ )  // Disc Phase
      {
       if ( (gmqo??(i??).qmh != 0 ) && ( iscurrthread(&gmqo??(i??)) == TRUE ) )
         {
          gmqo??(i??).disc()         ;
          gmqo??(i??).qmh = 0        ;
          gmqo??(i??).threadid = 0   ;

          if ( firstdisc == 0 )
            {
             firstdisc = i         ;
             mqrc = gmqo??(i??).rc ;
             mqac = gmqo??(i??).ac ;
             rc   = mqac           ;
            }
         }
      }
#else
    if ( gmqo??(1??).qmh != 0 )
      {
       gmqo??(1??).disc()         ;
       gmqo??(1??).qmh = 0        ;
       gmqo??(1??).threadid = 0   ;
       mqrc = gmqo??(1??).rc      ;
       mqac = gmqo??(1??).ac      ;
       rc   = mqac                ;
      }
#endif

    for ( i=MINQS ; i <= MAXQS ; i++ ) // Zap phase
       {
        if ( (gmqo??(i??).qmh != 0 ) && ( iscurrthread(&gmqo??(i??)) == TRUE ) )
          {
           gmqo??(i??).qmh = 0        ;
           gmqo??(i??).threadid = 0   ;
          }
       }
   }

#if _MSC_VER
//
// Now Exit the critical section, if within it
// Only in Windows!
//
 if ( gotmutex != 0 )
   {
    dosrc = TRUE ;
    DEBUG(traceid, "Releasing the Mutex " << mutex << " " ) ;
    dosrc = mutex->Unlock(1,NULL) ;
    DEBUG(traceid, "Mutex Release rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex-- ;
       DEBUG(traceid, "Now left Critical Section ") ;
      }
    else
      {
      mqac = GetLastError() ;
      rc = -95 ;
      }
   }
#endif

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQDISC function


//
// Do an Open   MQOPEN
//
//   Call:   rc = RXMQopen(iodesc,opts,handle,oodesc)
//
FTYPE  RXMQOPEN  RXMQPARM
 {

 MQULONG                 i                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = OPEN   ;  // This function trace id

 RXSTRING                viodesc          ;  // Stem Var Obj Desc Input
 RXSTRING                voodesc          ;  // Stem Var Obj Desc Output
 RXSTRING                dopts            ;  // Data     Options
 RXSTRING                vhandle          ;  //      Var Obj Handle

 MQHCONN                 qmh              ;  //Connected Queue Manager
 MQOD                    odopen           ;  //MQ object desc
 MQLONG                  options    =  0  ;  //MQ open options
 MQLONG                  theobj     = -1  ;  //gmqo object to use

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null Input OD/Qname"},
        {  -3, "Zero length input OD/Qname"},
        {  -4, "Null options"},
        {  -5, "Zero length options"},
        {  -6, "Null handle name"},
        {  -7, "Zero length handle name"},
        {  -8, "Null Output OD"},
        {  -9, "Zero length output OD"},
        { -10, "No available Q objects"},
        { -95, "Mutex release failed, see reason code"},
        { -96, "Mutex acquisition failed, see reason code"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 4 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -5 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc =  -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc =  -7 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(3??)) )    rc =  -8 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(3??)) ) rc =  -9 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) )   rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
    {
      memcpy(&viodesc ,&aargv??(0??),sizeof(viodesc)) ;
      memcpy(&dopts   ,&aargv??(1??),sizeof(dopts  )) ;
      memcpy(&vhandle ,&aargv??(2??),sizeof(vhandle)) ;
      memcpy(&voodesc ,&aargv??(3??),sizeof(voodesc)) ;

      DEBUG(traceid, "viodesc " << viodesc.strptr  )  ;
      DEBUG(traceid, "dopts   " << dopts.strptr    )  ;
      DEBUG(traceid, "vhandle " << vhandle.strptr  )  ;
      DEBUG(traceid, "voodesc " << voodesc.strptr  )  ;

      make_od_from_stem(&odopen,viodesc)            ;
      parm_to_ulong(dopts, &options)                ;
    }

#if _MSC_VER
//
// During the Scan and Open, the Queue Objects are locked
//       Wait for 30 seconds before timing out.
//       Only in Windows!
//
 MQULONG dosrc = TRUE ;
 MQULONG gotmutex = 0 ;
 if ( rc == 0 )
   {
    DEBUG(traceid, "Acquiring the Mutex " << mutex << " " ) ;
    dosrc = mutex->Lock(MUTEXWAIT) ;
    DEBUG(traceid, "Mutex Acquisition rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex = 1 ;
       DEBUG(traceid, "Now in Critical Section ") ;
      }
    else
      {
       mqac = GetLastError() ;
       rc = -96 ;
      }
   }
#endif

//
// Select the handle (ie: the gmqo to use)
//        Use one with a QMH but no QH, or else an unused one
//
 if (rc == 0)
   {
//
// Get hold of the QMH for the thread
//          (prior call shows that there is one)
//
    DEBUG(traceid, "Getting the QMH ")  ;
    qmh = isthreadconn() ;
    DEBUG(traceid, "Obtained the QMH ") ;

    for ( i=MINQS ; i <= MAXQS ; i++ )
      {
       if ( ( gmqo??(i??).qmh == qmh ) && ( gmqo??(i??).qh == 0 ) )
         {
          theobj = i ;
          DEBUG(traceid, "Selected QH Object " << theobj << " " ) ;
          break      ;
         }
      }

    if ( theobj == -1 ) for ( i=MINQS ; i <= MAXQS ; i++ )
      {
       if ( gmqo??(i??).qmh == 0 )
         {
          theobj = i ;
          DEBUG(traceid, "Selected Empty Object " << theobj << " " ) ;
          break      ;
         }
      }

    if ( (theobj == -1) ) rc = -10 ;
   }

//
// If the Object does not have a QMH, then setup the various entries via a copy
//
 if ( ( rc == 0 ) && ( gmqo??(theobj??).qmh == 0 ) )
   {
    gmqo??(theobj??).qmh      = qmh           ;
    gmqo??(theobj??).threadid = getthreadid() ;
   }

//
// Do the MQOPEN on the obtained Queue object
//
 if (rc == 0)
   {
    DEBUG(traceid, "Calling MQOPEN with options " << options << " " ) ;
    rc   = gmqo??(theobj??).open(&odopen,options) ;
    mqrc = gmqo??(theobj??).rc                    ;
    mqac = gmqo??(theobj??).ac                    ;

    if ( gmqo??(theobj??).qh != 0 )   //If the Open worked,
      {                               //then .....
       stem_from_long(traceid, NULL, vhandle, ""  , theobj) ;
       make_stem_from_od(&odopen,voodesc) ; //and update the OD
      }
   }

#if _MSC_VER
//
// Now Exit the critical section, if within it
// Only in Windows!
//
 if ( gotmutex != 0 )
   {
    dosrc = TRUE ;
    DEBUG(traceid, "Releasing the Mutex " << mutex << " " ) ;
    dosrc = mutex->Unlock(1,NULL) ;
    DEBUG(traceid, "Mutex Release rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex-- ;
       DEBUG(traceid, "Now left Critical Section ") ;
      }
    else
      {
      mqac = GetLastError() ;
      rc = -95 ;
      }
   }
#endif

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQOPEN function

#ifdef _v7
//
// Do a Subscribe MQSUB
//
//   Call:   rc = RXMQsub(isdesc,handle,osdesc)
//
FTYPE  RXMQSUB  RXMQPARM
 {

 MQULONG                 i                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = SUB    ;  // This function trace id

 RXSTRING                visdesc          ;  // Stem Var Sub Desc Input
 RXSTRING                vosdesc          ;  // Stem Var Sub Desc Output
 RXSTRING                vhandle          ;  //      Var Obj Handle

 MQHCONN                 qmh              ;  // Connected Queue Manager
 MQSD                    subdesc          ;  // MQ subscription desc
 MQLONG                  theobj     = -1  ;  // gmqo object to use

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null input SD"},
        {  -3, "Zero length input SD"},
        {  -6, "Null handle name"},
        {  -7, "Zero length handle name"},
        {  -8, "Null Output SD"},
        {  -9, "Zero length Output SD"},
        { -10, "No available objects"},
        { -95, "Mutex release failed, see reason code"},
        { -96, "Mutex acquisition failed, see reason code"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 3 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -7 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc =  -8 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc =  -9 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) )   rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
    {
      memcpy(&visdesc ,&aargv??(0??),sizeof(visdesc)) ;
      memcpy(&vhandle ,&aargv??(1??),sizeof(vhandle)) ;
      memcpy(&vosdesc ,&aargv??(2??),sizeof(vosdesc)) ;

      DEBUG(traceid, "visdesc " << visdesc.strptr  )  ;
      DEBUG(traceid, "vhandle " << vhandle.strptr  )  ;
      DEBUG(traceid, "vosdesc " << vosdesc.strptr  )  ;

      make_sd_from_stem(&subdesc,visdesc)             ;
    }

#if _MSC_VER
//
// During the Scan and Open, the Queue Objects are locked
//       Wait for 30 seconds before timing out.
//       Only in Windows!
//
 MQULONG dosrc = TRUE ;
 MQULONG gotmutex = 0 ;
 if ( rc == 0 )
   {
    DEBUG(traceid, "Acquiring the Mutex " << mutex << " " ) ;
    dosrc = mutex->Lock(MUTEXWAIT) ;
    DEBUG(traceid, "Mutex Acquisition rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex = 1 ;
       DEBUG(traceid, "Now in Critical Section ") ;
      }
    else
      {
       mqac = GetLastError() ;
       rc = -96 ;
      }
   }
#endif

//
// Get hold of the QMH for the thread
//          (prior call shows that there is one)
//
 DEBUG(traceid, "Getting the QMH ")  ;
 qmh = isthreadconn() ;
 DEBUG(traceid, "Obtained the QMH ") ;

//
// Select the handle (ie: the gmqo to use)
//        Use one with a QMH but no QH, or else an unused one
//
 if (rc == 0)
   {
    for ( i=MINQS ; i <= MAXQS ; i++ )
      {
       if ( ( gmqo??(i??).qmh == qmh ) && ( gmqo??(i??).qh == 0 ) )
         {
          theobj = i ;
          DEBUG(OPEN, "Selected Object " << theobj << " " ) ;
          break      ;
         }
      }

    if ( theobj == -1 ) for ( i=MINQS ; i <= MAXQS ; i++ )
      {
       if ( gmqo??(i??).qmh == 0 )
         {
          theobj = i ;
          DEBUG(OPEN, "Selected Empty Object " << theobj << " " ) ;
          break      ;
         }
      }

    if ( (theobj == -1) ) rc = -10;
   }

//
// If the Object does not have a QMH, then setup the various entries via a copy
//
 if ( ( rc == 0 ) && ( gmqo??(theobj??).qmh == 0 ) )
   {
    gmqo??(theobj??).qmh      = qmh           ;
    gmqo??(theobj??).threadid = getthreadid() ;
   }

//
// Do the MQSUB on the obtained Subscription object
//
 if (rc == 0)
   {
    DEBUG(traceid, "Calling MQSUB" )                  ;
    rc   = gmqo??(theobj??).sub(&subdesc)         ;
    mqrc = gmqo??(theobj??).rc                    ;
    mqac = gmqo??(theobj??).ac                    ;

    if ( gmqo??(theobj??).qh != 0 )   //If the Subscribe worked,
      {                               //then .....
       stem_from_long(traceid, NULL, vhandle, ""  , theobj) ;
       make_stem_from_sd(&subdesc,vosdesc) ; //and update the MD
      }
   }

#if _MSC_VER
//
// Now Exit the critical section, if within it
// Only in Windows!
//
 if ( gotmutex != 0 )
   {
    dosrc = TRUE ;
    DEBUG(traceid, "Releasing the Mutex " << mutex << " " ) ;
    dosrc = mutex->Unlock(1,NULL) ;
    DEBUG(traceid, "Mutex Release rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex-- ;
       DEBUG(traceid, "Now left Critical Section ") ;
      }
    else
      {
      mqac = GetLastError() ;
      rc = -95 ;
      }
   }
#endif

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQSUB function
#endif

//
// Do a Close   MQCLOSE
//
//   Call:   rc = RXMQclos(handle,opts)
//
FTYPE  RXMQCLOS  RXMQPARM
 {

 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = CLOSE  ;  // This function trace id

 RXSTRING                dopts            ;  // Data     Options
 RXSTRING                dhandle          ;  // Data     Obj Handle

 MQLONG                  options  = 0     ;  //MQ close options
 MQLONG                  handle   = 0     ;  //MQ object number

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null handle"},
        {  -3, "Zero length handle"},
        {  -4, "Null options"},
        {  -5, "Zero length options"},
        {  -6, "Handle out of range"},
        {  -7, "Invalid handle"},
        { -95, "Mutex release failed, see reason code"},
        { -96, "Mutex acquisition failed, see reason code"},
        { -97, "Handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 2 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -5 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) )   rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
   {
    memcpy(&dhandle,&aargv??(0??),sizeof(dhandle))   ;
    memcpy(&dopts  ,&aargv??(1??),sizeof(dopts  ))   ;

    DEBUG(traceid,  "dopts   " << dopts.strptr   )   ;
    DEBUG(traceid,  "dhandle " << dhandle.strptr )   ;

    parm_to_ulong(dopts,   &options)                 ;
    parm_to_ulong(dhandle, &handle)                  ;
   }

//
// See if the handle is valid (ie: the gmqo to use)
//
 if ( (rc == 0) && ( ( handle > MAXQS ) || ( handle <= 0 ) ) ) rc = -6 ;
 if ( (rc == 0) && ( gmqo??(handle??).qh == 0 ) )              rc = -7 ;

//
// The Handle points to a valid object, see if it is owned by the thread
//
 if ( (rc == 0) && ( iscurrthread(&gmqo??(handle??)) == FALSE ) ) rc = -97 ;

#if _MSC_VER
//
// During the Scan and Open, the Queue Objects are locked
//       Wait for 30 seconds before timing out.
//       Only in Windows!
//
 MQULONG dosrc = TRUE ;
 MQULONG gotmutex = 0 ;
 if ( rc == 0 )
   {
    DEBUG(traceid, "Acquiring the Mutex " << mutex << " " ) ;
    dosrc = mutex->Lock(MUTEXWAIT) ;
    DEBUG(traceid, "Mutex Acquisition rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex = 1 ;
       DEBUG(traceid, "Now in Critical Section ") ;
      }
    else
      {
       mqac = GetLastError() ;
       rc = -96 ;
      }
   }
#endif

//
// Now close the object
//
 if (rc == 0)
   {
    rc   = gmqo??(handle??).close(options)        ;
    mqrc = gmqo??(handle??).rc                    ;
    mqac = gmqo??(handle??).ac                    ;

    if ( mqac == 0 )                  //If the Close worked,
      {                               //then .....
       gmqo??(handle??).qh = 0 ;      //loose the MQ object
      }
   }

#if _MSC_VER
//
// Now Exit the critical section, if within it
// Only in Windows!
//
 if ( gotmutex != 0 )
   {
    dosrc = TRUE ;
    DEBUG(traceid, "Releasing the Mutex " << mutex << " " ) ;
    dosrc = mutex->Unlock(1,NULL) ;
    DEBUG(traceid, "Mutex Release rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex-- ;
       DEBUG(traceid, "Now left Critical Section ") ;
      }
    else
      {
      mqac = GetLastError() ;
      rc = -95 ;
      }
   }
#endif

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQCLOS function

//
// Do a Syncpoint    MQCMIT
//
//   Call:   rc = RXMQcmit()
//
FTYPE  RXMQCMIT  RXMQPARM
 {

 MQULONG                 i                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = CMIT   ;  // This function trace id

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 0 ) )               rc =  -1 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) ) rc = -98 ;

//
// Do the Actual Syncpoint on the required Queue Manager
//        This also forces it for all others on the thread
//
 if (rc == 0)
   {
    for ( i=MINQS ; i <= MAXQS ; i++ )
      {
       if ( (gmqo??(i??).qh != 0 ) && ( iscurrthread(&gmqo??(i??)) == TRUE ) )
         {
          DEBUG(traceid, "Doing the Synchpoint on object " << i << " " ) ;
          rc   = gmqo??(i??).commit();
          mqrc = gmqo??(i??).rc      ;
          mqac = gmqo??(i??).ac      ;

          break ; // Only bother with the first one
         }
      }
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQCMIT function

//
// Do a Rollback     MQBACK
//
//   Call:   rc = RXMQback()
//
FTYPE  RXMQBACK  RXMQPARM
 {

 MQULONG                 i                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = BACK   ;  // This function trace id

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 0 ) )               rc =  -1 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) ) rc = -98 ;

//
// Do the Actual Backout on the required Queue Manager
//        This also forces it for all others on the thread
//
 if (rc == 0)
   {
    for ( i=MINQS ; i <= MAXQS ; i++ )
      {
       if ( (gmqo??(i??).qh != 0 ) && ( iscurrthread(&gmqo??(i??)) == TRUE ) )
         {
          DEBUG(traceid, "Doing the Backout on object " << i << " " ) ;
          rc   = gmqo??(i??).rollback()  ;
          mqrc = gmqo??(i??).rc          ;
          mqac = gmqo??(i??).ac          ;

          break ; // Only bother for the first one
         }
      }
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

 return 0;
 } // End of RXMQBACK function

//
// Do a Put     MQPUT
//
//   Call:   rc = RXMQput(handle, data,
//                        input_msgdesc, output_msgdesc,
//                        input_pmo, output_pmo)
//
FTYPE  RXMQPUT  RXMQPARM
 {

 MQLONG                  rc   = 0         ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = PUT    ;  // This function trace id

 RXSTRING                dhandle          ;  // Obj Handle
 RXSTRING                vdata            ;  // Variable Data
 RXSTRING                vides            ;  // Variable Input  Msg Desc
 RXSTRING                vodes            ;  // Variable Output Msg Desc
 RXSTRING                vipmo            ;  // Variable Input  PMO
 RXSTRING                vopmo            ;  // Variable Output PMO

 MQLONG                  handle           ;  //MQ object number
 MQMD2                   desc             ;  //MQ Message descriptor
 MQPMO                   pmo              ;  //MQ Put Message options
 MQLONG                  data0 = 0        ;  // Variable Data len
 void                 *  data  = 0        ;  //-> Data buffer
 MQLONG                  datalen          ;  //   Data length

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null handle"},
        {  -3, "Zero data handle"},
        {  -4, "Null data stem var"},
        {  -5, "Zero data stem var"},
        {  -6, "Null input MsgDesc"},
        {  -7, "Zero length input MsgDesc"},
        {  -8, "Null output MsgDesc"},
        {  -9, "Zero length output MsgDesc"},
        { -10, "Null input PMO"},
        { -11, "Zero length input PMO"},
        { -12, "Null output PMO"},
        { -13, "Zero length output PMO"},
        { -14, "Handle out of range"},
        { -15, "Invalid handle"},
        { -16, "malloc failure, check reason code"},
        { -17, "Zero length input data buffer"},
        { -18, "Data length is not equal to specified value"},
        { -19, "Context handle out of range"},
        { -20, "Invalid Context handle"},
        { -96, "Context handle not owned by current thread"},
        { -97, "Handle not owned by current thread "},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 6 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -5 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc =  -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc =  -7 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(3??)) )    rc =  -8 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(3??)) ) rc =  -9 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(4??)) )    rc = -10 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(4??)) ) rc = -11 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(5??)) )    rc = -12 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(5??)) ) rc = -13 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) )   rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
   {
    memcpy(&dhandle,&aargv??(0??),sizeof(dhandle)) ;
    memcpy(&vdata,  &aargv??(1??),sizeof(vdata))   ;
    memcpy(&vides,  &aargv??(2??),sizeof(vides))   ;
    memcpy(&vodes,  &aargv??(3??),sizeof(vodes))   ;
    memcpy(&vipmo,  &aargv??(4??),sizeof(vipmo))   ;
    memcpy(&vopmo,  &aargv??(5??),sizeof(vopmo))   ;

    DEBUG(traceid,  "dhandle " << dhandle.strptr ) ;
    DEBUG(traceid,  "vdata   " << vdata.strptr   ) ;
    DEBUG(traceid,  "vides   " << vides.strptr   ) ;
    DEBUG(traceid,  "vodes   " << vodes.strptr   ) ;
    DEBUG(traceid,  "vimpo   " << vipmo.strptr   ) ;
    DEBUG(traceid,  "vopmo   " << vopmo.strptr   ) ;

    make_md_from_stem(&desc, vides )               ;
    make_po_from_stem(&pmo , vipmo )               ;

    parm_to_ulong(dhandle, &handle)                ;
    stem_to_long(traceid, vdata, "0" , &data0)     ;
   }

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && ( data0 <= 0 ) ) rc = -17    ;

//
// Allocate data buffer to store stem.1 variable data.
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Doing malloc for " << data0 << " bytes ") ;
    data = malloc(data0) ;
    if ( data == NULL ) mqac = errno ;
    DEBUG(traceid,"malloc RC " << mqac << " PTR is "<<hex<< data);
    if ( mqac != 0 ) rc = -16;
   }

 if (rc == 0)
   {
    datalen = stem_to_data(traceid, vdata, "1", (MQBYTE *)data, data0);

    DEBUG(traceid, "Value at " << vdata.strptr
                           << " Length " << datalen
                           << " Data <"  << data
                           << "> " ) ;
   }

//
// Now check if stem.0 specifies the same value as stem.1 length.
// If not, don't know what to do.
//
 if ( (rc == 0) && ( datalen != data0 ) ) rc = -18;

//
// See if the output queue handle is valid (ie: the gmqo to use)
//
 if ( (rc == 0) && ( ( handle > MAXQS ) || ( handle <= 0 ) ) )    rc = -14 ;
 if ( (rc == 0) && ( gmqo??(handle??).qh == 0 ) )                 rc = -15 ;
 if ( (rc == 0) && ( iscurrthread(&gmqo??(handle??)) == FALSE ) ) rc = -97 ;


//
// If Context is specified (for MQPMO_PASS_ALL_CONTEXT)
// see if the input queue handle is valid (ie: the gmqo to use)
//
 if ( (rc == 0) && ( pmo.Context != 0 ) )
   {
    if ( ( pmo.Context > MAXQS ) || ( pmo.Context < 0 ) )  rc = -19 ;
    if ( (rc == 0) && (gmqo??(pmo.Context??).qh == 0 ) )   rc = -20 ;
    if ( (rc == 0) && (iscurrthread(&gmqo??(pmo.Context??)) == FALSE ) ) rc = -96;
    if   (rc == 0) pmo.Context = gmqo??(pmo.Context??).qh;
    }

//
// Now put the data to the queue
//
 if (rc == 0)
   {
    DEBUG(traceid, "PUT Maxdatalen is " << data0 << " ") ;

    rc   = gmqo??(handle??).put(&desc,&pmo,
                                data0,(char *)data)  ;
    mqrc = gmqo??(handle??).rc                       ;
    mqac = gmqo??(handle??).ac                       ;

    DEBUG(traceid, "PUT RC " << mqrc << " AC " << mqac << " " ) ;

    make_stem_from_md(&desc, vodes ) ;   //Set the return Variables
    make_stem_from_po(&pmo , vopmo ) ;
   }

//
// Free data buffer for stem.1 variable data, if allocated.
//
 if ( data != 0 )
   {
    DEBUG(traceid, "Freemaining area " << data) ;
    free(data) ;
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

 return 0;
 } // End of RXMQPUT function

//
// Do a put    MQPUT1
//
//   Call:   rc = RXMQPUT1(input1_objdesc, output1_objdesc, data,
//                         input1_msgdesc,output1_msgdesc,
//                         input1_pmo,output1_pmo)
//        or
//           rc = RXMQPUT1(queue_name, output1_objdesc, data,
//                         input1_msgdesc,output1_msgdesc,
//                         input1_pmo,output1_pmo)
//
FTYPE RXMQPUT1  RXMQPARM
 {
 MQULONG                 i                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = PUT1   ;  // This function trace id

 RXSTRING         viodesc      ;  // Stem Var Obj Desc Input
 RXSTRING         voodesc      ;  // Stem Var Obj Desc Output

 RXSTRING                vdata            ;  // Variable Data
 RXSTRING                vides            ;  // Variable Input1  Msg Desc
 RXSTRING                vodes            ;  // Variable Output1 Msg Desc
 RXSTRING                vipmo            ;  // Variable Input1  PMO
 RXSTRING                vopmo            ;  // Variable Output1 PMO

 MQHCONN                 qmh              ;
 MQOD                    odopen           ;
 MQMD2                   desc             ;  //MQ Message descriptor
 MQPMO                   pmo              ;  //MQ put1 Message options
 void                 *  data   = 0       ;  //-> Data buffer
 MQLONG                  data0 = 0        ;  // Variable Data len
 MQLONG                  datalen          ;  //   Data length
 MQLONG                  theobj = -1      ;
 char                    theobjc??(10??)  ;
 RXSTRING                theobjx          ;

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null input OD/Qname"},
        {  -3, "Zero length input OD/Qname"},
        {  -4, "Null output OD"},
        {  -5, "Zero length output OD"},
        {  -6, "Null data stem var"},
        {  -7, "Zero data stem var"},
        {  -8, "Null input MsgDesc"},
        {  -9, "Zero length input MsgDesc"},
        { -10, "Null output MsgDesc"},
        { -11, "Zero length output MsgDesc"},
        { -12, "Null input PMO"},
        { -13, "Zero length input PMO"},
        { -14, "Null output PMO"},
        { -15, "Zero length output PMO"},
        { -16, "No available Q objects"},
        { -17, "malloc failure, check reason code"},
        { -18, "Zero length input data buffer"},
        { -19, "Data length is not equal to specified value"},
        { -20, "Context handle out of range"},
        { -21, "Invalid context handle"},
        { -97, "Context handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 7 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -5 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc =  -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc =  -7 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(3??)) )    rc =  -8 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(3??)) ) rc =  -9 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(4??)) )    rc = -10 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(4??)) ) rc = -11 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(5??)) )    rc = -12 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(5??)) ) rc = -13 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(6??)) )    rc = -14 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(6??)) ) rc = -15 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) )   rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
   {
    memcpy(&viodesc,&aargv??(0??),sizeof(viodesc)) ;
    memcpy(&voodesc,&aargv??(1??),sizeof(voodesc)) ;
    memcpy(&vdata,&aargv??(2??),sizeof(vdata))     ;
    memcpy(&vides,&aargv??(3??),sizeof(vides))     ;
    memcpy(&vodes,&aargv??(4??),sizeof(vodes))     ;
    memcpy(&vipmo,&aargv??(5??),sizeof(vipmo))     ;
    memcpy(&vopmo,&aargv??(6??),sizeof(vopmo))     ;

    DEBUG(traceid,  "viodesc " << viodesc.strptr ) ;
    DEBUG(traceid,  "voodesc " << voodesc.strptr ) ;
    DEBUG(traceid,  "vdata   " << vdata.strptr   ) ;
    DEBUG(traceid,  "vides   " << vides.strptr   ) ;
    DEBUG(traceid,  "vodes   " << vodes.strptr   ) ;
    DEBUG(traceid,  "vimpo   " << vipmo.strptr   ) ;
    DEBUG(traceid,  "vopmo   " << vopmo.strptr   ) ;

    make_od_from_stem(&odopen, viodesc           ) ;
    make_md_from_stem(&desc,   vides             ) ;
    make_po_from_stem(&pmo ,   vipmo             ) ;
    stem_to_long     (traceid, vdata, "0" , &data0  ) ;
   }

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && ( data0 <= 0 ) )  rc = -18 ;

//
// Allocate data buffer to store stem.1 variable data.
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Doing malloc for " << data0 << " bytes ") ;
    data = malloc(data0) ;
    if ( data == NULL ) mqac = errno ;
    DEBUG(traceid,"malloc RC " << mqac << " PTR is "<<hex<< data);
    if ( mqac != 0 )  rc = -17 ;
   }

 if (rc == 0)
   {
    datalen = stem_to_data(traceid, vdata, "1", (MQBYTE *)data, data0);

    DEBUG(traceid, "Value at " << vdata.strptr
                << " Length "  << datalen
                << " Data <"   << data
                << "> " ) ;
   }

//
// Now check if stem.0 specifies the same value as stem.1 length.
// If not, don't know what to do.
//
 if ( (rc == 0) && ( datalen != data0 ) ) rc = -19 ;

#if _MSC_VER
//
// During the Scan and Open, the Queue Objects are locked
//       Wait for 30 seconds before timing out.
//       Only in Windows!
//
 MQULONG dosrc = TRUE ;
 MQULONG gotmutex = 0 ;
 if ( rc == 0 )
   {
    DEBUG(traceid, "Acquiring the Mutex " << mutex << " " ) ;
    dosrc = mutex->Lock(MUTEXWAIT) ;
    DEBUG(traceid, "Mutex Acquisition rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex = 1 ;
       DEBUG(traceid, "Now in Critical Section ") ;
      }
    else
      {
       mqac = GetLastError() ;
       rc = -96 ;
      }
   }
#endif

//
// Get hold of the QMH for the thread
//          (prior call shows that there is one)
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Getting the QMH ")  ;
    qmh = isthreadconn() ;
    DEBUG(traceid, "Obtained the QMH ") ;
   }

//
// Select the handle (ie: the gmqo to use)
// Use one with the same QMH and same thread
//
 if (rc == 0)
   {
    for ( i=MINQS ; i <= MAXQS ; i++ )
      {
       if ( ( gmqo??(i??).qmh == qmh ) && ( iscurrthread(&gmqo??(i??)) ) )
         {
          theobj = i ;
          sprintf(theobjc,"%lu",theobj)                            ;
          DEBUG(traceid, "Selected QH Object " << theobjc << " " ) ;
          MAKERXSTRING(theobjx,theobjc,strlen(theobjc))            ;
          break      ;
         }
      }
    if ( (theobj == -1) ) rc = -16 ;
   }

 //
 // If Context is specified (for MQPMO_PASS_ALL_CONTEXT)
 // see if the input queue handle is valid (ie: the gmqo to use)
 //
  if ( (rc == 0) && ( pmo.Context != 0 ) )
    {
     if ( ( pmo.Context > MAXQS ) || ( pmo.Context < 0 ) ) rc = -20 ;
     if ( (rc == 0) & (gmqo??(pmo.Context??).qh == 0 ) )   rc = -21 ;
     if ( (rc == 0) & (iscurrthread(&gmqo??(pmo.Context??)) == FALSE ) ) rc = -97;
     if (rc == 0) pmo.Context = gmqo??(pmo.Context??).qh;
    }

//
// Now execute PUT1
//
 if (rc == 0)
   {
    DEBUG(traceid, "PUT1 Maxdatalen is " << data0 << " ")     ;
    rc   = gmqo??(theobj??).put1(&desc,&odopen,&pmo,
                                        data0,(char *)data)   ;
    mqrc = gmqo??(theobj??).rc                                ;
    mqac = gmqo??(theobj??).ac                                ;

    DEBUG(traceid, "PUT1 RC " << mqrc
                << " AC "     << mqac
                << " " ) ;

    make_stem_from_md(&desc, vodes      ) ;   //Set the return Variables
    make_stem_from_po(&pmo , vopmo      ) ;
    make_stem_from_od(&odopen , voodesc ) ;
   }

#if _MSC_VER
//
// Now Exit the critical section, if within it
// Only in Windows!
//
 if ( gotmutex != 0 )
   {
    dosrc = TRUE ;
    DEBUG(traceid, "Releasing the Mutex " << mutex << " " ) ;
    dosrc = mutex->Unlock(1,NULL) ;
    DEBUG(traceid, "Mutex Release rc " << dosrc << " (F0 T1) " ) ;

    if ( dosrc == TRUE )
      {
       gotmutex-- ;
       DEBUG(traceid, "Now left Critical Section ") ;
      }
    else
      {
      mqac = GetLastError() ;
      rc = -95 ;
      }
   }
#endif

//
// Free data buffer for stem.1 variable data, if allocated.
//
 if ( data != 0 )
   {
    DEBUG(traceid, "Freemaining area " << data) ;
    free(data) ;
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
} // End of RXMQPUT1 function

//
// Do a Get     MQGET
//
//   Call:   rc = RXMQget(handle, data,
//                        input_msgdesc,output_msgdesc,
//                        input_gmo,output_gmo)
//
FTYPE  RXMQGET  RXMQPARM
 {

 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = GET    ;  // This function trace id

 RXSTRING                vdata            ;  // Variable Data
 RXSTRING                vides            ;  // Variable Input  Msg Desc
 RXSTRING                vodes            ;  // Variable Output Msg Desc
 RXSTRING                vigmo            ;  // Variable Input  GMO
 RXSTRING                vogmo            ;  // Variable Output GMO
 RXSTRING                dhandle          ;  // Obj Handle
 MQLONG                  handle       = 0 ;  //MQ object number

 MQMD2                   desc             ;  //MQ Message descriptor
 MQGMO                   gmo              ;  //MQ Get Message options
 void                 *  buffer       = 0 ;  //-> Data buffer
 MQLONG                  bufflen      = 0 ;  //   Data length max
 MQLONG                  buffcurlen   = 0 ;  //   Data length actual

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null handle"},
        {  -3, "Zero data handle"},
        {  -4, "Null data stem var"},
        {  -5, "Zero data stem var"},
        {  -6, "Null input MsgDesc"},
        {  -7, "Zero length input MsgDesc"},
        {  -8, "Null output MsgDesc"},
        {  -9, "Zero length output MsgDesc"},
        { -10, "Null input GMO"},
        { -11, "Zero length input GMO"},
        { -12, "Null output GMO"},
        { -13, "Zero length output GMO"},
        { -14, "Handle out of range"},
        { -15, "Invalid handle"},
        { -16, "malloc failure, check reason code"},
        { -17, "Zero length input data buffer"},
        { -97, "Handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//

 if ( (rc == 0) && (aargc != 6 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -5 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc =  -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc =  -7 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(3??)) )    rc =  -8 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(3??)) ) rc =  -9 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(4??)) )    rc = -10 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(4??)) ) rc = -11 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(5??)) )    rc = -12 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(5??)) ) rc = -13 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) )   rc = -98 ;

//
// Now the parms are correct, get them
//

 if (rc == 0)
   {
    memcpy(&dhandle,&aargv??(0??),sizeof(dhandle)) ;
    memcpy(&vdata,&aargv??(1??),sizeof(vdata))     ;
    memcpy(&vides,&aargv??(2??),sizeof(vides))     ;
    memcpy(&vodes,&aargv??(3??),sizeof(vodes))     ;
    memcpy(&vigmo,&aargv??(4??),sizeof(vigmo))     ;
    memcpy(&vogmo,&aargv??(5??),sizeof(vogmo))     ;

    DEBUG(traceid,  "dhandle " << dhandle.strptr ) ;
    DEBUG(traceid,  "vdata   " << vdata.strptr   ) ;
    DEBUG(traceid,  "vides   " << vides.strptr   ) ;
    DEBUG(traceid,  "vodes   " << vodes.strptr   ) ;
    DEBUG(traceid,  "vigmo   " << vigmo.strptr   ) ;
    DEBUG(traceid,  "vogmo   " << vogmo.strptr   ) ;


    make_md_from_stem(&desc, vides )               ;
    make_go_from_stem(&gmo , vigmo )               ;
    parm_to_ulong(dhandle, &handle)                ;
    stem_to_long(traceid, vdata, "0" , &bufflen)       ;
   }

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
  if ( (rc == 0) && ( bufflen <= 0 ) ) rc = -17 ;

//
// See if the handle is valid (ie: the gmqo to use)
//
 if ( (rc == 0) && ( ( handle > MAXQS ) || ( handle <= 0 ) ) ) rc = -14 ;
 if ( (rc == 0) && ( gmqo??(handle??).qh == 0 ) )              rc = -15 ;

//
// The Handle points to a valid object, see if it is owned by the thread
//
 if ( (rc == 0) && ( iscurrthread(&gmqo??(handle??)) == FALSE ) ) rc = -97 ;

//
// Now GETMAIN the buffer to receive the data records
//
 if (( rc == 0) && ( bufflen != 0 ) )
   {
    DEBUG(traceid, "Doing malloc for " << bufflen << " bytes ") ;
    buffer = malloc(bufflen) ;
    if ( buffer == NULL ) mqac = errno ;
    DEBUG(traceid,"malloc RC " << mqac << " PTR is "<<hex<< buffer);
    if ( mqac != 0 ) rc = -16 ;
   }

//
// Now get the data from the queue
//
 if (rc == 0)
   {
    DEBUG(traceid, "GET Maxdatalen is " << bufflen << " ")    ;
    rc   = gmqo??(handle??).get(&desc, &gmo,
                                        bufflen,
                                        (char *)buffer,
                                        &buffcurlen)      ;
    mqrc = gmqo??(handle??).rc                            ;
    mqac = gmqo??(handle??).ac                            ;

    DEBUG(traceid, "GET RC " << mqrc << " AC "
                         << mqac << " Datalen "
                         << buffcurlen << " " ) ;

    make_stem_from_md(&desc, vodes          ) ;  //Set the return Variables
    make_stem_from_go(&gmo , vogmo          ) ;

    stem_from_long (traceid, NULL, vdata, "0" , buffcurlen)                   ;
    if (buffcurlen > bufflen) buffcurlen = bufflen                        ;
    stem_from_bytes(traceid, NULL, vdata, "1" , (MQBYTE *)buffer, buffcurlen) ;
   }

//
// Free data buffer for stem.1 variable data, if allocated.
//
 if ( buffer != 0 )
   {
    DEBUG(traceid, "Freemaining area " << buffer) ;
    free(buffer) ;
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQGET function

//
// Do an inquire MQINQ
//
//   Call:   rc = RXMQinq(handle, attribute, atttrsetting )
//
//           NB: only 1 attribute at a time!
//
//
FTYPE  RXMQINQ  RXMQPARM
 {

 MQLONG           rc = 0               ;  // Function Return Code
 MQLONG           mqrc = 0             ;  // MQ RC
 MQLONG           mqac = 0             ;  // MQ AC
 MQULONG          traceid = INQ        ;  // This function trace id

 RXSTRING         voattr               ;  // Variable Output Attr
 RXSTRING         viattr               ;  // Variable Input  Attr
 RXSTRING         dhandle              ;  // Data     Obj Handle
 MQLONG           handle       = 0     ;  //MQ object number
 MQLONG           curattr      = 0     ;  // Current Attr

 MQLONG           inqselcount  = 1     ;  //INQ - number of sels
 MQLONG           inqsels      = 0     ;  //INQ - the selectors
 MQLONG           inqselicount = 0     ;  //INQ - number of int sels
 MQLONG           inqints      = 0     ;  //INQ - integer return
 MQLONG           inqcharlen   = 600   ;  //INQ - char return length
 char             inqchars??(601??)    ;  //INQ - char return

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null handle"},
        {  -3, "Zero data handle"},
        {  -4, "Null data input attr"},
        {  -5, "Zero data input attr"},
        {  -6, "Null output attr"},
        {  -7, "Zero length output attr"},
        {  -8, "No attribute supplied"},
        {  -9, "Attribute out of valid range"},
        { -10, "Handle out of range"},
        { -11, "Invalid handle"},
        { -97, "Handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 3 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -5 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc =  -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc =  -7 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) )   rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
   {
    memcpy(&dhandle,&aargv??(0??),sizeof(dhandle)) ;
    memcpy(&viattr,&aargv??(1??),sizeof(viattr))   ;
    memcpy(&voattr,&aargv??(2??),sizeof(voattr))   ;

    DEBUG(traceid,  "dhandle " << dhandle.strptr ) ;
    DEBUG(traceid,  "viattr  " << viattr.strptr  ) ;
    DEBUG(traceid,  "voattr  " << voattr.strptr  ) ;

    parm_to_ulong(dhandle, &handle)            ;
    parm_to_ulong(viattr,  &curattr)           ;

    if (curattr == 0) rc = -8                  ;

    if ( !( ( (curattr >= MQIA_FIRST) && (curattr <= MQIA_LAST  ) )
         || ( (curattr >= MQCA_FIRST) && (curattr <= MQCA_LAST  ) ) ) ) rc = -9;
   }
//
// See if the handle is valid (ie: the gmqo to use)
//
 if ( (rc == 0) && ( ( handle > MAXQS ) || ( handle <= 0 ) ) ) rc = -10 ;
 if ( (rc == 0) && ( gmqo??(handle??).qh == 0 ) )              rc = -11 ;

//
// The Handle points to a valid object, see if it is owned by the thread
//
 if ( (rc == 0) && ( iscurrthread(&gmqo??(handle??)) == FALSE ) ) rc = -97 ;

//
// Format up either a character or an integer area
//
 if (rc == 0)
   {
    if ( (curattr >= MQCA_FIRST  ) && (curattr <= MQCA_LAST  ) )
      {
       inqselcount  = 1                    ;
       inqsels      = curattr              ;
       inqselicount = 0                    ;
       inqints      = 0                    ;
       inqcharlen   = 600                  ;
       memset(inqchars,0,sizeof(inqchars)) ;
      }      //End of Char Attr processing
    else     //Int attr processing
      {
       inqselcount  = 1                    ;
       inqsels      = curattr              ;
       inqselicount = 1                    ;
       inqints      = 0                    ;
       inqcharlen   = 0                    ;
       memset(inqchars,0,sizeof(inqchars)) ;
      }     //End of Int Attr processing

   } //End of Attribute building

//
// Now do the Inquiry and return the setting
//
 if (rc == 0)
   {
    DEBUG(traceid, "Attr " << inqsels
            << " IntSelNum " << inqselicount
            << " ChrAtrLen " << inqcharlen   );

    rc   = gmqo??(handle??).inq(inqselcount , inqsels           ,
                                inqselicount, &inqints          ,
                                inqcharlen  , (char *)inqchars) ;
    mqrc = gmqo??(handle??).rc                                  ;
    mqac = gmqo??(handle??).ac                                  ;

    DEBUG(traceid, "INQ RC "   << mqrc
            << " AC "      << mqac
            << " Intval "  << inqints
            << " Charval " << inqchars ) ;

    if ( (curattr >= MQCA_FIRST  ) && (curattr <= MQCA_LAST) )
      stem_from_string(traceid, NULL, voattr, "", inqchars, inqcharlen) ;
    else
      stem_from_long  (traceid, NULL, voattr, "", inqints) ;
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQINQ function

//
// Do a set      MQSET
//
//   Call:   rc = RXMQset(handle, attribute, atttrsetting )
//
//           NB: only 1 attribute at a time!
//
//
FTYPE  RXMQSET  RXMQPARM
 {

 MQLONG           rc = 0               ;  // Function Return Code
 MQLONG           mqrc = 0             ;  // MQ RC
 MQLONG           mqac = 0             ;  // MQ AC
 MQULONG          traceid = SET        ;  // This function trace id

 RXSTRING         viattr               ;  // Variable Input  Attr
 RXSTRING         vsattr               ;  // Variable Set    Attr
 RXSTRING         dhandle              ;  // Data     Obj Handle
 MQLONG           handle       = 0     ;  //MQ object number
 MQLONG           curattr      = 0     ;  // Current Attr

 MQLONG           number       = 0     ;  // Char->Integer Conver
 MQLONG           setselcount  = 1     ;  //SET - number of sels
 MQLONG           setsels      = 0     ;  //SET - the selectors
 MQLONG           setselicount = 0     ;  //SET - number of int sels
 MQLONG           setints      = 0     ;  //SET - integer return
 MQLONG           setcharlen   = 600   ;  //SET - char return length
 char             setchars??(601??)    ;  //SET - char return

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null handle"},
        {  -3, "Zero data handle"},
        {  -4, "Null data attribute"},
        {  -5, "Zero data attribute"},
        {  -6, "Null setting"},
        {  -7, "Zero length setting"},
        {  -8, "No attribute supplied"},
        {  -9, "Attribute out of valid range"},
        { -10, "Handle out of range"},
        { -11, "Invalid handle"},
        { -97, "Handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//

 if ( (rc == 0) && (aargc != 3 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -5 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc =  -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc =  -7 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE ) )   rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
   {
    memcpy(&dhandle,&aargv??(0??),sizeof(dhandle)) ;
    memcpy(&viattr,&aargv??(1??),sizeof(viattr))   ;
    memcpy(&vsattr,&aargv??(2??),sizeof(vsattr))   ;

    DEBUG(traceid,  "dhandle " << dhandle.strptr ) ;
    DEBUG(traceid,  "viattr  " << viattr.strptr  ) ;
    DEBUG(traceid,  "vsattr  " << vsattr.strptr  ) ;

    parm_to_ulong(dhandle, &handle)            ;
    parm_to_ulong(viattr,  &curattr)           ;

    DEBUG(traceid, "Value at " << &vsattr
            << " Length " << vsattr.strlength
            << " Data <"  << vsattr.strptr
            << "> " ) ;

    if (curattr == 0) rc = -8 ;

    if ( !( ( (curattr >= MQIA_FIRST) && (curattr <= MQIA_LAST  ) )
         || ( (curattr >= MQCA_FIRST) && (curattr <= MQCA_LAST  ) ) ) ) rc = -9;
   }

//
// See if the handle is valid (ie: the gmqo to use)
//
 if ( (rc == 0) && ( ( handle > MAXQS ) || ( handle <= 0 ) ) ) rc = -10 ;
 if ( (rc == 0) && ( gmqo??(handle??).qh == 0 ) )              rc = -11 ;

//
// The Handle points to a valid object, see if it is owned by the thread
//
if ( (rc == 0) && ( iscurrthread(&gmqo??(handle??)) == FALSE ) ) rc = -97 ;

//
// Format up either a character or an integer area
//
 if (rc == 0)
   {
    if ( (curattr >= MQCA_FIRST  ) && (curattr <= MQCA_LAST  ) )
      {
       setselcount  = 1                    ;
       setsels      = curattr              ;
       setselicount = 0                    ;
       setints      = 0                    ;
       setcharlen   = 600                  ;
       memset(setchars,' ',sizeof(setchars))              ;
       memcpy(setchars,vsattr.strptr,vsattr.strlength)    ;

      }      //End of Char Attr processing
    else      //Int attr processing
      {
       setselcount  = 1                                   ;
       setsels      = curattr                             ;
       setselicount = 1                                   ;
       setints      = 0                                   ;
       setcharlen   = 0                                   ;
       memset(setchars,0,sizeof(setchars))                ;
       parm_to_ulong(vsattr,  &number)                    ;
       setints = number                                   ;
      }      //End of Int Attr processing

   } //End of Attribute building

//
// Now do the Setting
//
 if (rc == 0)
   {
    DEBUG(traceid, "Attr " << setsels
            << " IntSelNum " << setselicount
            << " IntSetVal " << setints
            << " ChrAtrLen " << setcharlen
            << " ChrSetVar " << setchars     );

    rc   = gmqo??(handle??).set(setselcount , setsels           ,
                                setselicount, &setints          ,
                                setcharlen  , (char *)setchars) ;
    mqrc = gmqo??(handle??).rc                                  ;
    mqac = gmqo??(handle??).ac                                  ;

    DEBUG(traceid, "SET RC " << mqrc << " AC " << mqac ) ;
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQSET function

//
// Extension functions
//

//
// Do a Browse
//
//   Call:   rc = RXMQbrws(handle, data)
//
//           This function issues a browse on the Queue, and
//                returns the next message (if available). If you
//                need the Message descriptor as well, then use the
//                native Get interface.
//
//
FTYPE  RXMQBRWS  RXMQPARM
 {

 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = BRO    ;  // This function trace id

 RXSTRING                vdata            ;  // Variable Data
 RXSTRING                dhandle          ;  // Data     Obj Handle
 MQLONG                  handle       = 0 ;  //MQ object number
 void                 *  buffer       = 0 ;  //-> Data buffer
 MQLONG                  bufflen      = 0 ;  //   Data length max
 MQLONG                  buffcurlen   = 0 ;  //   Data length actual

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null handle"},
        {  -3, "Zero data handle"},
        {  -4, "Null data stem var"},
        {  -5, "Zero data stem var"},
        {  -6, "Handle out of range"},
        {  -7, "Invalid handle"},
        {  -8, "malloc failure, check reason code"},
        {  -9, "Zero length input data buffer"},
        { -97, "Handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 2 ) )                 rc =  -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc =  -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc =  -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc =  -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc =  -5 ;
 if ( (rc == 0) && ( isthreadconn() == FALSE) )    rc = -98 ;

//
// Now the parms are correct, get them
//
 if (rc == 0)
   {
    memcpy(&dhandle,&aargv??(0??),sizeof(dhandle)) ;
    memcpy(&vdata,&aargv??(1??),sizeof(vdata))     ;

    DEBUG(traceid,  "dhandle " << dhandle.strptr ) ;
    DEBUG(traceid,  "vdata   " << vdata.strptr   ) ;

    parm_to_ulong(dhandle, &handle)            ;
    stem_to_long(traceid, vdata, "0" , &bufflen)   ;
   }

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && ( bufflen <= 0 ) ) rc = -9 ;

//
// See if the handle is valid (ie: the gmqo to use)
//
 if ( (rc == 0) && ( ( handle > MAXQS ) || ( handle <= 0 ) ) ) rc = -6 ;
 if ( (rc == 0) && ( gmqo??(handle??).qh == 0 ) )              rc = -7 ;

//
// The Handle points to a valid object, see if it is owned by the thread
//
 if ( (rc == 0) && ( iscurrthread(&gmqo??(handle??)) == FALSE ) ) rc = -97;

//
// Now GETMAIN the buffer to receive the data records
//
 if (( rc == 0) && ( bufflen != 0 ) )
   {
    DEBUG(traceid, "Doing malloc for " << bufflen << " bytes ") ;
    buffer = malloc(bufflen) ;
    if ( buffer == NULL ) mqac = errno ;
    DEBUG(traceid,"malloc RC " << mqac << " PTR is "<<hex<< buffer);
    if ( mqac != 0 ) rc = -8 ;
   }

//
// Now get the data from the queue
//
 if (rc == 0)
   {
    DEBUG(traceid, "BROWSE Maxdatalen is " << bufflen << " ") ;

    rc   = gmqo??(handle??).browse(bufflen,
                                   (char *)buffer,
                                   &buffcurlen)         ;
    mqrc = gmqo??(handle??).rc                          ;
    mqac = gmqo??(handle??).ac                          ;

    DEBUG(traceid, "GET RC " << mqrc
            << " AC "    << mqac
            << " Datalen " << buffcurlen << " " ) ;

    stem_from_long  (traceid, NULL, vdata, "0" , buffcurlen)                   ;
    if (buffcurlen > bufflen) buffcurlen = bufflen                         ;
    stem_from_bytes (traceid, NULL, vdata, "1" , (MQBYTE *)buffer, buffcurlen) ;
   }

//
// Free data buffer for stem.1 variable data, if allocated.
//
  if ( buffer != 0 )
    {
     DEBUG(traceid, "Freemaining area " << buffer) ;
     free(buffer) ;
    }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQBRWS function

//
// Do a Header extract  MQHXT
//
//   Call:   rc = RXMQhxt( input_Stem, output_stem )
//
//
//
//   This function takes an input Stem variable, and builds an
//        output stem variable, extracting the Headers from the
//        'message' data.
//
//        The actual message & its length are placed in .1 and .0 , with
//            all the other info being expanded into the .item variables.
//            In addition, for compatability with the EVENT splitting up
//            capability .NAME and .TYPE are also set (both to either
//            DLH or XQH). .ZLIST is also set to contain the exploded
//            components (including TYPE NAME 0 1).
//
//
//
//   The headers processed are:
//
//       XQH - Transmission Header
//
//             .0      -> Data Length
//             .1      -> Message Data
//
//             .RQM    -> Remote Queue Manager Name
//             .RQN    -> Remote Queue Name
//
//             .TYPE   -> XQH     for compatability with
//             .NAME   -> XQH    Event splitting up
//
//             + all the MD options
//
//             .ZLIST  -> 0 1 AID AOD AT BC CID CCSI ENC EXP FBK
//                        FORM MSG MSGID NAME PAN PAT PD PER PRI
//                        PT REP RQM RQN RTOQ RTOQM TYPE UID
//
//
//       DLH - Dead Letter Header
//
//             .0      -> Data Length
//             .1      -> Message Data
//
//             .TYPE   -> DLH     for compatability with
//             .NAME   -> DLH     Event splitting up
//
//             .REA    -> Reason
//             .DQM    -> Destination Queue Manager Name
//             .DQN    -> Destination Queue Name
//             .ENC    -> Encoding
//             .CCSI   -> Coded Character Set ID
//             .FORM   -> Format
//             .PAT    -> Put Appl Type
//             .PAN    -> Put Appl Name
//             .PD     -> Put Date
//             .PT     -> Put Time
//
//             .ZLIST  -> 0 1 CCSI DQM DQN ENC FORM NAME PAN
//                        PAT PD PT REA TYPE
//
//
FTYPE  RXMQHXT  RXMQPARM
 {

 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = HXT    ;  // This function trace id

 char                    header??(5??)    ;  // Header Type
 MQXQH                 * thexqh           ;  // -> XQH
 MQDLH                 * thedlh           ;  // -> DLH

 RXSTRING                vinp             ;  // Variable Data - Input
 MQLONG                  inp0             ;  // Input Data - len
 RXSTRING                vout             ;  // Variable Data - Output
 MQLONG                  out0             ;  // Output Data - len

 char                    zlist??(200??)   ;  // Char version of .ZLIST

 void                 *  data  = 0        ;  //-> Data buffer
 MQULONG                 datalen          ;  //   Data length

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null input stem var"},
        {  -3, "Zero input stem var"},
        {  -4, "Null output stem var"},
        {  -5, "Zero output stem var"},
        {  -6, "No input data"},
        {  -8, "Cannot verify Header"},
        { -10, "Unknown Header"},
        { -11, "Too short for a DLH"},
        { -12, "Too short for a XQH"},
        { -13, "malloc failure, check reason code"},
        { -14, "Data length is not equal to specified value"},
        { -97, "Handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

 sprintf((char *)zlist,"0 1") ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 2 ) )                 rc = -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc = -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc = -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc = -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc = -5 ;

//
// The input parms are OK, so obtain them
//
 if (rc == 0)
   {
    memcpy(&vinp,&aargv??(0??),sizeof(vinp)) ;
    memcpy(&vout,&aargv??(1??),sizeof(vout)) ;
    stem_to_long  (traceid, vinp, "0" , &inp0)   ;
   }

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && ( inp0 <= 0 ) ) rc = -6  ; // No input data

//
// Allocate data buffer to store stem.1 variable data
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Doing malloc for " << inp0 << " bytes ") ;
    data = malloc(inp0) ;
    if ( data == NULL ) mqac = errno ;
    DEBUG(traceid,"malloc RC " << mqac << " PTR is "<<hex<< data);
    if ( mqac != 0 ) rc = -13 ;
   }

 if ( rc == 0 )
   {
    datalen = stem_to_data(traceid, vinp, "1", (MQBYTE *)data, inp0) ;
    DEBUG(traceid, "Value at " << vinp.strptr
            << " Length " << datalen
            << " Data <"  << data
            << "> " ) ;
   }

//
// Now check if stem.0 specifies the same value as stem.1 length.
// If not, don't know what to do.
//
 if ((rc == 0) && ( datalen != (MQULONG) inp0 ) ) rc = -14 ;

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && ( datalen <= 3 ) ) rc = -8 ; // Cannot verify header

//
// There is the possibility of a Header, so see if it is a known
//       one, and ignore it, if it is not one to process.
//
 if ( rc == 0 )
   {
    memset(header, 0, sizeof(header)                    )        ;
    strncpy(header, (const char *)data, sizeof(MQCHAR4) )        ;

    if (    ( strcmp(header, MQDLH_STRUC_ID) != 0 )
         && ( strcmp(header, MQXQH_STRUC_ID) != 0 ) ) rc = -10   ;
   }

//
// Lets's try DLH
//
 if ( ( rc == 0 ) && ( strcmp(header, MQDLH_STRUC_ID) == 0 ) )
   {
    if ( datalen < sizeof(MQDLH) ) rc = -11 ; // Too short for DLH
    else
      {
       DEBUG(traceid, "Unravelling a DLH"  ) ;
       thedlh = (MQDLH *)data            ; // Set Header pointer
       out0 = inp0 - sizeof(MQDLH)       ; // Calc actual Data length

       stem_from_long  (traceid, NULL, vout, "0" , out0)                               ;
       stem_from_bytes (traceid, NULL, vout, "1" , (MQBYTE *)data + sizeof(MQDLH), out0) ;

       DEBUG( traceid, " Msg  length " << inp0
                    << " DLH  length " << (long)sizeof(MQDLH)
                    << " Data length " << out0
                    << " " ) ;

       DEBUG( traceid, " Data area "     << hex << (void *)data
                    << " Start of DLH "  << thedlh
                    << " Start of data " << hex << (char *)data + sizeof(MQDLH)
                    << " <"              << hex << (char *)data + sizeof(MQDLH)
                    << "> " ) ;

       stem_from_string(traceid, zlist, vout, "TYPE" , "DLH",                   3               ) ;
       stem_from_string(traceid, zlist, vout, "NAME" , "DLH",                   3               ) ;
       stem_from_long  (traceid, zlist, vout, "REA"  , thedlh->Reason)                            ;
       stem_from_string(traceid, zlist, vout, "DQM"  , thedlh->DestQMgrName,    sizeof(MQCHAR48)) ;
       stem_from_string(traceid, zlist, vout, "DQN"  , thedlh->DestQName,       sizeof(MQCHAR48)) ;
       stem_from_long  (traceid, zlist, vout, "ENC"  , thedlh->Encoding)                          ;
       stem_from_long  (traceid, zlist, vout, "CCSI" , thedlh->CodedCharSetId)                    ;
       stem_from_string(traceid, zlist, vout, "FORM" , thedlh->Format,          sizeof(MQCHAR8) ) ;
       stem_from_long  (traceid, zlist, vout, "PAT"  , thedlh->PutApplType)                       ;
       stem_from_string(traceid, zlist, vout, "PAN"  , thedlh->PutApplName,     sizeof(MQCHAR28)) ;
       stem_from_string(traceid, zlist, vout, "PD"   , thedlh->PutDate,         sizeof(MQCHAR8) ) ;
       stem_from_string(traceid, zlist, vout, "PT"   , thedlh->PutTime,         sizeof(MQCHAR8) ) ;
       stem_from_string(traceid, zlist, vout, "ZLIST", zlist, strlen(zlist))                      ;

       DEBUG(traceid, "Unravelled the DLH" ) ;
      }
   }

 //
 // Lets's try XQH
 //
 if ( ( rc == 0 ) && ( strcmp(header, MQXQH_STRUC_ID) == 0 ) )
   {
    if ( datalen < sizeof(MQXQH) ) rc = -12 ; // Too short for XQH
      {
       DEBUG(traceid, "Unravelling a XQH"  ) ;
       thexqh = (MQXQH *)data            ; // Set Header pointer
       out0 = inp0 - sizeof(MQXQH)       ; // Calc actual Data length

       stem_from_long  (traceid, NULL, vout, "0" , out0)                               ;
       stem_from_string(traceid, NULL, vout, "1" , (char *)data + sizeof(MQXQH), out0) ;
       DEBUG( traceid, " Msg  length " << inp0
                    << " XQH  length " << (long)sizeof(MQXQH)
                    << " Data length " << out0
                    << " " ) ;

       DEBUG( traceid, " Data area "     << hex << (void *)data
                    << " Start of XQH "  << thexqh
                    << " Start of data " << hex << (char *)data + sizeof(MQXQH)
                    << " <"              << hex << (char *)data + sizeof(MQXQH)
                    << "> " ) ;

       stem_from_string(traceid, zlist, vout, "TYPE" , "XQH",                   3               ) ;
       stem_from_string(traceid, zlist, vout, "NAME" , "XQH",                   3               ) ;
       stem_from_string(traceid, zlist, vout, "RQN"  , thexqh->RemoteQName,     sizeof(MQCHAR48)) ;
       stem_from_string(traceid, zlist, vout, "RQM"  , thexqh->RemoteQMgrName,  sizeof(MQCHAR48)) ;
       DEBUG(traceid, "Generating the XQH.MD" )             ;

       // Only Version 1 of MQMD is created here
       stem_from_long  (traceid, zlist, vout, "VER"  , thexqh->MsgDesc.Version)                               ;
       stem_from_long  (traceid, zlist, vout, "REP"  , thexqh->MsgDesc.Report)                                ;
       stem_from_long  (traceid, zlist, vout, "MSG"  , thexqh->MsgDesc.MsgType)                               ;
       stem_from_long  (traceid, zlist, vout, "EXP"  , thexqh->MsgDesc.Expiry)                                ;
       stem_from_long  (traceid, zlist, vout, "FBK"  , thexqh->MsgDesc.Feedback)                              ;
       stem_from_long  (traceid, zlist, vout, "ENC"  , thexqh->MsgDesc.Encoding)                              ;
       stem_from_long  (traceid, zlist, vout, "CCSI" , thexqh->MsgDesc.CodedCharSetId)                        ;
       stem_from_string(traceid, zlist, vout, "FORM" , thexqh->MsgDesc.Format,              sizeof(MQCHAR8))  ;
       stem_from_long  (traceid, zlist, vout, "PRI"  , thexqh->MsgDesc.Priority)                              ;
       stem_from_long  (traceid, zlist, vout, "PER"  , thexqh->MsgDesc.Persistence)                           ;
       stem_from_bytes (traceid, zlist, vout, "MSGID", thexqh->MsgDesc.MsgId,               sizeof(MQBYTE24)) ;
       stem_from_bytes (traceid, zlist, vout, "CID"  , thexqh->MsgDesc.CorrelId,            sizeof(MQBYTE24)) ;
       stem_from_long  (traceid, zlist, vout, "BC"   , thexqh->MsgDesc.BackoutCount)                          ;
       stem_from_string(traceid, zlist, vout, "RTOQ" , thexqh->MsgDesc.ReplyToQ,            sizeof(MQCHAR48)) ;
       stem_from_string(traceid, zlist, vout, "RTOQM", thexqh->MsgDesc.ReplyToQMgr,         sizeof(MQCHAR48)) ;
       stem_from_string(traceid, zlist, vout, "UID"  , thexqh->MsgDesc.UserIdentifier,      sizeof(MQCHAR12)) ;
       stem_from_bytes (traceid, zlist, vout, "AT"   , thexqh->MsgDesc.AccountingToken,     sizeof(MQBYTE32)) ;
       stem_from_string(traceid, zlist, vout, "AID"  , thexqh->MsgDesc.ApplIdentityData,    sizeof(MQCHAR32)) ;
       stem_from_long  (traceid, zlist, vout, "PAT"  , thexqh->MsgDesc.PutApplType)                           ;
       stem_from_string(traceid, zlist, vout, "PAN"  , thexqh->MsgDesc.PutApplName,         sizeof(MQCHAR28)) ;
       stem_from_string(traceid, zlist, vout, "PD"   , thexqh->MsgDesc.PutDate,             sizeof(MQCHAR8))  ;
       stem_from_string(traceid, zlist, vout, "PT"   , thexqh->MsgDesc.PutTime,             sizeof(MQCHAR8))  ;
       stem_from_string(traceid, zlist, vout, "AOD"  , thexqh->MsgDesc.ApplOriginData,      sizeof(MQCHAR4))  ;

       DEBUG(traceid, "End of XQH.MD generation")           ;
       stem_from_string(traceid, zlist, vout, "ZLIST", zlist, strlen(zlist))                      ;
       DEBUG(traceid, "Unravelled the XQH" ) ;
      }
   }

//
// Free data buffer for stem.1 variable data, if allocated.
//
 if ( data != 0 )
   {
    DEBUG(traceid, "Freemaining area " << data ) ;
    free(data) ;
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQHXT function

//
// Do an Event extract  RXMQEVNT
//
//   Call:   rc = RXMQevnt( input_Stem, output_stem )
//
//
//
//   This function takes an input Stem variable, and builds an
//        output stem variable, extracting the Event Data from the
//        'message' data.
//
//        Stem.TYPE variable is set to 'EVENT'
//        Stem.REA variable is set to the event numerical value
//        Stem.NAME variable is set to the name of the event
//        corresponding to the numerical value expressed as a
//        MQRC_... constant name with MQRC_ prefix stripped.
//        All other info is expanded into Stem.item variables
//        where item is parameter attribute name like MQCA_...
//        with MQRC_prefix stripped.
//
//        Additionally (.)ZLIST is set to contain a list of all the
//        components (without the leading dot).
//
//     The events are splitup into the normal stem.component name, but
//                if the component is not a SINGLE string or Integer (as
//                defined by the IN & ST PCF types for the given field), then
//                the explosion is stem.component.0 for the number of items
//                                 stem.component.n for the individual items
//
//     The naming of the fields within the event is controlled by the
//                geteventname routine, which converts a known set of
//                fields (as defined by the Event part of the PCF book)
//                into character component names. Note that wherever the
//                field occurs generates its expansion - no matter whether
//                or not the PCF book say it should or should not occur!
//
FTYPE  RXMQEVNT  RXMQPARM
 {

 MQLONG                  i                ;  // Looper
 MQLONG                  j                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = EVENT  ;  // This function trace id

 RXSTRING                vinp             ;  // Variable Data - Input
 MQLONG                  inp0             ;  // Input Data - len
 RXSTRING                vout             ;  // Variable Data - Output

 MQCFH                 * bufpcf1          ;  // Pointers to
 MQCFIN                * bufpcf2i         ;  //  PCF areas within the
 MQCFST                * bufpcf2s         ;  //  Event Message
 MQCFIL                * bufpcf2il        ;
 MQCFSL                * bufpcf2sl        ;

 MQLONG                  parms            ;  //Number of elements
 MQLONG                  parmtype         ;  //Type of element
 MQLONG                  parmsize         ;  //Length of element
 MQLONG                  parmnumb         ;  //Key of element
 MQCFSL                * sp               ;  //-> String element
 unsigned char           comp??(31??)     ;  //Component name

 char                    zvars??(900??)   ;  //List of components

 void                 *  data  = 0        ;  //-> Data buffer
 MQULONG                 datalen          ;  //   Data length

 RXSTRING                varname          ;  // REXX string of varnamc
 RXSTRING                varvalu          ;  // REXX string of varvalc
 char                    varnamc??(100??) ;  // Char version of variable name
 char                    varvalc??(100??) ;  // Char version of variable value

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null input stem var"},
        {  -3, "Zero input stem var"},
        {  -4, "Null output stem var"},
        {  -5, "Zero output stem var"},
        {  -6, "No input data"},
        {  -8, "Cannot verify Header"},
        { -10, "Not an Event Header"},
        { -11, "Too short for an Event"},
        { -12, "Unknown Event Category"},
        { -13, "Unknown Event Type"},
        { -16, "No elements in the Event"},
        { -97, "Handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 2 ) )                 rc = -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc = -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc = -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc = -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc = -5 ;

//
// The input parms are OK, so obtain them
//
 if (rc == 0)
   {
    memcpy(&vinp,&aargv??(0??),sizeof(vinp)) ;
    memcpy(&vout,&aargv??(1??),sizeof(vinp)) ;
    stem_to_long(traceid, vinp, "0" , &inp0) ;
   }

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && ( inp0 <= 0 ) ) rc = -6 ;

//
// Allocate data buffer to store stem.1 variable data
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Doing malloc for " << inp0 << " bytes ") ;
    data = malloc(inp0) ;
    if ( data == NULL ) mqac = errno ;
    DEBUG(traceid,"malloc RC " << mqac << " PTR is "<<hex<< data);
    if ( mqac != 0 ) rc = -15 ;
   }

 if (rc == 0)
   {
    datalen = stem_to_data(traceid, vinp, "1", (MQBYTE *)data, inp0) ;
    DEBUG(traceid, "Value at " << vinp.strptr
                << " Length "  << datalen
                << " Data <"   << data
                << "> " ) ;
   }

//
// Now check if stem.0 specifies the same value as stem.1 length.
// If not, don't know what to do.
//
 if ((rc == 0) && ( datalen != (MQULONG) inp0 ) ) rc = -16 ;

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && ( datalen <= 3 ) ) rc = -8 ; // Cannot verify header

//
// There is the possibility of a Header, so see if it is an event
//       one, and ignore it it is not one to process.
//
 if ( ( rc == 0 ) && ( ((MQCFH *)data)->Type != MQCFT_EVENT ) ) rc = -10 ;

//
// Although we have a valid header, just check the prefix length
//          to ensure the WHOLE header is present
//
 if ( ( rc == 0 ) && ( datalen < MQCFH_STRUC_LENGTH) ) rc = -11 ;

//
// Now there is an event Header, check the general category
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Determining Event general Category " ) ;

    bufpcf1   = (MQCFH  *)data                 ; //Point to
    bufpcf2i  = (MQCFIN *)&((char *)data)??(MQCFH_STRUC_LENGTH??) ; //Buffer
    bufpcf2s  = (MQCFST *)&((char *)data)??(MQCFH_STRUC_LENGTH??) ; //Structures
    bufpcf2il = (MQCFIL *)&((char *)data)??(MQCFH_STRUC_LENGTH??) ;
    bufpcf2sl = (MQCFSL *)&((char *)data)??(MQCFH_STRUC_LENGTH??) ;
    parms     = bufpcf1->ParameterCount ;

    switch ( bufpcf1->Command )  //See what the event category is
      {
       case MQCMD_Q_MGR_EVENT   : DEBUG(traceid,"It is a QMGR    Event ") ; break ;
       case MQCMD_PERFM_EVENT   : DEBUG(traceid,"It is a PERF    Event ") ; break ;
       case MQCMD_CHANNEL_EVENT : DEBUG(traceid,"It is a CHANNEL Event ") ; break ;
       case MQCMD_CONFIG_EVENT  : DEBUG(traceid,"It is a CONFIG  Event ") ; break ;
       case MQCMD_COMMAND_EVENT : DEBUG(traceid,"It is a COMMAND Event ") ; break ;
       case MQCMD_LOGGER_EVENT  : DEBUG(traceid,"It is a LOGGER  Event ") ; break ;
       default :
         {
          rc   = -12 ;
          break  ;
         }
       } // End of Category determination select
   }

//
// Now there is a valid event Header, decide which one it is
//     and then create the NAME (and TYPE=EVENT) components
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Starting to examine the Event ") ;

    sprintf(varnamc,"RXMQ.RCMAP.%lu",bufpcf1->Reason)              ; // Construct variable name
    MAKERXSTRING(varname,varnamc,strlen(varnamc))                  ; // Construct REXX variable name structure
    stem_to_string(traceid, varname, "", varvalc, sizeof(varvalc)) ; // Get variable value

    if ( strlen(varvalc) == 0) rc = -13;
    varvalu.strptr = strstr(varvalc, "_") + 1    ; // Bypass prefix
    varvalu.strlength = strlen(varvalu.strptr)   ; // New string length

    DEBUG(traceid, "Event " << bufpcf1->Reason
                << " maps to " << varvalu.strptr
                << ( (varvalu.strlength != 0) ? " "
                      : " which is unknown ") ) ;
   }

//
// There is something that looks like a valid event, but just
//       ensure there are some elements in it
//
if ( rc == 0 )
  {
   DEBUG(traceid,"There are " << bufpcf1->ParameterCount << " elements in the Event ") ;
   if ( bufpcf1->ParameterCount == 0 ) rc = -14 ;
  }

//
// Now there is a valid event , create the TYPE and NAME components,
//           as well as REAson from the Prefix Header. Also initialize
//           the zvars (to goto .ZLIST) component store.
//
if ( rc == 0 )
  {
   zvars??(0??) = '\0'                                                                ;
   stem_from_string(traceid, zvars, vout, "TYPE" , "EVENT", strlen("EVENT"))          ;
   stem_from_string(traceid, zvars, vout, "NAME" , varvalu.strptr, varvalu.strlength) ;
   stem_from_long  (traceid, zvars, vout, "REA"  , bufpcf1->Reason)                   ;
  }

//
// Now parse the event and create the relevant components
//
//               Build the ZLIST results into zvars for each parm
//
//               List entries are setup as stem.component.0 = number of items
//                                                       .n = the nth item
//
 if ( rc == 0 )                      // Scan and Extract
   {
    DEBUG(traceid, "Now starting to scan the Event ") ;

    memset(&comp,0,sizeof(comp));
    for (i=0 ; i < parms ; i++ )
      {
       parmtype = bufpcf2i->Type        ; //All types share a
       parmsize = bufpcf2i->StrucLength ; // common prefix
       parmnumb = bufpcf2i->Parameter   ;
       switch ( parmtype )
         {
          case MQCFT_INTEGER  : //Convert to attr=value
            geteventname((char *)comp,parmnumb)          ;
            switch ( parmnumb ) //Display in Hex or (mostly) decimal
              {
               case MQIACF_AUX_ERROR_DATA_INT_1      :
               case MQIACF_AUX_ERROR_DATA_INT_2      :
               case MQIACF_ERROR_IDENTIFIER          :
               case MQIACH_CHANNEL_ERROR_DATA        :
                 sprintf(varvalc,"%08lX",bufpcf2i->Value) ; // Display in HEX
                 DEBUG(traceid, " Integer Parm "     << bufpcf2i->Parameter
                             << " Value "            << bufpcf2i->Value
                             << " displayed as hex " << varvalc << " " ) ;
                 break ;

               default  :
                 sprintf(varvalc,"%lu",bufpcf2i->Value)  ; // Display in decimal
                 break ;
              }
            stem_from_string (traceid, zvars, vout, (char*)comp ,
                              varvalc, strlen((const char *)varvalc));
            DEBUG(traceid, " Integer Parm " << bufpcf2i->Parameter
                        << " Value "        << bufpcf2i->Value
                        << " into  "        << comp << " " ) ;
            break ;

          case MQCFT_INTEGER_LIST : //Convert to attr=value pairs
            if ( bufpcf2il->Count != 0 )
              {
               DEBUG(traceid, " Integer List " << bufpcf2il->Count
                           << " Values "  ) ;
               geteventname((char *)comp,bufpcf2il->Parameter)     ;
               sprintf(varnamc,"%s.0",comp)                        ;
               stem_from_long(traceid, zvars, vout, varnamc , bufpcf2il->Count) ;

               for ( j=0 ; j < bufpcf2il->Count ; j++ )
                 {
                  geteventname((char *)comp,bufpcf2il->Parameter)     ;
                  sprintf(varnamc,"%s.%lu",comp,(j+1))                ;
                  switch ( parmnumb ) //Display in Hex or (mostly) decimal
                    {
                     case MQIACF_AUX_ERROR_DATA_INT_1      :
                     case MQIACF_AUX_ERROR_DATA_INT_2      :
                     case MQIACF_ERROR_IDENTIFIER          :
                     case MQIACH_CHANNEL_ERROR_DATA        :
                       sprintf(varvalc,"%08lX",bufpcf2il->Values??(j??)) ; // HEX
                       DEBUG(traceid, " Integer List "     << bufpcf2il->Parameter
                                   << " Value "            << bufpcf2il->Values??(j??)
                                   << " displayed as hex " << varvalc << " " ) ;
                       break ;
                     default  :
                       sprintf(varvalc,"%lu",bufpcf2il->Values??(j??)) ; // Decimal
                       break ;
                    }
                  stem_from_string(traceid, zvars, vout, (char*)varnamc ,
                                   varvalc, strlen((const char *)varvalc));
                  DEBUG(traceid, " Integer List " << bufpcf2il->Parameter
                              << " Value "        << bufpcf2il->Values??(j??)
                              << " into  "        << comp << " " ) ;
                 }
              }
            break        ;

          case MQCFT_STRING : //Print in the Stem variable
            geteventname((char *)comp,parmnumb)                   ;
            stem_from_string(traceid, zvars, vout, (char*)comp,
                             (char *)&(bufpcf2s->String) , bufpcf2s->StringLength) ;
            DEBUG(traceid, " String Parm " << bufpcf2s->Parameter
                        << " Value "       << bufpcf2s->String
                        << " into  "       << comp << " " )       ;
            break                                                 ;

          case MQCFT_STRING_LIST : //Print Strings in the Stem Variable
            if ( bufpcf2sl->Count != 0 )
              {
               DEBUG(traceid, " String List " << bufpcf2sl->Count
                           << " Values "  ) ;
               geteventname((char *)comp,bufpcf2sl->Parameter)   ;
               sprintf(varnamc,"%s.0",comp)                      ;
               stem_from_long(traceid, zvars, vout, (char*)varnamc, bufpcf2sl->Count) ;

               for ( j=0 ; j < bufpcf2sl->Count ; j++ )
                 {
                  sp = bufpcf2sl + ( (bufpcf2sl->StringLength) * j)   ;
                  geteventname((char *)comp,bufpcf2sl->Parameter)     ;
                  sprintf(varnamc,"%s.%lu",comp,(j+1))                ;
                  stem_from_string(traceid, zvars, vout, (char*)varnamc,
                                  (char *)&(sp->Strings) , bufpcf2sl->StringLength);
                  DEBUG(traceid, " String List " << bufpcf2sl->Parameter
                              << " Value "       << sp->Strings
                              << " into  "       << comp << " " ) ;
                 }
              }
            break        ;
          default : break ;
         }

       bufpcf2i  = (MQCFIN *)( (char *)bufpcf2i  + parmsize ) ; //Bump
       bufpcf2s  = (MQCFST *)( (char *)bufpcf2s  + parmsize ) ; // up
       bufpcf2il = (MQCFIL *)( (char *)bufpcf2il + parmsize ) ; // structure
       bufpcf2sl = (MQCFSL *)( (char *)bufpcf2sl + parmsize ) ; // pointers

      } ; //End of Parameter Printing loop


// All Parms/Components extracted, so create ZLIST

    stem_from_string(traceid, NULL, vout, "ZLIST", (char *)&zvars , strlen((const char *)zvars));

    DEBUG(traceid, "All Event fields extracted ") ;
   } // End of Event Processing Block

//
// Free data buffer for stem.1 variable data, if allocated.
//
 if ( data != 0 )
   {
    DEBUG(traceid, "Freemaining area " << data ) ;
    free(data) ;
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQEVNT function


//
// Process a Trigger Message RXMQTM
//
//   Call:   rc = RXMQtm( input_Stem,   output_stem )
//
//           rc = RXMQtm( trig_message, output_stem )
//
//
//   This function works in two ways, depending upon whether or not the first
//        parameter ends with a dot (indicating a stem variable or some real
//        data).
//
//        If its a stem variable, then the usual .0 .1 convention applies and
//           and the message is deemed to be a Trigger Message derived from a
//           Queue. In this case, it is broken up into its components of
//           QN PN TD AT AID ED UD. These components are only generated if
//           the relevant fields are non-blank and non-zero. ZLIST processing
//           is provided for these components. Additionally, a component called
//           PL is created, which is a string (in TM2 format) suitable for
//           passing to an attached procedure (the QM field will be filled in
//           with the current Queue manager, if connected). PL is NOT part of
//           ZLIST (as it should not be fiddled about with in any way).
//
//        If its NOT a stem variable (not ending in a dot), then the input is
//           deamed to be the TM2 (PL) structure (which must be of the correct
//           length). In this case, the data (and its NOT a REXX variable) is
//           parsed for components QN PN TD AT ED UD QM with ZLIST processing;
//           if the item is all blanks (or all nulls!?!) then the component is
//           not built. Each component should be stripped of blanks before
//           usage in the Rexx environment.
//
//
FTYPE  RXMQTM  RXMQPARM
 {

 MQULONG                 i                ;  // Looper
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = TM     ;  // This function trace id

 MQCHAR48                qm               ;  // Connected QM

 char                    header??(5??)    ;  // Header  Type
 char                    versionc??(5??)  ;  // Version Type in Char
 MQLONG                  version          ;  // Version Type

 MQTM                  * thetm            ;  // -> Trigger Message
 MQTMC2                * thetm2           ;  // -> Trigger Parm

 RXSTRING                oldtm2           ;  // Supplied Trigger Message
 MQTMC2                  newtm2           ;  // Build Trigger Parm

 int                     action = '1'     ;  // TM or TM2 processing

 RXSTRING                vinp             ;  // Variable Data - Input
 MQLONG                  inp0             ;  // Input Data - len
 RXSTRING                vout             ;  // Variable Data - Output

 char                    zlist??(200??)   ;  // Char version of .ZLIST

 void                 *  data  = 0        ;  //-> Data buffer
 MQULONG                 datalen          ;  //   Data length

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null input stem var"},
        {  -3, "Zero input stem var"},
        {  -4, "Null output stem var"},
        {  -5, "Zero output stem var"},
        {  -6, "No input data"},
        {  -7, "Zero input data"},
        {  -8, "Cannot locate Header"},
        {  -9, "Cannot find Header"},
        { -10, "Not a TM Header"},
        { -11, "Cannot find Header"},
        { -12, "Unknown Header"},
        { -13, "Unknown Version"},
        { -14, "Header mismatch (1<>1)"},
        { -15, "Version mismatch (1<>1)"},
        { -16, "Header mismatch (2<>C)"},
        { -17, "Version mismatch (2<>C)"},
        { -18, "Too short for a TM"},
        { -19, "Too short for a TMC"},
        { -20, "malloc failure, check reason code"},
        { -21, "Data length is not equal to specified value"},
        { -97, "Handle not owned by current thread"},
        { -98, "Not connected to a QM"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

 action = 1 ;
 zlist??(0??) = '\0'                      ;

 memset(qm,       ' ', sizeof(qm)       ) ;
 memset(&newtm2,  ' ', sizeof(newtm2)   ) ;

//
// Check the parms
//
 if ( (rc == 0) && (aargc != 2 ) )                 rc = -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc = -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc = -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc = -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc = -5 ;

//
// The input parms are OK, so obtain them, and obtain the resolutions
//
 if (rc == 0)
   {
    int dotat ;

    memcpy(&vinp,&aargv??(0??),sizeof(vinp)) ;
    memcpy(&vout,&aargv??(1??),sizeof(vout)) ;

    dotat = vinp.strlength - 1 ;

    if ( vinp.strptr??(dotat??) == '.' )
      {
       DEBUG(traceid , "Processing a Trigger Message" ) ;
       action = 1                                  ;
       stem_to_long  (traceid, vinp, "0" , &inp0)       ;

//
// Allocate data buffer to store stem.1 variable data
//
       if (( inp0 > 0 ) )
         {
          DEBUG(traceid, "Doing malloc for " << inp0 << " bytes ") ;
          data = malloc(inp0) ;
          if ( data == NULL ) mqac = errno ;
          DEBUG(traceid,"malloc RC " << mqac << " PTR is "<<hex<< data);
          if ( mqac != 0 ) rc = -20 ;
         }

       if (rc == 0 && (inp0 > 0))
         {
          datalen = stem_to_data(traceid, vinp, "1",(MQBYTE *)data, inp0) ;
          DEBUG(traceid , "Obtained the Trigger Message. Length " << inp0
                     << "/"       << datalen ) ;
          DEBUGW(traceid , write("Data /",6), write((const char *)data,inp0) ,
                           write("/\n",2) ) ;
         }
       if ((rc == 0) && ( datalen != (MQULONG) inp0 ) ) rc = -21 ;

      }
    else
      {    // vinp.strptr??(dotat??) != '.'
       DEBUG(traceid , "Processing Trigger Data"   ) ;
       action = 2                        ;
       oldtm2.strptr    = vinp.strptr    ;
       oldtm2.strlength = vinp.strlength ;
       DEBUG(traceid , "Obtained the Trigger Data. Length " << vinp.strlength ) ;
       DEBUGW(traceid , write("Data /",6) , write((const char *)vinp.strptr,vinp.strlength) ,
                   write("/\n",2 ) ) ;
      }
   }

//
// Now check the input Stem variable to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && (action == 1) && ( inp0 <= 0 ) )    rc = -6 ;
 if ( (rc == 0) && (action == 1) && ( datalen == 0 ) ) rc = -7 ;
 if ( (rc == 0) && (action == 1) && ( inp0 <= 3 ) )    rc = -8 ;
 if ( (rc == 0) && (action == 1) && ( datalen <= 3 ) ) rc = -9 ;

//
// Now check the input data to see that there is
//     some valid data to obtain
//
 if ( (rc == 0) && (action == 2) && ( oldtm2.strlength == 0 ) ) rc = -10 ;
 if ( (rc == 0) && (action == 2) && ( oldtm2.strlength <= 3 ) ) rc = -11 ;

//
// There is the possibility of an Trigger Message, so see if it is a known
//       one, and ignore it it is not one to process.
//
 if ( rc == 0 )
   {
    memset(header,    0, sizeof(header)   ) ;
    memset(&versionc, 0, sizeof(versionc) ) ;
    version = 0                             ;

    if ( action == 1 ) strncpy( header,   (const char *) data         , sizeof(MQCHAR4) ) ;
    else               strncpy( header,   (const char *)oldtm2.strptr , sizeof(MQCHAR4) ) ;

    if ( action == 1 ) version = ((MQTM *)data)->Version  ;
    else    strncpy( versionc, (const char *)&(((MQTMC2 *)oldtm2.strptr )->Version) , sizeof(MQCHAR4) ) ;

    DEBUG(traceid, "Action " << action
               << " Header /" << header
               << "/ Versionc /" << versionc
               << "/ Version /" << version << "/ "  ) ;

    if (    ( rc == 0 )
         && ( strcmp(header, MQTM_STRUC_ID ) != 0 )
         && ( strcmp(header, MQTMC_STRUC_ID) != 0 ) ) rc = -12 ;

    if (    ( rc == 0 )
         && ( version  != MQTM_VERSION_1 )
         && ( strncmp(versionc, MQTMC_VERSION_2, sizeof(MQCHAR4) ) != 0 ) ) rc = -13 ;
   }

//
// Now there is a trigger area of some sort, just check to see its the right one
//
 if ( ( rc == 0 ) && (action == 1) && ( strcmp(header, MQTM_STRUC_ID ) != 0 ) )  rc = -14 ;
 if ( ( rc == 0 ) && (action == 1) && ( version != MQTM_VERSION_1 ) )            rc = -15 ;
 if ( ( rc == 0 ) && (action == 2) && ( strcmp(header, MQTMC_STRUC_ID ) != 0 ) ) rc = -16 ;
 if ( ( rc == 0 ) && (action == 2) && ( strncmp(versionc, MQTMC_VERSION_2, sizeof(MQCHAR4) ) != 0 ) ) rc = -17 ;

//
// Although we have a valid item, just check the lengths
//          to ensure the WHOLE thing is present
//
 if ( ( rc == 0 ) && ( action == 1 ) && ( datalen < sizeof(MQTM) ) )             rc = -18 ;
 if ( ( rc == 0 ) && ( action == 2 ) && ( oldtm2.strlength <  sizeof(MQTMC2) ) ) rc = -19 ;

//
// Now we have got a valid Trigger Message, split it up
//
if ( ( rc == 0 ) && ( action == 1 ) )
  {
   DEBUG(traceid, "Unravelling a TM message"  ) ;

   DEBUG(traceid, "Finding the QM "  ) ;

   if ( isthreadconn() != FALSE ) for ( i=MINQS ; i <= MAXQS ; i++ )
     {
      if (    ( iscurrthread(&gmqo??(i??)) != FALSE )
           && ( gmqo??(i??).qm??(0??)      != 0     ) )
        {
         DEBUG(traceid, "Found QM in Object " << i << " "      ) ;
         memcpy(qm, &gmqo??(i??).qm, strlen(gmqo??(i??).qm) ) ;
         break      ;
        }
     }

   DEBUG(traceid, "QM name is /" << qm << "/ " ) ;

   DEBUG(traceid, "Now starting the unpacking ") ;
   thetm = (MQTM *) data; // Set Header pointer

                         // Build the output TMC2 area Header

   memcpy(&newtm2.StrucId,     MQTMC_STRUC_ID,     sizeof(MQCHAR4)  ) ;
   memcpy(&newtm2.Version,     MQTMC_VERSION_2,    sizeof(MQCHAR4)  ) ;
   memcpy(&newtm2.QName,       thetm->QName,       sizeof(MQCHAR48) ) ;
   memcpy(&newtm2.ProcessName, thetm->ProcessName, sizeof(MQCHAR48) ) ;
   memcpy(&newtm2.TriggerData, thetm->TriggerData, sizeof(MQCHAR64) ) ;
   memcpy(&newtm2.ApplId,      thetm->ApplId,      sizeof(MQCHAR256)) ;
   memcpy(&newtm2.EnvData,     thetm->EnvData,     sizeof(MQCHAR128)) ;
   memcpy(&newtm2.UserData,    thetm->UserData,    sizeof(MQCHAR128)) ;
   memcpy(&newtm2.QMgrName,    qm,                 sizeof(qm)       ) ;

                         //Build the components

   DEBUG(traceid, "Looking at QN  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm->QName,sizeof(MQCHAR48)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "QN"  , thetm->QName,          sizeof(MQCHAR48)) ;

   DEBUG(traceid, "Looking at PN  ") ;
   DEBUGW(traceid , write("Data /",6),  write((const char *)thetm->ProcessName,sizeof(MQCHAR48)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "PN"  , thetm->ProcessName,          sizeof(MQCHAR48)) ;

   DEBUG(traceid, "Looking at TD  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm->TriggerData,sizeof(MQCHAR64)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "TD"  , thetm->TriggerData,          sizeof(MQCHAR64)) ;

   DEBUG(traceid, "Looking at AT  ") ;
   stem_from_long  (traceid, zlist, vout, "AT"  , thetm->ApplType)      ;

   DEBUG(traceid, "Looking at AID ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm->ApplId,sizeof(MQCHAR256)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "AID" , thetm->ApplId,              sizeof(MQCHAR256)) ;

   DEBUG(traceid, "Looking at ED  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm->EnvData,sizeof(MQCHAR128)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "ED"  , thetm->EnvData,             sizeof(MQCHAR128)) ;

   DEBUG(traceid, "Looking at UD  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm->UserData,sizeof(MQCHAR128)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "UD"  , thetm->UserData,            sizeof(MQCHAR128)) ;

   DEBUG(traceid, "Looking at ZLIST ") ;
   stem_from_string(traceid, zlist, vout, "ZLIST", zlist, strlen(zlist))                         ;

   DEBUG(traceid, "Setting the PL ") ;
   stem_from_string(traceid, zlist, vout, "PL"  , (MQCHAR *)&newtm2,sizeof(MQTMC2)) ;

   DEBUG(traceid, "Unravelled the TM" ) ;
  }

//
// Now we have got valid Trigger Data, split it up
//
if ( ( rc == 0 ) && ( action == 2 ) )
  {
   DEBUG(traceid, "Unravelling Trigger Data"  ) ;

   thetm2 = (MQTMC2 *)oldtm2.strptr   ; // Set Header pointer

//Build the components

   DEBUG(traceid, "Doing QN  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm2->QName,sizeof(MQCHAR48)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "QN"  , thetm2->QName,          sizeof(MQCHAR48)) ;

   DEBUG(traceid, "Doing PN  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm2->ProcessName,sizeof(MQCHAR48)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "PN"  , (char*)thetm2->ProcessName, sizeof(MQCHAR48)) ;

   DEBUG(traceid, "Doing TD  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm2->TriggerData,sizeof(MQCHAR64)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "TD"  , thetm2->TriggerData,          sizeof(MQCHAR64)) ;

   DEBUG(traceid, "Doing AID ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm2->ApplId,sizeof(MQCHAR256)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "AID" , thetm2->ApplId,              sizeof(MQCHAR256)) ;

   DEBUG(traceid, "Doing ED  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm2->EnvData,sizeof(MQCHAR128)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "ED"  , thetm2->EnvData,             sizeof(MQCHAR128)) ;

   DEBUG(traceid, "Doing UD  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm2->UserData,sizeof(MQCHAR128)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "UD"  , thetm2->UserData,            sizeof(MQCHAR128)) ;

   DEBUG(traceid, "Doing QM  ") ;
   DEBUGW(traceid , write("Data /",6) , write((const char *)thetm2->QMgrName,sizeof(MQCHAR48)) ,
                    write("/\n",2) ) ;
   stem_from_string(traceid, zlist, vout, "QM"  , thetm2->QMgrName,             sizeof(MQCHAR48)) ;

   DEBUG(traceid, "Doing ZLIST ") ;
   stem_from_string(traceid, zlist, vout, "ZLIST", zlist, strlen(zlist))                         ;

   DEBUG(traceid, "Unravelled the Trigger Data" ) ;
  }

//
// Free data buffer for stem.1 variable data, if allocated.
//
 if ( data != 0 )
   {
    DEBUG(traceid, "Freemaining area " << data ) ;
    free(data) ;
   }

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
 } // End of RXMQTM function


#ifdef __MVS__
//
// Execute a Command (use MQSC interface for MVS)
//
//   Call:   rc = RXMQC(parms, input_command, output_response )
//
FTYPE RXMQC  RXMQPARM
{

 MQULONG                 count   = 0      ;  // Responses received
 MQULONG                 linecnt = 0      ;  // Line count in reply msg
 MQULONG                 lines   = 0      ;  // Lines processed in a group
 MQLONG                  rc      = 0      ;  // Function Return Code
 MQLONG                  mqrc    = 0      ;  // MQ RC or CSQN205I rc
 MQLONG                  mqac    = 4      ;  // MQ AC or CSQN205I rsn
 MQULONG                 traceid = COM    ;  // This function trace id

 RXSTRING                vparm            ;  // Variable Parms
 RXSTRING                vcomm            ;  // Variable Command
 RXSTRING                vresp            ;  // Variable Reply Stem Var

 char                 *  buffer   =     0 ;  //-> Data buffer
 MQLONG                  bufflen  = 15000 ;  // Default buffer length
 MQLONG                  reclen   =     0 ;  // Received record length

 MQOD                    cod              ;  // Command queue object descriptor
 MQOD                    rod              ;  // Reply queue object descriptor
 MQMD2                   md               ;  // Message descriptor for PUT & GET
 MQPMO                   pmo              ;  // PUT message options
 MQGMO                   gmo              ;  // GET message options

 rahmqntr      cmq                        ; // MQ instances - command Queue
 rahmqntr      rmq                        ; //              - replyto Queue
 int           wasconnected               ; // Was it connected before ?

 char        qm   ??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name
 char        cq   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - command
 char        rq   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - replyToq
 MQBYTE24    CorrelMsg                         ; //PUT MsgId = GET CorrelId
 char        var  ??(10??)                     ;
 MQLONG      to   = 5000                       ; //Timeout for MQ Get

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null parms"},
        {  -3, "Zero parms"},
        {  -4, "Null command var"},
        {  -5, "Zero command var"},
        {  -6, "Null response stem var"},
        {  -7, "Zero length response stem var"},
        {  -8, "No command supplied"},
        {  -9, "Too big a Command supplied"},
        { -10, "malloc failure, check reason code"},
        { -11, "Connect to QMgr failed, check rc/rsn"},
        { -12, "Open command queue failed, check rc/rsn"},
        { -13, "Open response queue failed, check rc/rsn"},
        { -14, "Put command to queue failed, check rc/rsn"},
        { -15, "Get response from queue failed, check rc/rsn"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

 if ( (rc == 0) && (aargc != 3 ) )                 rc = -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc = -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc = -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc = -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc = -5 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc = -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc = -7 ;

 if (rc == 0)
   {
    memcpy(&vparm,&aargv??(0??),sizeof(vparm)) ; //QM Name (or stem.)
    memcpy(&vcomm,&aargv??(1??),sizeof(vcomm)) ; //Command to issue
    memcpy(&vresp,&aargv??(2??),sizeof(vresp)) ; //Return stem.

    DEBUG(traceid,  "vparm   " << vparm.strptr   ) ;
    DEBUG(traceid,  "vcomm   " << vcomm.strptr   ) ;
    DEBUG(traceid,  "vresp   " << vresp.strptr   ) ;

    memset(qm,0,MQ_Q_MGR_NAME_LENGTH+1     ) ; //Clear data areas
    memset(cq,0,MQ_Q_NAME_LENGTH+1         ) ;
    memset(rq,0,MQ_Q_NAME_LENGTH+1         ) ;

    strcpy(cq,"SYSTEM.COMMAND.INPUT" ) ; // and set defaults
    strcpy(rq,"SYSTEM.COMMAND.REPLY.MODEL" ) ;

 //If the given variable is not a stem. variable, then take the
 //   quicker option of assuming it's the Queue Manager name

    if ( vparm.strptr??((strlen((const char *)vparm.strptr)-1)??) != '.' )
      {
       memcpy(qm, vparm.strptr, strlen((const char *)vparm.strptr) ) ;
       DEBUG(traceid, "Queue Manager " << qm  )      ;
      }

 //The given variable is a stem. variable, so get its contents
    else
      {
       stem_to_string(traceid, vparm, "QM"  , qm, MQ_Q_MGR_NAME_LENGTH) ;
       stem_to_string(traceid, vparm, "CQ"  , cq, MQ_Q_NAME_LENGTH    ) ;
       stem_to_string(traceid, vparm, "RQ"  , rq, MQ_Q_NAME_LENGTH    ) ;
       stem_to_long  (traceid, vparm, "TO"  , &to                     ) ;
      }

    DEBUG(traceid, "QM is " << qm   )          ;
    DEBUG(traceid, "CQ is " << cq   )          ;
    DEBUG(traceid, "RQ is " << rq   )          ;
    DEBUG(traceid, "Command is <" << vcomm.strptr << ">" )  ;

   }

 if ( (rc == 0 ) && ( vcomm.strlength == 0 ) )                     rc = -8 ;
 if ( (rc == 0 ) && ( vcomm.strlength > MQ_COMMAND_MQSC_LENGTH ) ) rc = -9 ;

//
//The return stem variable is initially set to no info
//
 if ( rc == 0 ) stem_from_long (traceid, NULL, vresp, "0" , 0)                 ;

// 1) Connect to queue manager, if not yet connected

 if (rc == 0)
   if (isthreadconn() == FALSE)
     {
      wasconnected = 0          ;
      DEBUG (traceid, "Connecting to QM " << qm ) ;
      rc   = cmq.conn(qm)       ;
      DEBUG( traceid, "RC=" << rc ) ;

      if ( rc == 0 ) rmq.qmh = cmq.qmh ;
      else
        {
         rc = -11               ;
         mqrc = cmq.rc          ;
         mqac = cmq.ac          ;
        }
     }
   else
     {
      wasconnected = 1             ;
      MQHCONN qmh = isthreadconn() ;
      rmq.qmh = cmq.qmh = qmh      ;
     }


// 2) Open the command queue for PUT access

 if ( rc == 0 )
   {
    memcpy(&cod, &od_default, sizeof(MQOD))     ;
    strncpy(cod.ObjectQMgrName, qm, strlen(qm)) ;
    strncpy(cod.ObjectName, cq, strlen(cq))     ;

    DEBUG ( traceid, "Opening Command Q " << cq << " for Output " ) ;
    rc   = cmq.open(&cod, MQOO_OUTPUT) ;
    DEBUG ( traceid, "RC=" << rc ) ;
    if ( rc != 0 )
      {
       rc = -12                ;
       mqrc = cmq.rc           ;
       mqac = cmq.ac           ;
      }
   }

// 3) Open/Create the ReplyToQ for Get access using model

 if ( rc == 0)
   {
    memcpy(&rod, &od_default, sizeof(MQOD));
    strcpy(rod.ObjectName, rq)             ;
    strcpy(rod.DynamicQName,"RXMQ.*")      ;

    DEBUG(traceid,"Opening Response Q by model " << rq << " for Destructive access " ) ;
    rc   = rmq.open(&rod,MQOO_INPUT_SHARED);
    DEBUG(traceid, "RC=" << rc ) ;
    if ( rc != 0 )
      {
       rc = -13              ;
       mqrc = rmq.rc         ;
       mqac = rmq.ac         ;
      }
    else
      {
       memcpy(rq, rod.ObjectName, MQ_Q_NAME_LENGTH);
       DEBUG(traceid, "The ReplyToQ name is " << rq )  ;
      }
   }

// 4) Put command on command queue

 if ( rc == 0)
   {
    memcpy(&md,  &md_default,  sizeof(MQMD2))            ;
    md.MsgType = MQMT_REQUEST                            ;
    memcpy(md.Format, MQFMT_STRING, sizeof(MQCHAR8))     ;
    memcpy(md.ReplyToQ, rq, MQ_Q_NAME_LENGTH)            ;
    memcpy(md.ReplyToQMgr, qm, MQ_Q_MGR_NAME_LENGTH)     ;

    memcpy(&pmo, &pmo_default, sizeof(MQPMO))            ;
    pmo.Options = MQPMO_NO_SYNCPOINT
                | MQPMO_DEFAULT_CONTEXT                  ;

    DEBUG(traceid, "Now issuing the MQPUT to the Command Queue") ;
    rc   = cmq.put( &md, &pmo, vcomm.strlength, vcomm.strptr);
    DEBUG (traceid, "RC=" << rc ) ;

    cmq.close(MQCO_NONE)      ;

    if ( rc != 0 )
      {
       rc = -14               ;
       mqrc = cmq.rc          ;
       mqac = cmq.ac          ;
      }
   }

//
// Allocate data buffer to receive the ReplyToQ records
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Doing malloc for " << bufflen << " bytes ")  ;
    buffer = (char *) malloc(bufflen)                        ;
    if ( buffer == NULL )
      {
       mqac = errno                       ;
       DEBUG(traceid,"malloc RC " << mqac)    ;
       rc = -10                           ;
      }
   }

// 5) Get results

 if ( rc == 0 )
   {
    DEBUG(traceid, "Now starting to obtain the ReplyToQ messages");

    memcpy(CorrelMsg, md.MsgId, sizeof(MQBYTE24))             ;
    memcpy(&gmo, &gmo_default, sizeof(MQGMO));
    gmo.Options = MQGMO_NO_SYNCPOINT         |
                  MQGMO_WAIT                 ;
    gmo.WaitInterval =    to                 ;  // 5 sec

    while ( ( (mqrc == 0) && (mqac == 4) ) || (linecnt < lines) )
      {
       DEBUG(traceid, "Issuing a MQGET to the ReplyToQ " << rq ) ;
       memcpy(&md,  &md_default,  sizeof(MQMD2))             ;
       memcpy(md.CorrelId, CorrelMsg, sizeof(MQBYTE24))      ;
       memset(buffer, 0, bufflen)                            ;
       rc   = rmq.get(&md, &gmo, bufflen, buffer, &reclen )  ;
       DEBUG(traceid, "GET RC "   << rmq.rc
               << " AC "      << rmq.ac
               << " Datalen " << reclen
               << " " ) ;

       if (rc == MQRC_TRUNCATED_MSG_FAILED)            // buffer is too small
         {
          DEBUG(traceid, "Reallocating buffer ") ;
          free(buffer)                       ;         // free existing buffer
          rc = 0                             ;
          bufflen = bufflen * 2              ;         // double the size of buffer
          DEBUG(traceid, "Doing malloc for " << bufflen << " bytes ")  ;
          buffer = (char *) malloc(bufflen)                        ;
          if ( buffer == NULL )
            {
             mqac = errno                       ;
             DEBUG(traceid, "malloc RC " << mqac)   ;
             rc = -10                           ;
             break;
            }
          continue;
         }

    if (rc != 0)
      {
       rc = -15 ;
       mqrc = rmq.rc      ;
       mqac = rmq.ac      ;
       break;
      }

    if (!strncmp ("CSQN205I", (const char *)buffer, 8))
      {
       linecnt = 1;
       sscanf(((const char *) buffer)+17, "%8u", &lines) ;
       sscanf(((const char *) buffer)+34, "%8X", &mqrc) ;
       sscanf(((const char *) buffer)+51, "%8X", &mqac);
       DEBUG(traceid, "CSQN205I COUNT = " << lines <<
                  ", RETURN = " << mqrc <<
                  ", REASON = " << mqac) ;
      }
    else
      if ( lines == 0 )              // CSQN205I must be 1st response
        {
         DEBUG(traceid, "No CSQN205I response received ")   ;
         break;
        }
      else
        {
         count++   ;
         linecnt++ ;
         sprintf(var,"%d", count );
         stem_from_long  (traceid, NULL, vresp, "0" , count)                           ;
         stem_from_bytes (traceid, NULL, vresp, var , (unsigned char *)buffer, reclen) ;
        }
      }
   }

//
// Free data buffer, if allocated.
//
 if ( buffer != 0 )
   {
    DEBUG(traceid, "Doing free for buffer ") ;
    free(buffer) ;
   }

 rmq.close(MQCO_DELETE_PURGE);

 if( !wasconnected ) cmq.disc();

//
// Set the LAST variables, and the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

return 0;
}

#else
//
// Execute a Command (use PCF interface for Windows)
//
//   Call:   rc = RXMQC(parms, input_command, output_response )
//
FTYPE RXMQC RXMQPARM
{

 MQLONG                  rc      = 0      ;  // Function Return Code
 MQLONG                  mqrc    = 0      ;  // MQ RC
 MQLONG                  mqac    = 0      ;  // MQ AC
 MQULONG                 traceid = COM    ;  // This function trace id

 RXSTRING                vparm            ;  // Variable Parms
 RXSTRING                vcomm            ;  // Variable Command
 RXSTRING                vresp            ;  // Variable Reply Stem Var

 rahmqntr      cmq                          ; //MQ instances - command Queue
 rahmqntr      rmq                          ; //             - replyto Queue

 char        qm   ??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name
 char        cq   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - command
 char        rq   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - replyToq
 MQLONG      to   = 10000                      ; //Timeout for MQ Get (10 sec)

 char        pcfarea  ??(MQCFH_STRUC_LENGTH        +    //PCF area - header
                         MQCFIN_STRUC_LENGTH       +    //         - integer parm
                         MQCFST_STRUC_LENGTH_FIXED +    //         - string parm
                         MAXCOMMLEN+5             ??) ; //         - command

 MQCFH     * ptrpcf1 = (MQCFH  *) pcfarea                 ; //Set PCF area pointers
 MQCFIN    * ptrpcf2 = (MQCFIN *)&pcfarea??(MQCFH_STRUC_LENGTH??)    ;
 MQCFST    * ptrpcf3 = (MQCFST *)&pcfarea??(MQCFH_STRUC_LENGTH+MQCFIN_STRUC_LENGTH??) ;
 char      * ptrpcf4 = (char   *)&pcfarea??(MQCFH_STRUC_LENGTH+MQCFIN_STRUC_LENGTH+MQCFST_STRUC_LENGTH_FIXED??) ;

 MQCFH     * bufpcf1         ; //PCF pointers
 MQCFIN    * bufpcf2i        ; // for the returned
 MQCFST    * bufpcf2s        ; // info
 MQCFIL    * bufpcf2il       ;
 MQCFSL    * bufpcf2sl       ;

 MQLONG      pcflen   = 0    ; //Size of PCF to be sent
 MQLONG      recno   =  0    ; //Obtained record number

 MQOD                    od                   ;  //MQ Open Descriptor
 MQLONG                  gopts        = 0     ;  //MQ Get Message options
 MQLONG                  popts        = 0     ;  //MQ Put Message options
 MQGMO                   gmo                  ;  //MQ Get Message options
 MQPMO                   pmo                  ;  //MQ Put Message options
 void                 *  buffer       = 0     ;  //-> Data buffer
 MQLONG                  bufflen      = 10000 ;  //   Data length max

 char                    command??(MAXCOMMLEN+5??)  ; //Padded Command
 MQLONG                  commandlen                 ; // to send to QM

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null parms"},
        {  -3, "Zero parms"},
        {  -4, "Null command var"},
        {  -5, "Zero command var"},
        {  -6, "Null response stem var"},
        {  -7, "Zero length response stem var"},
        {  -8, "No command supplied"},
        {  -9, "Too big a Command supplied"},
        { -10, "malloc failure, check reason code"},
        { -11, "Connect to QMgr failed, check rc/rsn"},
        { -12, "Open command queue failed, check rc/rsn"},
        { -13, "Open response queue failed, check rc/rsn"},
        { -14, "Put command to queue failed, check rc/rsn"},
        { -15, "Get response from queue failed, check rc/rsn"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

//
// Check the parms
//

 if ( (rc == 0) && (aargc != 3 ) )                 rc = -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc = -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc = -3 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(1??)) )    rc = -4 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(1??)) ) rc = -5 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(2??)) )    rc = -6 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(2??)) ) rc = -7 ;

//
// Now the parms are correct, get them
//

 if (rc == 0)
   {
    memcpy(&vparm,&aargv??(0??),sizeof(vparm)) ; //QM Name (or stem.)
    memcpy(&vcomm,&aargv??(1??),sizeof(vcomm)) ; //Command to issue
    memcpy(&vresp,&aargv??(2??),sizeof(vresp)) ; //Return stem.

    DEBUG(traceid,  "vparm   " << vparm.strptr   ) ;
    DEBUG(traceid,  "vcomm   " << vcomm.strptr   ) ;
    DEBUG(traceid,  "vresp   " << vresp.strptr   ) ;


    memset(qm,0,MQ_Q_MGR_NAME_LENGTH+1     ) ; //Clear data areas
    memset(cq,0,MQ_Q_NAME_LENGTH+1         ) ;
    memset(rq,0,MQ_Q_NAME_LENGTH+1         ) ;

    strcpy(cq,"SYSTEM.ADMIN.COMMAND.QUEUE" ) ; // and set defaults
    strcpy(rq,"SYSTEM.MQSC.REPLY.QUEUE"    ) ;

 //If the given variable is not a stem. variable, then take the
 //   quicker option of assuming it's the Queue Manager name

    if ( vparm.strptr??((strlen((const char *)vparm.strptr)-1)??) != '.' )
      {
       memcpy(qm, vparm.strptr, strlen((const char *)vparm.strptr) ) ;
       DEBUG(traceid, "Queue Manager " << qm  )      ;
      }

 //The given variable is a stem. variable, so get its contents
    else
      {
       stem_to_string(traceid, vparm, "QM"  , qm, MQ_Q_MGR_NAME_LENGTH) ;
       stem_to_string(traceid, vparm, "CQ"  , cq, MQ_Q_NAME_LENGTH    ) ;
       stem_to_string(traceid, vparm, "RQ"  , rq, MQ_Q_NAME_LENGTH    ) ;
       stem_to_long  (traceid, vparm, "TO"  , &to                     ) ;
      }

    DEBUG(traceid, "QM is " << qm   )          ;
    DEBUG(traceid, "CQ is " << cq   )          ;
    DEBUG(traceid, "RQ is " << rq   )          ;
    DEBUG(traceid, "Command is <" << vcomm.strptr << ">" )  ;
   }

 //If no command, then nothing to do

 if ( (rc == 0 ) && ( vcomm.strlength == 0 ) ) rc = -8 ;

 //If command is too big, error

 if ( (rc == 0 ) && ( vcomm.strlength > MAXCOMMLEN ) ) rc = -9 ;

//Ensure Command is a multiple of 4 bytes long (PCF String requirement)

if ( rc == 0 )
  {
   memset(command,0,sizeof(command)) ;
   memcpy(command,vcomm.strptr,vcomm.strlength) ;
   switch ( vcomm.strlength % 4 )
     {
      case 1  : strcat(command,"   ") ; break ;
      case 2  : strcat(command,"  " ) ; break ;
      case 3  : strcat(command," "  ) ; break ;
      default : break                         ;
     }
   commandlen = strlen(command) ;
  }

//Set the Put options (for the Command) and Get options (for the reply)

 if ( rc == 0 )                      // Connect to QM
   {
    gopts  = MQGMO_NO_SYNCPOINT         +
             MQGMO_WAIT                 +
             MQGMO_ACCEPT_TRUNCATED_MSG +
             MQGMO_FAIL_IF_QUIESCING    ;

    popts  = MQPMO_NO_SYNCPOINT         +
             MQPMO_DEFAULT_CONTEXT      + //Needed to bluff PCF Security!
             MQPMO_FAIL_IF_QUIESCING    ;
   }

//
// Allocate data buffer to receive the ReplyToQ records
//
 if ( rc == 0 )
   {
    DEBUG(traceid, "Doing malloc for " << bufflen << " bytes ") ;
    buffer = malloc(bufflen) ;
    if ( buffer == NULL ) mqac = errno ;
    DEBUG(traceid,"malloc RC " << mqac << " PTR is "<<hex<< buffer);
    if ( mqac != 0 ) rc = -10 ;
   }

//The return stem variable is initially set to no info

 if ( rc == 0 ) stem_from_long  (traceid, NULL, vresp, "0" , 0);

//Connect to the QM, open the Command Queue, and create the ReplyToQ

 if ( rc == 0 )                      // Connect to QM
   {
    DEBUG (traceid, "Connecting to QM " << qm ) ;
    rc   = cmq.conn(qm)       ;
    mqrc = cmq.rc             ;
    mqac = cmq.ac             ;
    DEBUG( traceid, "RC=" << rc ) ;

    if ( rc == 0 ) rmq.qmh = cmq.qmh ;
    else rc = -11                    ;
   }

 if ( rc == 0 )                      // Open the command Q for Put access
   {
    DEBUG ( traceid, "Opening Command Q " << cq << " for Output " ) ;
    rc   = cmq.open(cq,(MQOO_OUTPUT+MQOO_FAIL_IF_QUIESCING))   ;
    mqrc = cmq.rc              ;
    mqac = cmq.ac              ;
    DEBUG ( traceid, "RC=" << rc ) ;
    if ( rc != 0 ) rc = -12    ;
   }

 if ( rc == 0 )                      // Open/Create the ReplyToQ for Get access
   {
    memset(&od,0,sizeof(od))                             ;
    memcpy(od.StrucId,MQOD_STRUC_ID,sizeof(od.StrucId))  ;
    od.Version        = MQOD_VERSION_1                   ;
    od.ObjectType     = MQOT_Q                           ;
    strcpy(od.ObjectName,rq)                             ;
    strcpy(od.DynamicQName,"RXMQC.*")                    ;

    DEBUG(traceid,"Opening Model ReplyToQ " << od.ObjectName << " for Destructive access " ) ;

    rc   = rmq.open(&od,(MQOO_INPUT_EXCLUSIVE+MQOO_SAVE_ALL_CONTEXT+MQOO_FAIL_IF_QUIESCING)) ;
    mqrc = rmq.rc            ;
    mqac = rmq.ac            ;
    DEBUG(traceid, "RC=" << rc ) ;

    if ( rc == 0 )
      {
       memset(&rq,0,sizeof(rq))                        ;
       strncpy(rq,od.ObjectName,sizeof(od.ObjectName)) ;
       DEBUG(traceid, "The ReplyToQ name is " << rq )      ;
      }
    else rc = -13 ;
  }

 //Build the PCF ESCAPE command

 if ( rc == 0 )                      // Now build the PCF command
   {
    DEBUG(traceid, "Now building the PCF area") ;
    memset(pcfarea,0,sizeof(pcfarea))             ;
    ptrpcf1->Type           = MQCFT_COMMAND       ; // Header - 2 parms
    ptrpcf1->StrucLength    = MQCFH_STRUC_LENGTH  ;
    ptrpcf1->Version        = MQCFH_VERSION_1     ;
    ptrpcf1->Command        = MQCMD_ESCAPE        ;
    ptrpcf1->MsgSeqNumber   = 1                   ;
    ptrpcf1->Control        = MQCFC_LAST          ;
    ptrpcf1->ParameterCount = 2                   ;
    ptrpcf2->Type           = MQCFT_INTEGER       ; // Parm1 is ESCAPE
    ptrpcf2->StrucLength    = MQCFIN_STRUC_LENGTH ;
    ptrpcf2->Parameter      = MQIACF_ESCAPE_TYPE  ;
    ptrpcf2->Value          = MQET_MQSC           ;
    ptrpcf3->Type           = MQCFT_STRING        ; // Parm2 is ESCAPE data
    ptrpcf3->StrucLength    = MQCFST_STRUC_LENGTH_FIXED + commandlen  ;
    ptrpcf3->Parameter      = MQCACF_ESCAPE_TEXT  ;
    ptrpcf3->CodedCharSetId = MQCCSI_DEFAULT      ;
    ptrpcf3->StringLength   = commandlen          ;
    memcpy(ptrpcf4,command,commandlen)            ;

    pcflen =   MQCFH_STRUC_LENGTH
             + MQCFIN_STRUC_LENGTH
             + MQCFST_STRUC_LENGTH_FIXED
             + commandlen                         ;

   }

// Now write the Command to the relevant Command Queue

 if ( rc == 0)                       //Now write to the Queue
   {
    DEBUG(traceid, "Now building the PMD") ;
    memset(&pmo,0,sizeof(pmo))  ;               //Set Put Message Options
    memcpy(&pmo.StrucId,MQPMO_STRUC_ID,sizeof(pmo.StrucId))          ;
    pmo.Version        = MQPMO_VERSION_1                             ;
    pmo.Options        = popts                                       ;

    memset(&cmq.pmd,0,sizeof(cmq.pmd))  ;     //Set Put Message Desc
    memcpy(&cmq.pmd.StrucId,MQMD_STRUC_ID,sizeof(cmq.pmd.StrucId))   ;
    memcpy(&cmq.pmd.Format ,MQFMT_ADMIN  ,sizeof(cmq.pmd.Format ))   ;
    cmq.pmd.Version        = MQMD_VERSION_1                          ;
    cmq.pmd.Report         = MQRO_NONE                               ;
    cmq.pmd.MsgType        = MQMT_REQUEST                            ;
    cmq.pmd.Persistence    = MQPER_NOT_PERSISTENT                    ;
    cmq.pmd.Expiry         = MQEI_UNLIMITED                          ;
    cmq.pmd.Feedback       = MQFB_NONE                               ;
    cmq.pmd.Encoding       = MQENC_NATIVE                            ;
    cmq.pmd.CodedCharSetId = MQCCSI_Q_MGR                            ;
    cmq.pmd.Priority       = MQPRI_PRIORITY_AS_Q_DEF                 ;
    memcpy(&cmq.pmd.ReplyToQ,rq,sizeof(cmq.pmd.ReplyToQ))            ;

    DEBUG(traceid, "Now issuing the MQPUT to the Command Queue") ;
    rc = cmq.put( &pmo, pcflen, (char *)ptrpcf1 )            ;
    mqrc = cmq.rc             ;
    mqac = cmq.ac             ;
    DEBUG (traceid, "RC=" << rc ) ;
    if ( rc != 0 ) rc = -14   ;
   }

 //Read the ReplyToQ, and display the contained messages

 if ( rc == 0 )                      // Read Q and Display
   {
    DEBUG(traceid, "Now starting to obtain the ReplyToQ messages") ;
    for ( recno = 1 ; ; recno++ )
      {
       char     exiter = 'N' ;
       MQLONG   reclen = 0   ;

       bufpcf1   = (MQCFH  *) buffer                         ; //Point to
       bufpcf2i  = (MQCFIN *)&((char *)buffer)??(MQCFH_STRUC_LENGTH??) ; //Buffer
       bufpcf2s  = (MQCFST *)&((char *)buffer)??(MQCFH_STRUC_LENGTH??) ; //Structures
       bufpcf2il = (MQCFIL *)&((char *)buffer)??(MQCFH_STRUC_LENGTH??) ;
       bufpcf2sl = (MQCFSL *)&((char *)buffer)??(MQCFH_STRUC_LENGTH??) ;

       memset(&gmo,0,sizeof(gmo))                               ; //Build the
       memcpy(&gmo.StrucId,MQGMO_STRUC_ID,sizeof(gmo.StrucId))  ; //GMO, allow
       gmo.Version        = MQGMO_VERSION_1                     ; //? seconds
       gmo.Options        = gopts                               ; //for the
       if ( to != 0 ) gmo.WaitInterval = to                     ; //reply
       else           gmo.WaitInterval = MQWI_UNLIMITED         ;

       memset(&rmq.gmd,0,sizeof(rmq.gmd))                              ;
       memcpy(&rmq.gmd.StrucId,MQMD_STRUC_ID,sizeof(rmq.gmd.StrucId))  ;
       rmq.gmd.Version        = MQMD_VERSION_1                         ;

       memset(buffer,0,bufflen)                                        ;
       reclen   = 0                                                    ;
       DEBUG(traceid, "Issuing a MQGET to the ReplyToQ " << rq )           ;
       rc = rmq.get ( &gmo, bufflen, (char *)buffer, &reclen )         ;
       mqrc = rmq.rc     ;
       mqac = rmq.ac     ;
       DEBUG(traceid, "GET RC "   << mqrc
                   << " AC "      << mqac
                   << " Datalen " << reclen
                   << " " ) ;

       switch ( rc ) //Print obtained message
         {
          case MQRC_NONE      :
            DEBUG(traceid, "Message "       << recno
                        << " = MQ Message " << bufpcf1->MsgSeqNumber
                        << ((bufpcf1->Control == MQCFC_LAST) ? " (last)" : " (more)")
                        << " has MQ RC "    << bufpcf1->Reason
                        << ", "             << bufpcf1->ParameterCount
                        << " parms" ) ;

            if ( bufpcf1->ParameterCount != 0 ) //Do not attempt to ignore null messages
              {                                // when printing the results
               MQLONG   parms           ;
               int      i               ;
               int      j               ;
               MQLONG   parmtype        ;
               MQLONG   parmsize        ;
               MQCFIL * ip              ;
               MQCFSL * sp              ;
               char     resvar??(100??) ;
               parms = bufpcf1->ParameterCount ;

               for (i=0 ; i < parms ; i++ )
                {
                 parmtype = bufpcf2i->Type        ; //All types share a
                 parmsize = bufpcf2i->StrucLength ; // common prefix
                 switch ( parmtype )
                   {
                    case MQCFT_INTEGER  : //Convert to attr=value
                      DEBUG(traceid, " Integer Parm " << bufpcf2i->Parameter
                                  << " Value "        << bufpcf2i->Value ) ;
                      break        ;

                    case MQCFT_INTEGER_LIST : //Convert to attr=value pairs
                      if ( bufpcf2il->Count != 0 )
                        {
                         DEBUG(traceid, " Integer parm " << bufpcf2il->Parameter
                                     << " Values "  ) ;
                         for ( j=0 ; j < bufpcf2il->Count ; j++ )
                           {
                            ip = bufpcf2il + (sizeof(MQLONG) * j) ;
                            DEBUG(traceid, ip->Values ) ;
                           }
                        }
                      break        ;

                    case MQCFT_STRING : //Print in the Stem variable
                      stem_from_long  (traceid, NULL, vresp, "0" , recno);
                      sprintf(resvar,"%ld", recno)                   ;
                      stem_from_bytes(traceid, NULL, vresp, resvar,
                                     (MQBYTE *)&(bufpcf2s->String) , bufpcf2s->StringLength) ;
                      break        ;

                    case MQCFT_STRING_LIST : //Print Strings in the Stem Variable
                      if ( bufpcf2sl->Count != 0 )
                        {
                         for ( j=0 ; j < bufpcf2sl->Count ; j++ )
                          {
                           sp = bufpcf2sl + ( (bufpcf2sl->StringLength) * j)   ;
                           stem_from_long  (traceid, NULL, vresp, "0" , recno)   ;
                           sprintf(resvar,"%ld", recno)                      ;
                           stem_from_bytes(traceid, NULL, vresp, resvar ,
                                          (MQBYTE *)&(sp->Strings), bufpcf2s->StringLength);
                           break                                      ;
                          }
                        }
                      break        ;
                    default : break ;
                   }

                 bufpcf2i  = (MQCFIN *)( (char *)bufpcf2i  + parmsize ) ; //Bump
                 bufpcf2s  = (MQCFST *)( (char *)bufpcf2s  + parmsize ) ; // up
                 bufpcf2il = (MQCFIL *)( (char *)bufpcf2il + parmsize ) ; // structure
                 bufpcf2sl = (MQCFSL *)( (char *)bufpcf2sl + parmsize ) ; // pointers

                }  //End of Parameter Printing loop
              }  //End of Parameter Listing

            if ( bufpcf1->Control == MQCFC_LAST) exiter = 'L' ; //Last means just that!
            break          ;

          default             :
            DEBUG(traceid, "MQGET on ReplyToQ error RC=" << rc <<  " on message " << recno ) ;
            rc = -15      ;
            exiter = 'G'  ;
            break         ;
           }

         if ( exiter != 'N' ) break ; //Get next message
                                      //unless error/EOQ occured
        } // End of Message Processing FOR loop
    } // End of Queue/Queue Processing Block

 DEBUG(traceid, "All ReplyToQ records obtained" ) ;

//End of processing

 DEBUG(traceid, "Closing the ReplyToQ " ) ; //Close/Delete the ReplyToQ
 rmq.close(MQCO_DELETE_PURGE)         ;

 DEBUG(traceid, "Closing the command Q " ) ; //Close the Command Queue
 cmq.close(MQCO_NONE)                  ;

 DEBUG(traceid, "Disconnecting from Command QM " ) ; //Disconnect from QM
 cmq.disc()                                    ;

//
// Free data buffer, if allocated.
//
 if ( buffer != 0 )
   {
    DEBUG(traceid, "Freemaining area " << buffer) ;
    free(buffer) ;
   }
//
// Set the function return string
//
 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

//
// Return to caller
//
return 0;

 } // End of RXMQC function
#endif

//
// Perform one of RXMQ operations  RXMQV
//
//   Call:   rc = RXMQV(<RXMQ operation name>, <RXMQ operation's parameters>)
//                   Available operations:
//
//                     INIT     ->  RXMQINIT, performs REXX environment initialization
//                     TERM     ->  RXMQTERM, performs REXX environment termination
//                     CONS     ->  RXMQCONS, setup MQ constants for REXX
//                     CONN     ->  RXMQCONN, connect to QManager
//                     OPEN     ->  RXMQOPEN, open MQ queue
//                     CLOSE    ->  RXMQCLOS, close MQ queue
//                     DISC     ->  RXMQDISC, disconnect from QManager
//                     CMIT     ->  RXMQCMIT, do a syncpoint
//                     BACK     ->  RXMQBACK, do a rollback
//                     PUT      ->  RXMQPUT,  do a MQPUT operation
//                     PUT1     ->  RXMQPUT1, do a MQPUT1 operation
//                     GET      ->  RXMQGET,  do a MQGET operation
//                     INQ      ->  RXMQINQ,  perform a MQINQ operation
//                     SET      ->  RXMQSET,  do a MQSET operation
//                     SUB      ->  RXMQSUB,  do a MQSUB operation
//                     BROWSE   ->  RXMQBRWS, do a MQBROWSE operation
//                     HXT      ->  RXMQHXT,  do a Header extract, MQHXT
//                     EVENT    ->  RXMQEVNT, extract the Event Data from the 'message' data
//                     TM       ->  RXMQTM,   process a Trigger Message
//
FTYPE RXMQV  RXMQPARM
{
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = MQV    ;  // This function trace id
 char                    name??(8??)      ;  // Uppercased function name
 MQULONG                 namelen          ;  // Name length

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null MQ function name"},
        {  -3, "Zero length MQ function name"},
        { -20, "Unknown MQ function name"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

 // Check the parms
 if ( (rc == 0) && (aargc == 0 ) )                 rc = -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc = -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc = -3 ;

 typedef struct
         {
          char * func_name;
          int (APIENTRY *func_ptr) RXMQPARM;

 } mqftypes;

 mqftypes funclist[] = {
          {"INIT"  , RXMQINIT},
          {"TERM"  , RXMQTERM},
          {"CONS"  , RXMQCONS},
          {"CONN"  , RXMQCONN},
          {"OPEN"  , RXMQOPEN},
          {"CLOSE" , RXMQCLOS},
          {"DISC"  , RXMQDISC},
          {"CMIT"  , RXMQCMIT},
          {"BACK"  , RXMQBACK},
          {"PUT1"  , RXMQPUT1},
          {"PUT"   , RXMQPUT},
          {"GET"   , RXMQGET},
          {"INQ"   , RXMQINQ},
          {"SET"   , RXMQSET},
#ifdef _v7
          {"SUB"   , RXMQSUB},
#endif
          {"BROWSE", RXMQBRWS},
          {"HXT"   , RXMQHXT},
          {"EVENT" , RXMQEVNT},
          {"TM"    , RXMQTM},
          {"?"     , NULL}  };

// Uppercase specified function name
 namelen = (aargv??(0??).strlength < 8) ? aargv??(0??).strlength : 8 ;
 for (MQULONG i = 0; i < namelen; i++)
   name??(i??) = toupper(aargv??(0??).strptr??(i??));
 name??(namelen??) = '\0' ;

// Find and call appropriate function
 if (rc == 0)  for (int i = 0; ; i++)
   {
    if (funclist[i].func_name[0] == '?') break;
    if ( ( strlen(funclist??(i??).func_name) == namelen ) &&
         ( memcmp(funclist??(i??).func_name, name, namelen) == 0) )
      return funclist??(i??).func_ptr(name, aargc-1, &(aargv??(1??)), aqname, aretstr);
   }

 rc = -20 ;

 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

 return 0;
}

//
// Perform one of RXMQ Command operations  RXMQVC
//
//   Call:   rc = RXMQVC(<RXMQ Command operation name>, <RXMQ Command operation's parameters>)
//                   Available operations:
//
//                     INIT     ->  RXMQINIT, performs REXX environment initialization
//                     TERM     ->  RXMQTERM, performs REXX environment termination
//                     COMMAND  ->  RXMQC,    implements MQ command line
//
FTYPE RXMQVC  RXMQPARM
{
 MQLONG                  rc = 0           ;  // Function Return Code
 MQLONG                  mqrc = 0         ;  // MQ RC
 MQLONG                  mqac = 0         ;  // MQ AC
 MQULONG                 traceid = MQV    ;  // This function trace id
 char                    name??(8??)      ;  // Uppercased function name
 MQULONG                 namelen          ;  // Name length

 RETMSG ReturnMsg[] = {
        {  -1, "Bad number of parms" },
        {  -2, "Null MQ function name"},
        {  -3, "Zero length MQ function name"},
        { -20, "Unknown MQ function name"},
        { -99, "UNKNOWN FAILURE"}} ;
 set_entry (afuncname, traceid)    ;

// Check the parms
 if ( (rc == 0) && (aargc == 0 ) )                 rc = -1 ;
 if ( (rc == 0) && RXNULLSTRING(aargv??(0??)) )    rc = -2 ;
 if ( (rc == 0) && RXZEROLENSTRING(aargv??(0??)) ) rc = -3 ;

 typedef struct
         {
          char * func_name;
          int (APIENTRY *func_ptr) RXMQPARM;
         } mqftypes;

 mqftypes funclist[] = {
          {"INIT"   , RXMQINIT},
          {"TERM"   , RXMQTERM},
          {"COMMAND", RXMQC},
          {"?"      , NULL}  };

// Uppercase specified function name
 namelen = (aargv??(0??).strlength < 8) ? aargv??(0??).strlength : 8 ;
 for (MQULONG i = 0; i < namelen; i++)
   name??(i??) = toupper(aargv??(0??).strptr??(i??));
 name??(namelen??) = '\0' ;

// Find and call appropriate function
 if (rc == 0) for (int i = 0; ; i++)
   {
    if (funclist[i].func_name[0] == '?') break;
    if ( ( strlen(funclist??(i??).func_name) == namelen ) &&
         ( memcmp(funclist??(i??).func_name, name, namelen) == 0) )
      return funclist??(i??).func_ptr(name, aargc-1, &(aargv??(1??)), aqname, aretstr);
    }

 rc = -20;

 set_return(rc,mqrc,mqac,afuncname,ReturnMsg,aretstr,traceid,"") ;

 return 0;
}

//
// Try to support long Windows REXX function names
//
#ifdef _RXMQN
FTYPE  RXMQNINIT  RXMQPARM
 {
  return RXMQINIT(afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNCONS  RXMQPARM
 {
  return RXMQCONS (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNTERM  RXMQPARM
 {
   return RXMQTERM (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNCONN  RXMQPARM
 {
  return RXMQCONN (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNOPEN  RXMQPARM
 {
  return RXMQOPEN (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNCLOSE  RXMQPARM
 {
  return RXMQCLOS (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNDISC  RXMQPARM
 {
  return RXMQDISC (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNCMIT  RXMQPARM
 {
  return RXMQCMIT (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNBACK  RXMQPARM
 {
  return RXMQBACK (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNPUT  RXMQPARM
 {
  return RXMQPUT (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNPUT1  RXMQPARM
 {
  return RXMQPUT1 (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNGET  RXMQPARM
 {
  return RXMQGET (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNINQ  RXMQPARM
 {
  return RXMQINQ (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNSET  RXMQPARM
 {
  return RXMQSET (afuncname,aargc,aargv,aqname,aretstr);
 }

#ifdef _v7
FTYPE  RXMQNSUB  RXMQPARM
 {
  return RXMQSUB (afuncname,aargc,aargv,aqname,aretstr);
 }
#endif

FTYPE  RXMQNBROWSE  RXMQPARM
 {
  return RXMQBRWS (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNHXT  RXMQPARM
 {
  return RXMQHXT (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNEVENT  RXMQPARM
 {
  return RXMQEVNT (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNTM  RXMQPARM
 {
  return RXMQTM (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQNC  RXMQPARM
 {
  return RXMQC (afuncname,aargc,aargv,aqname,aretstr);
 }
#endif

#ifdef _RXMQT
FTYPE  RXMQTINIT  RXMQPARM
 {
  return RXMQINIT(afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTCONS  RXMQPARM
 {
  return RXMQCONS (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTTERM  RXMQPARM
 {
  return RXMQTERM (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTCONN  RXMQPARM
 {
  return RXMQCONN (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTOPEN  RXMQPARM
 {
  return RXMQOPEN (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTCLOSE  RXMQPARM
 {
  return RXMQCLOS (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTDISC  RXMQPARM
 {
  return RXMQDISC (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTCMIT  RXMQPARM
 {
  return RXMQCMIT (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTBACK  RXMQPARM
 {
  return RXMQBACK (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTPUT  RXMQPARM
 {
  return RXMQPUT (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTPUT1  RXMQPARM
 {
  return RXMQPUT1 (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTGET  RXMQPARM
 {
  return RXMQGET (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTINQ  RXMQPARM
 {
  return RXMQINQ (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTSET  RXMQPARM
 {
  return RXMQSET (afuncname,aargc,aargv,aqname,aretstr);
 }

#ifdef _v7
FTYPE  RXMQTSUB  RXMQPARM
 {
  return RXMQSUB (afuncname,aargc,aargv,aqname,aretstr);
 }
#endif

FTYPE  RXMQTBROWSE  RXMQPARM
 {
  return RXMQBRWS (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTHXT  RXMQPARM
 {
  return RXMQHXT (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTEVENT  RXMQPARM
 {
  return RXMQEVNT (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTTM  RXMQPARM
 {
  return RXMQTM (afuncname,aargc,aargv,aqname,aretstr);
 }

FTYPE  RXMQTC  RXMQPARM
 {
  return RXMQC (afuncname,aargc,aargv,aqname,aretstr);
 }
#endif
